#!/bin/bash

# Must have sudo permissions to run

# SETUP ------------------------------------------

# This script inherits the following variables from configure-git-fusion.sh
# If running this script by itself, set values for these required variables

# P4 variables:
# P4PORT=
# P4CHARSET=

# Unix user for Git Fusion:
# GFUSER=

if [ "$P4PORT" == '' ]; then
	echo 'P4PORT is undefined'
	exit 1
fi
if [ "$P4CHARSET" == '' ]; then
	echo 'P4CHARSET is undefined'
	exit 1
fi
if [ "$GFUSER" == '' ]; then
	echo 'GFUSER is undefined'
	exit 1
fi

# ASSESS ENVIRONMENT -----------------------------

if [ -e /etc/httpd/conf.d/git-fusion-ssl.conf ]; then
    echo 'Apache is already configured for Helix Git Fusion.'
    exit 0
fi

if [ -e /etc/redhat-release ]; then
    CENTOS_VERSION=$(grep -oE '[0-9]+\.[0-9]+' /etc/redhat-release)
else
    echo 'Could not determine your OS version. Aborting...'
    exit 0
fi

if [ $(echo "$CENTOS_VERSION" | cut -d '.' -f 1) != '6' ]; then
    echo "Your CentOS version is $CENTOS_VERSION. HTTPS authentication for Helix Git Fusion can only be configured on CentOS 6.x. Exiting..."
    exit 0
fi

# Get GFUSER_HOME
GFUSER_HOME=$(getent passwd $GFUSER | cut -d ':' -f 6)
if [ ! -e "$GFUSER_HOME/.bashrc" ]; then
    echo "$GFUSER_HOME/.bashrc file is either not defined or does not exist."
    exit 1
fi

# Git Fusion requires valid LANG and LC_ALL variables to be set in ~git/.bashrc
GFUSER_LANG=$(grep "^\s*export\s\+LANG" "$GFUSER_HOME/.bashrc" | egrep -i '[.](utf-8|utf8)$' | sed 's;^export LANG=;;')
GFUSER_LC_ALL=$(grep "^\s*export\s\+LC_ALL" "$GFUSER_HOME/.bashrc" | egrep -i '[.](utf-8|utf8)$' |  sed 's;^export LC_ALL=;;')
if [ -z "$GFUSER_LANG" ] || [ -z "$GFUSER_LC_ALL" ]; then
    echo "LANG or LC_ALL variables are not set in $GFUSER_HOME/.bashrc"
    exit 1
fi

# Check that P4GF_ENV file is there
GFUSER_P4GF_ENV=$(grep "^\s*export\s\+P4GF_ENV" "$GFUSER_HOME/.bashrc" | sed 's;^export P4GF_ENV=;;')
# In case path is something like ~git/p4gf_environment.cfg, eval forces tilde interpretation
GFUSER_P4GF_ENV=$(eval "echo $GFUSER_P4GF_ENV")
if [ ! -e "$GFUSER_P4GF_ENV" ]; then
    echo "P4GF_ENV file is either not defined or does not exist."
    exit 1
fi

# Check if Apache is installed or if the Apache directory exists. If yes, quit
APACHE_PKG=$(rpm -qa 'httpd' 2>&1 | grep 'httpd')
if [ -n "$APACHE_PKG" ]; then
    MSG_PKG='Apache is already installed! Git Fusion cannot configure HTTPS authentication. You must uninstall Apache before proceeding.'
fi
if [ -e '/etc/httpd' ]; then
    MSG_DIR='The Apache configuration directory /etc/httpd already exists. This directory must be removed before proceeding.'
fi
if [ -e '/etc/sysconfig/httpd' ]; then
    MSG_SYSCONFIG='The Apache configuration file /etc/sysconfig/httpd already exists. This directory must be removed before proceeding.'
fi
if [ "$MSG_PKG" ] || [ "$MSG_DIR" ] || [ "$MSG_SYSCONFIG" ]; then
    if [ "$MSG_PKG" ]; then
        echo "$MSG_PKG"
    fi
    if [ "$MSG_DIR" ]; then
        echo "$MSG_DIR"
    fi
    if [ "$MSG_SYSCONFIG" ]; then
        echo "$MSG_SYSCONFIG"
    fi
    exit 0
else
    echo 'This script will now configure HTTPS authentication for Helix Git Fusion using Apache.
It will download and install Apache 2 and the "mod_authnz_external" module. If you have existing Apache configuration, it may be overwritten.
It is not recommended that you use this Apache server for any other purpose than for use with Helix Git Fusion.'
fi

# CONFIG -----------------------------------------

# Install EPEL repository for mod_authnz_external
MOD_AUTHNZ_AVAILABLE=$(yum list available 2>&1 | grep 'mod_authnz_external')
if [ -n "$MOD_AUTHNZ_AVAILABLE" ]; then
    echo 'Required package mod_authnz_external is available for installation.'
else
    echo 'Required package mod_authnz_external is not available for installation.'
    echo 'This script will now install EPEL repository.'
    wget -O /tmp/epel-release-latest-6.noarch.rpm http://dl.fedoraproject.org/pub/epel/epel-release-latest-6.noarch.rpm > /dev/null
    if [ -e  /tmp/epel-release-latest-6.noarch.rpm ]; then
        echo 'EPEL was successfully downloaded'
    else
        echo 'Could not download EPEL, required for Apache installation!'
        exit 1
    fi
    rpm -Uvh /tmp/epel-release-latest-6.noarch.rpm > /dev/null
    if [ $? -ne 0 ]; then
        echo 'EPEL could not be installed!'
        rm -rf /tmp/epel-release-latest-6.noarch.rpm
        exit 1
    else
        echo 'Installed EPEL repository, verifying...'
        rm -rf /tmp/epel-release-latest-6.noarch.rpm
        EPEL_REPO_ENABLED=$(yum list available 2>&1 | grep 'mod_authnz_external')
        if [ -n "$EPEL_REPO_ENABLED" ]; then
            echo 'Package mod_authnz_external from the EPEL repository is available for installation.'
        else
            echo 'Package mod_authnz_external is not available.'
            echo 'EPEL repo could not be enabled.'
            echo 'This script will attempt to recover by upgrading SSL certificates.'
            yum upgrade -y ca-certificates --disablerepo=epel
            if [ $? -eq 0 ]; then
                echo 'The certificates were upgraded successfully. Verifying EPEL repo...'
                EPEL_REPO_ENABLED_AFTER_CERT_UPGRADE=$(yum list available 2>&1 | grep 'mod_authnz_external')
                if [ -n "$EPEL_REPO_ENABLED_AFTER_CERT_UPGRADE" ]; then
                    echo 'Package mod_authnz_external from the EPEL repository is available for installation.'
                else
                    yum remove -y "epel-release-*" --disablerepo=epel
                    if [ $? -eq 0 ]; then
                        echo 'EPEL repo was removed successfully.'
                    else
                        echo 'EPEL repo could not be removed, please uninstall it manually before re-running this script.'
                    fi
                    exit 1
                fi
            else
                yum remove -y "epel-release-*" --disablerepo=epel
                if [ $? -eq 0 ]; then
                    echo 'EPEL repo was removed successfully.'
                else
                    echo 'EPEL repo could not be removed, please uninstall it manually before re-running this script.'
                fi
                echo 'The SSL certificates could not be upgraded.'
                exit 1
            fi
        fi
    fi
fi

# Install Apache and necessary modules
yum install -y httpd mod_ssl mod_authnz_external
ret=$?
APACHE_PKG=$(rpm -qa 'httpd' 2>&1 | grep 'httpd')
if [ -n "$APACHE_PKG" ] && [ $ret -eq 0 ]; then
    echo 'Apache was installed successfully'
else
    echo 'Apache or one of its components could not be installed!'
    exit 1
fi

if ! grep -q 'ServerName GF' /etc/httpd/conf/httpd.conf; then
    echo 'ServerName GF' >> /etc/httpd/conf/httpd.conf
fi

service httpd stop > /dev/null

 # Disable SELinux for the current session
setenforce 0 > /dev/null 2>&1 || true

if [ $? -ne 0 ]; then
    echo 'SELinux could not be disabled!'
    exit 1
else
    echo 'SELinux has been disabled'
fi

# Delete default ssl.conf if it exists
if [ -e /etc/httpd/conf.d/ssl.conf ]; then
    rm -rf /etc/httpd/conf.d/ssl.conf
    if [ $? -ne 0 ]; then
        echo 'Could not delete default configuration file /etc/httpd/conf.d/ssl.conf'
        exit 1
    fi
fi

# Create the VirtualHost from the template delivered with Helix Git Fusion
cp /opt/perforce/git-fusion/libexec/centos-git-fusion-ssl.conf.template /etc/httpd/conf.d/git-fusion-ssl.conf
sed -i "s|HOME /opt/perforce/git-fusion/home/perforce-git-fusion|HOME $GFUSER_HOME|" /etc/httpd/conf.d/git-fusion-ssl.conf
sed -i "s|LANG en_US.UTF-8|LANG $GFUSER_LANG|" /etc/httpd/conf.d/git-fusion-ssl.conf
sed -i "s|LC_ALL en_US.UTF-8|LC_ALL $GFUSER_LC_ALL|" /etc/httpd/conf.d/git-fusion-ssl.conf
sed -i "s|P4GF_ENV /opt/perforce/git-fusion/home/perforce-git-fusion/p4gf_environment.cfg|P4GF_ENV $GFUSER_P4GF_ENV|" /etc/httpd/conf.d/git-fusion-ssl.conf
sed -i "s/ssl:myperforceserver:port p4charset/$P4PORT $P4CHARSET/" /etc/httpd/conf.d/git-fusion-ssl.conf
if [ -e /etc/httpd/conf.d/git-fusion-ssl.conf ]; then
    echo 'Apache site configuration for Helix Git Fusion has been created'
else
    echo 'Could not create Apache site configuration for Helix Git Fusion!'
    exit 1
fi

# Configure Apache to find Git Fusion tools and to run as git:perforce
cat >> /etc/sysconfig/httpd <<TEXT
# Find tools needed by Git Fusion, such as Python3 and Git
export PATH=/opt/perforce/git-fusion/bin:\$PATH
TEXT
env_vars_ret_1=$?
GFUSER_GROUP=$(groups "$GFUSER" | cut -d ':' -f 2 | awk '{ print $1 }')
sed -i "s|User apache|User $GFUSER|" /etc/httpd/conf/httpd.conf
env_vars_ret_2=$?
sed -i "s|Group apache|Group $GFUSER_GROUP|" /etc/httpd/conf/httpd.conf
env_vars_ret_3=$?
if [[ $env_vars_ret_1 -ne 0 || $env_vars_ret_2 -ne 0 || $env_vars_ret_3 -ne 0 ]]; then
    echo 'Errors while updating Apache environment configuration!'
    exit 1
else
    echo 'Apache environment configuration updated'
fi

# If P4D is ssl-enabled, trust it
P4SSL='^ssl'
P4HOST=(${P4PORT//:/ })
if [[ ${P4HOST[0]} =~ $P4SSL ]] ; then
    echo "Trusting P4 server as user $GFUSER"
    su - "$GFUSER" -c "p4 -p $P4PORT trust -yf" 2>&1
    if [ $? -ne 0 ]; then
        echo "Unable to trust the server $P4PORT as user $GFUSER"
        exit 1
    fi
fi

# Create authentication script from template
cp /opt/perforce/git-fusion/libexec/p4auth.sh.template /opt/perforce/git-fusion/libexec/p4auth.sh
sed -i "s|export P4TRUST=/opt/perforce/git-fusion/home/perforce-git-fusion/.p4trust|export P4TRUST=$GFUSER_HOME/.p4trust|" /opt/perforce/git-fusion/libexec/p4auth.sh
if [ -e /opt/perforce/git-fusion/libexec/p4auth.sh ]; then
    echo 'Created Helix Git Fusion authentication script'
else
    echo 'Helix Git Fusion authentication script could not be created!'
    exit 1
fi

# Generate self-signed certs
mkdir -p /etc/httpd/ssl
openssl req -x509 -subj '/C=US/ST=CA/L=Alameda/O=Perforce/OU=GF/CN=GF/E=git-fusion@perforce.com' -nodes -days 365 -newkey rsa:2048 -keyout /etc/httpd/ssl/apache.key -out /etc/httpd/ssl/apache.crt > /dev/null 2>&1
if [ $? -ne 0 ]; then
    echo 'Errors while creating self-signed certificates!'
    exit 1
else
    echo 'Self-signed certificates have been created'
fi

service httpd start > /dev/null 2>&1
if [ $? -ne 0 ]; then
    echo 'Apache could not start correctly!'
    exit 1
else
    echo 'Apache started successfully'
fi

cat << __SUCCESS_MESSAGE__
:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
::
::  Apache has been successfully configured for Helix Git Fusion.
::
::  Make sure that Git Fusion submit triggers are installed on your 
::  Helix Server and verify your configuration by running the following
::  command, where P4PASSWD is the password for the super user $P4USER:
::
::  $ curl -k --user $P4USER:P4PASSWD https://localhost/@info
::
:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
__SUCCESS_MESSAGE__