#! /usr/bin/env python3.3
"""Perform useful admin tasks.

Intially manage the Git Fusion prevent new session keys.
"""

import locale
import os
import sys
import datetime
from argparse import RawDescriptionHelpFormatter

import p4gf_env_config  # pylint: disable=unused-import
import p4gf_branch
import p4gf_create_p4
import p4gf_p4key
from   p4gf_l10n             import _, NTR
import p4gf_log
import p4gf_const
import p4gf_util

DESCRIPTION = _("""Tool for Git Fusion administration.

Disable, enable, and report on new repo sessions.

Use: p4gf_admin.py <subcommand> --help for subcommand options

""")


def user_has_admin(p4):
    """Does this user have admin permissions somewhere?"""
    r = p4.run(['protects'])
    for rr in r:
        if rr['perm'] == 'admin' or rr['perm'] == 'super':
            return True
    print("User '{}' requires admin permissions. Stopping".format(p4.user))
    sys.exit(os.EX_USAGE)


def noop():
    """noop"""
    print("noop: this function should not be called")


class Admin():

    """"Estimate the git and P4 disk size of cloning a new repo."""

    def __init__(self, args):
        self.debug                                                = args.debug
        p4gf_const.P4GF_USER                                      = args.p4user
        self.p4port                                               = args.p4port
        self.p4user                                               = args.p4user
        self.locale                                               = args.locale
        self.action                                               = args.action
        self.repo_name                                            = None
        self.server_id                                            = None
        self.all_instances                                        = None
        if self.action in ['disable-sessions', 'enable-sessions']:
            self.repo_name                                        = args.repo
            self.server_id                                        = args.server_id
            self.all_instances                                    = args.all_instances
                                                                  # convert 'action' name
                                                                  #    to legal function name
        self.action                                               = self.action.replace('-', '_')
        self.p4               = None
        self.funct            = getattr(self, self.action, noop)
        locale.setlocale(locale.LC_ALL, self.locale)

    def get_prevent_key(self):
        """Produce the prevent key based in this class's properties."""
        prevent_key = None
        if self.all_instances:
            prevent_key = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_ALL
        elif self.repo_name and self.server_id:
            prevent_key = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_INSTANCE_REPO.format(
                    server_id=self.server_id, repo_name=self.repo_name)
        elif self.repo_name:
            prevent_key = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_REPO + self.repo_name
        elif self.server_id:
            prevent_key = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_INSTANCE + self.server_id
        return prevent_key

    def report(self):
        """Report on this Git Fusion's prevented instances/repos."""
        prevent_keys = self.get_all_prevent_keys()
        if not prevent_keys:
            print("Disabled repo sessions:")
            print("    None")
        else:
            print("Disabled repo sessions:")
            for k in sorted(prevent_keys.keys()):
                print("  {} {}".format(k, prevent_keys[k]))

    def disable_sessions(self):
        """Disable new sessions by setting prevent key."""
        user_has_admin(self.p4)
        prevent_key = self.get_prevent_key()
        if prevent_key:
            ddn = str(datetime.datetime.now())
            p4gf_p4key.set(self.p4, prevent_key, ddn[:ddn.rfind('.')])

    def enable_sessions(self):
        """Enable new session by removing prevent key."""
        user_has_admin(self.p4)
        prevent_key = self.get_prevent_key()
        if prevent_key:
            if p4gf_p4key.is_set(self.p4, prevent_key):
                p4gf_p4key.delete(self.p4, prevent_key)
            else:
                print("enable_sessions: key '{}' not set - no need to unset".format(prevent_key))

    def run(self):
        """Call the requested method."""
        p4 = p4gf_create_p4.create_p4(port=self.p4port, user=self.p4user)
        self.p4 = p4
        if self.debug:
            print("p4 connection : p4port={} p4user={}".format(p4.port, p4.user))
        p4gf_branch.init_case_handling(p4)
        self.funct()

    def get_all_prevent_keys(self):
        """Return list of prevent keys."""
        keylist = p4gf_p4key.get_all(self.p4, p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS + '*')
        return keylist


def main():
    """Parse the command-line arguments and report on repo metrics."""
    # pylint: disable=too-many-branches
    parser = p4gf_util.create_arg_parser(
        desc=DESCRIPTION, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-d', '--debug', action=NTR("store_true"),
                        help=_('debugging'))
    parser.add_argument('-u', '--p4user', metavar='p4user',
                        help=_('Perforce user'))
    parser.add_argument('-p', '--p4port', metavar='p4port',
                        help=_('Perforce server'))
    parser.add_argument('--locale', metavar='locale', default='en_US.UTF-8',
                        help=_('system locale setting'))

    subparsers = parser.add_subparsers(dest='action')

    subparsers.add_parser('report', help='Report Git Fusion instance information')

    disable_parser = subparsers.add_parser(
        'disable-sessions',
        help=_("Disable new sessions for a repo or an instance or Git Fusion(all instances)"))
    disable_parser.add_argument('--repo', metavar='repo',
                                help=_('Repo name'))
    disable_parser.add_argument('--server-id', metavar='server-id',
                                help=_('Server Id of the Git Fusion instance'))
    disable_parser.add_argument('--all-instances', action=NTR("store_true"),
                                help=_('All Git Fusion instances'))

    enable_parser = subparsers.add_parser(
        'enable-sessions',
        help=_("Enable new sessions for a repo or an instance or Git Fusion(all instances)"))
    enable_parser.add_argument('--repo', metavar='repo',
                               help=_('Repo name'))
    enable_parser.add_argument('--server-id', metavar='server-id',
                               help=_('Server Id of the Git Fusion instance'))
    enable_parser.add_argument('--all-instances', action=NTR("store_true"),
                               help=_('All Git Fusion instances'))

    args = parser.parse_args()
    if args.action in ['disable-sessions', 'enable-sessions']:
        if args.all_instances and (args.server_id or args.repo):
            print("all-instances may not be combined with --server-id or --repo")
            sys.exit(os.EX_USAGE)

    if args.debug:
        print("args={}".format(args))

    # If connect args not passed, check that the environment is set.
    if not args.p4port:
        if 'P4PORT' not in os.environ and 'P4GF_ENV' not in os.environ:
            print(
              _("Neither --p4port is an argument nor are P4GF_ENV and P4PORT in the environment."))
            sys.exit(os.EX_USAGE)
        if 'P4PORT' in os.environ:
            args.p4port = os.environ['P4PORT']
            if args.debug:
                print("setting P4PORT={} from os.environ".format(args.p4port))
    else:
        # Set the requested port for Git Fusion's environment
        os.environ['P4PORT'] = args.p4port

    if not args.p4user:
        if 'P4USER' not in os.environ:
            # Defaulting to p4gf_const.P4GF_USER
            args.p4user = p4gf_const.P4GF_USER
        else:
            args.p4user = os.environ['P4USER']

            if args.debug:
                print("setting P4USER={} from os.environ".format(args.p4user))
    else:
        # Set the requested user for Git Fusion's environment
        os.environ['P4USER'] = args.p4user

    admin = Admin(args)
    admin.run()


if __name__ == "__main__":
    p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
