#! /usr/bin/env python3.3
"""Delete the local repository files on this Git Fusion instance.

Clears the cached objects directory, locks, and views for the named
repository (or all if using --all option). Also removes repo/instance
specific p4 keys.

"""

import logging
import os
import shutil
import sys

from P4 import P4Exception

import p4gf_const
import p4gf_create_p4
import p4gf_git_repo_lock
from p4gf_l10n import _, NTR
import p4gf_log
import p4gf_p4cache
import p4gf_p4key
import p4gf_proc
import p4gf_repo_dirs
import p4gf_translate
import p4gf_util

# cannot use __name__ since it will often be "__main__"
LOG = logging.getLogger("p4gf_clear_cache")


def _rmtree(path):
    """Quietly remove the named directory, without raising errors."""
    if os.path.exists(path):
        try:
            shutil.rmtree(path)
            LOG.debug('deleted directory %s', path)
        except:  # pylint: disable=bare-except
            pass


def _process(p4, server_id, repo_name):
    """Process the named repository."""
    # prevent access to this instance/repo combination
    prevent_key_name = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_INSTANCE_REPO.format(
        server_id=server_id, repo_name=repo_name)
    p4gf_p4key.increment(p4, prevent_key_name)

    # wait for access to this instance/repo
    with p4gf_git_repo_lock.write_lock(repo_name):
        LOG.debug('access to repository %s acquired', repo_name)

    # remove all repo-specific files and directories
    repo_dirs = p4gf_repo_dirs.from_p4gf_dir(p4gf_const.P4GF_HOME, repo_name)
    _rmtree(repo_dirs.repo_container)
    _rmtree(os.path.join(p4gf_const.P4GF_HOME, "branches", repo_name))
    _rmtree(os.path.join(p4gf_const.P4GF_HOME, "locks", repo_name))
    _rmtree(os.path.join(p4gf_const.P4GF_HOME, "objects", "repos", repo_name))
    _rmtree(os.path.join(p4gf_const.P4GF_HOME, "repos", repo_name))

    # remove all of the instance/repo specific keys
    key_name_templates = [
        p4gf_const.P4GF_P4KEY_LAST_COPIED_CHANGE,
        p4gf_const.P4GF_P4KEY_LAST_COPIED_TAG,
        p4gf_const.P4GF_P4KEY_LAST_SEEN_CHANGE,
        p4gf_const.P4GF_P4KEY_REPO_SERVER_CONFIG_REV,
    ]
    for tmplt in key_name_templates:
        key_name = tmplt.format(repo_name=repo_name, server_id=server_id)
        with p4.at_exception_level(p4.RAISE_NONE):
            if p4gf_p4key.delete(p4, key_name):
                LOG.debug('key %s deleted', key_name)
    cp_pattern = p4gf_const.P4GF_P4KEY_CHECKPOINT.format(
        repo_name=repo_name, server_id=server_id, name='*')
    cp_keys = p4gf_p4key.get_all(p4, cp_pattern)
    for key_name in cp_keys.keys():
        with p4.at_exception_level(p4.RAISE_NONE):
            if p4gf_p4key.delete(p4, key_name):
                LOG.debug('key %s deleted', key_name)

    # enable this instance/repo combination
    p4gf_p4key.delete(p4, prevent_key_name)


def main():
    """Check for missing cache objects, with the option to add them."""
    # pylint:disable=too-many-branches,too-many-statements
    p4gf_util.has_server_id_or_exit()
    desc = _("Delete the local repository files on this Git Fusion instance.")
    parser = p4gf_util.create_arg_parser(desc=desc)
    parser.add_argument(NTR("repo"), metavar=NTR("R"), nargs="*",
                        help=_("name of the repositories to be deleted"))
    parser.add_argument("-a", "--all", action="store_true",
                        help=_("process all known Git Fusion repositories"))
    args = parser.parse_args()

    # Check that either --all or "repo" was specified, but not both.
    if not args.all and len(args.repo) == 0:
        sys.stderr.write(_("Missing repo names; try adding --all option.\n"))
        sys.exit(2)
    if args.all and len(args.repo) > 0:
        sys.stderr.write(_("Ambiguous arguments. Choose --all or a repo name.\n"))
        sys.exit(2)

    p4gf_proc.init()
    with p4gf_create_p4.Closer():
        p4 = p4gf_create_p4.create_p4_temp_client()
        if not p4:
            sys.exit(1)
        # Sanity check the connection (e.g. user logged in?) before proceeding.
        try:
            p4gf_p4cache.fetch_client(p4)
        except P4Exception as e:
            sys.stderr.write(_("P4 exception occurred: {exception}\n").format(exception=e))
            sys.exit(1)
        if args.all:
            repos = p4gf_util.repo_config_list(p4)
            if len(repos) == 0:
                print(_("No Git Fusion repositories found, nothing to do."))
                sys.exit(0)
        else:
            repos = args.repo
        server_id = p4gf_util.get_server_id()
        for repo in repos:
            repo_name = p4gf_translate.TranslateReponame.git_to_repo(repo)
            print(_("Processing repository {repo_name}...").format(repo_name=repo_name))
            _process(p4, server_id, repo_name)


if __name__ == "__main__":
    p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
