#! /usr/bin/env python
"""Git Fusion Read only Reporting Tool."""
import sys
import uuid

# Determine python version
PYTHON3 = True
if sys.hexversion < 0x03000000:
    PYTHON3 = False

def NTR(x):         # pylint: disable=invalid-name
    """No-TRanslate: Localization marker for string constants."""
    return x
_ = NTR

P4PORT = None
P4PASS = 0
P4FAIL = 1
DEBUG = False

# For unicode servers uncomment the following line
# CHARSET = ['-C', 'utf8']
CHARSET = []

# Create a unique client name to use with 'p4' calls.
# The client is not required, but it prevents failures
# should the default client (hostname)
# be identical to a Perforce depot name.
P4CLIENT = "GF-ADMIN-{0}".format(str(uuid.uuid1())[0:12])

P4GF_CONFIG_GLOBAL                  = NTR('//{P4GF_DEPOT}/p4gf_config')
P4GF_CONFIG_REPO                    = NTR('//{P4GF_DEPOT}/repos/{repo_name}/p4gf_config')
P4GF_CONFIG_REPO2                   = NTR('//{P4GF_DEPOT}/repos/{repo_name}/p4gf_config2')

P4GF_P4KEY_LAST_COPIED_CHANGE  = \
    NTR('git-fusion-{repo_name}-{server_id}-last-copied-changelist-number')
P4USER=''
P4GF_P4_BIN='p4'
import os
import re

                        # pylint:enable=W9903



# regex
# Edit these as needed for non-English p4d error messages
NOLOGIN_REGEX         = re.compile(r'Perforce password \(P4PASSWD\) invalid or unset')
PERMISSION_REGEX         = re.compile(r'You don\'t have permission for this operation.')
CONNECT_REGEX         = re.compile(r'.*TCP connect to.*failed.*')
NOSUCHKEY_REGEX         = re.compile(r'No such key.*')
TRUST_REGEX  = re.compile(r"^.*authenticity of '(.*)' can't.*fingerprint.*p4 trust.*$",
                          flags=re.DOTALL)
TRUST_MSG  = _("""
\nThe Git Fusion trigger has not established trust with its ssl enabled server.
Contact your adminstrator and have them run {command}""").format(command=NTR("'p4 trust'."))


OBJPATH_COMMIT_REGEX = re.compile("/objects/repos/(?P<repo>[^/]+)/commits/(?P<slashed_sha1>[^-]+)"
                                  "-(?P<branch_id>[^,]+),(?P<change_num>\\d+)")
REPO=0
SHA1=1
CHANGE=2
BRANCH=3
def commit_from_filepath(filepath):
    """Take a (client or depot) file path and parse off the "xxx-commit-nnn" suffix.

    Return an ObjectType instance.
    """
    #pylint: disable=unused-variable
    m = OBJPATH_COMMIT_REGEX.search(filepath)
    if not m:
        return (None, None)
    sha1       = m.group(NTR('slashed_sha1')).replace('/', '')
    repo       = m.group(NTR('repo'))
    change_num = m.group(NTR('change_num'))
    branch_id  = m.group(NTR('branch_id'))
    return (repo, sha1, change_num, branch_id)

def fix_text_encoding(text):
    """Ensure the string is encodable using the stdout encoding.

    :param str text: string to be re-coded, if necessary.
    :return: a safe version of the string which can be printed.

    """
    encoding = sys.stdout.encoding
    if encoding is None:
        encoding = sys.getdefaultencoding()
    if not PYTHON3:
        try:
            text = unicode(text, encoding, 'replace')  # pylint:disable=undefined-variable
        except TypeError:
            # If text itself is unicode, the unicode() will fail
            # thus text is already unicode
            pass
    try:
        text.encode(encoding)
        return text
    except UnicodeEncodeError:
        bites = text.encode(encoding, 'backslashreplace')
        result = bites.decode(encoding, 'strict')
        return result


def print_log(msg):
    """Print and optionally log to debug log."""
    msg = fix_text_encoding(msg)
    print(msg)


# import there here to avoid unneeded processing before the potential early exit above
import marshal
from   subprocess import Popen, PIPE, check_output
import getopt

# pylint: disable=invalid-name
# Invalid constant name

P4GF_OBJECTS_ROOT                   = NTR('//{P4GF_DEPOT}/objects')
P4GF_DEPOT              = NTR('.git-fusion')


def _encoding_list():
    """Return a list of character encodings, in preferred order.

    Use when attempting to read bytes of unknown encoding.
    """
    # latin_1 decodes anything, so keep it last in the list.
    return ['utf8', 'latin_1']


def encode(data):
    """Attempt to encode using one of several code encodings."""
    if not PYTHON3:
        return data

    for encoding in _encoding_list():
        try:
            s = data.encode(encoding)
            return s
        except UnicodeEncodeError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            print_log(str(e))
    # Give up, re-create and raise the first error.
    data.encode(_encoding_list[0])


def decode(bites):
    """Attempt to decode using one of several code pages."""
    for encoding in _encoding_list():
        try:
            s = bites.decode(encoding)
            return s
        except UnicodeDecodeError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            print_log(str(e))
    # Give up, re-create and raise the first error.
    bites.decode(_encoding_list[0])


def _convert_bytes(data):
    """For python3, convert the keys in maps from bytes to strings.

    Recurses through
    the data structure, processing all lists and maps. Returns a new
    object of the same type as the argument. Any value with a decode()
    method will be converted to a string.
    For python2 - return data
    """
    def _maybe_decode(key):
        """Convert the key to a string using its decode() method.

        If decode() not available return the key as-is.
        """
        return decode(key) if 'decode' in dir(key) else key

    if not PYTHON3:
        return data

    if isinstance(data, dict):
        newdata = dict()
        for k, v in data.items():
            newdata[_maybe_decode(k)] = _convert_bytes(v)
    elif isinstance(data, list):
        newdata = [_convert_bytes(d) for d in data]
    else:
        # convert the values, too
        newdata = _maybe_decode(data)
    return newdata


def p4_print(depot_path):
    """Accumulate multiple 'data' entries to assemble content from p4 print."""
    result = p4_run(['print', '-q', depot_path])
    contents = ''
    for item in result:
        if 'data' in item and item['data']:
            contents += item['data']
    return contents


_unicode_error = [{'generic': 36,
                   'code': NTR('error'),
                   'data': _('Unicode server permits only unicode enabled clients.\n'),
                   'severity': 3}]



def p4_run(cmd, stdin=None, user=P4USER, exit_on_error=True, client=P4CLIENT):
    """Use the -G option to return a list of dictionaries."""
    # pylint: disable=too-many-statements, too-many-branches
    user = P4USER
    raw_cmd = cmd

    # Data passed to stdin needs to be marshalled if it is form data.
    # Other stdin is expected to be plain text
    stdin_data = None
    if stdin is not None and stdin is not PIPE:
        stdin_data = stdin
        if not isinstance(stdin_data, bytes):
            stdin_data = encode(stdin_data)
        stdin = PIPE

    global CHARSET
    while True:
        #cmd = [P4GF_P4_BIN, "-p", P4PORT, "-u", user, "-G" , "-c", client] + CHARSET + raw_cmd
        cmd = [P4GF_P4_BIN]
        if P4PORT:
            cmd = cmd + [ "-p", P4PORT]
        if P4USER:
            cmd = cmd + [ "-u", P4USER]
        cmd = cmd + ["-G" , "-c", client] + CHARSET + raw_cmd
        if DEBUG:
            print('p4_run: cmd : {0}'.format(cmd))
        try:
            process = Popen(cmd, shell=False, stdin=stdin, stdout=PIPE, stderr=PIPE)
            if stdin_data is not None:
                process.stdin.write(stdin_data)
                process.stdin.close()

        except (OSError, ValueError) as e:
            print_log(_("Error calling Popen with cmd: {command}").format(command=cmd))
            print_log(_("Error: {exception}").format(exception=e))
            sys.stdout.flush()
            sys.exit(P4FAIL)

        data = []

        try:
            while True:
                data.append(marshal.load(process.stdout))
        except EOFError:
            pass
        ret = process.wait()
        if data:
            data = _convert_bytes(data)
        if ret != 0:
            # check for unicode error:
            if (not CHARSET) and (not stdin) and data and data == _unicode_error:
                # set charset and retry
                CHARSET = ['-C', 'utf8']
                continue

            else:
                error = process.stderr.read().splitlines()
                if exit_on_error and error and len(error) > 1:
                    for err in error:
                        if CONNECT_REGEX.match(_convert_bytes(err)):
                            print_log(_("Cannot connect to P4PORT: {p4port}").format(p4port=P4PORT))
                            sys.stdout.flush()
                            os._exit(P4FAIL)    # pylint: disable=protected-access
            data.append({"Error": ret})
        break
    if exit_on_error and len(data) and 'code' in data[0] and data[0]['code'] == 'error':
        if NOLOGIN_REGEX.match(data[0]['data']):
            errdata = data[0]['data'].strip()
            print_log(_("\n        p4 error : {error_data}")
                      .format(error_data=errdata))
            p4tickets = os.environ['P4TICKETS'] if 'P4TICKETS' in os.environ else 'unset'
            # pylint: disable = line-too-long
            print_log(_("        User '{user}' is not logged in. "
                        "P4PORT={p4port}   P4TICKETS={p4tickets}")
                      .format(user=user, p4port=P4PORT,  p4tickets=p4tickets))
            sys.exit(P4FAIL)
        m = TRUST_REGEX.match(data[0]['data'])
        if m:
            print_log(TRUST_MSG)
            sys.exit(P4FAIL)
    if exit_on_error and ret:
        # decode the 'data' in the error message
        d = None
        if data and 'data' in data[0]:
            d = data[0]['data']
            d = d.decode() if isinstance(d, bytes) else str(d)
        # Perforce reports an error when deleting a non-existing key
        # Detect this case and return without error
        if d and NOSUCHKEY_REGEX.match(d):
            return None
        print_log(_("P4 error : P4PORT={p4port}").format(p4port=P4PORT))
        if data:
            if d:   # decoded 'data' ?
                print_log(d.strip())
                if PERMISSION_REGEX.match(d):
                    print_log(_("Check the Perforce protects table."))
            if 'Error' in data[0]:
                d = data[0]['Error']
                d = d.decode() if isinstance(d, bytes) else str(d)
                print_log(_("error={error}").format(error=d.strip()))
        print_log(_("Contact your administrator."))
        sys.exit(P4FAIL)
    return data



def get_p4key(name, user=P4USER):
    """Get p4key."""
    if not name:
        return None
    p4key = p4_run(['key',  name], user=user)[0]
    return p4key['value']


def get_configureable(name, user=P4USER):
    """Get configurable."""
    if not name:
        return None
    name_value = p4_run(['configure', 'show',  name], user=user)[0]
    return name_value['Value']


def git_to_repo(viewname, forward=True, translate_p4=True):
    """Translate  '/'   and optionally    ( ':'  and  ' ' )  ... as shown in translatables[].



    This supports git urls containing '/' and P4 clients containing ':' and ' '
    Example:   git clone:gitfusion@localhost:/my/teste:er/repo
               git           /my/teste:er/repo
               repo          my_0xS_test_0xC_er_0xS_repo
    Note: Leading slash in git name is stripped. So this clones the same repo
    Example:   git clone:gitfusion@localhost:my/teste:er/repo
               git           my/teste:er/repo
               repo          my_0xS_test_0xC_er_0xS_repo

    """
    # Slash to support git urls
    translatables = [['/', '_0xS_'], [' ','_0xSP_']]

    # these are the P4 client legal chars
    # which are not legal for the internal file path of the GF repos/...
    #  '_' mapping must be last of all target items containing '_' ]
    # p4_translatables = [ [':' ,  '_0xC_'] , [' '  , '_']]
    p4_translatables = [[':',  '_0xC_']]
    if translate_p4:
        translatables.extend(p4_translatables)

    translated = ''.join(viewname)  # copy
    for pair in translatables:
        if forward:
            translated = translated.replace(pair[0], pair[1])
        else:
            # Do not re-translate '_' back to SPACE
            if pair[0] == ' ':
                continue
            translated = translated.replace(pair[1], pair[0])

    return translated

def slashify_sha1(sha1):
    """Convert a SHA1 to the path form for use in Perforce.

    For instance, 60eaf72224a34f592636271fa957b6c4acaee5f3
    becomes 60/ea/f72224a34f592636271fa957b6c4acaee5f3
    which can then be used to build a file path.
    """
    if sha1 == '*':
        return '*/*/*'
    return sha1[:2] + "/" + sha1[2:4] + "/" + sha1[4:]

def objects_root():
    """Return //P4GF_DEPOT/objects."""
    return P4GF_OBJECTS_ROOT.format(P4GF_DEPOT=P4GF_DEPOT)

def commit_depot_path(commit_change, repo):
    """Return depot path to a commit."""

    if re.match(r'^[0-9]+$', commit_change):   # changenum
        object_path = (NTR('{objects_root}/repos/{repo}/commits/.../*-*,{change}')
            .format(objects_root=objects_root(),
                    repo=repo,change=commit_change))
    elif re.match(r'^[0-9a-zA-Z]{7,}$', commit_change): # sha1
        object_path =  (NTR('{objects_root}/repos/{repo}/commits/{slashed}*-*')
                .format(objects_root=objects_root(),
                        repo=repo,
                        slashed=slashify_sha1(commit_change)))
    else:
        print("Appears this value {0} is not a P4 change nor a sha1".
                format(commit_change))
        sys.exit(1)
    return object_path

def convert_sha1_changenum(commit_change, repo):
    """Look up changenum or sha1 in object cache and report the complement."""
    object_path = commit_depot_path(commit_change, repo)
    result = p4_run(['files', object_path])
    if 'code' in result[0] and result[0]['code'] == 'error':
        print ("No commit/changelist for repo={0} change/commit={1} ".format(repo, commit_change))
        sys.exit(0)
    commit_sha1_list = []
    for r in result:
        if 'depotFile' in r:
            df = r['depotFile']
            if DEBUG:
                print("depotFile={0}".format(df))
            commit_sha1_list.append(commit_from_filepath(df))
    return commit_sha1_list

def print_global_config():
    """Print global config files."""

    # Strip comment lines for printed config.

    config_path = '//' + P4GF_DEPOT + '/p4gf_config'

    # Exists p4gf_config?
    r = p4_run(['files', '-e',  config_path])
    if 'code' in r[0] and r[0]['code'] == 'error':
        print("No global p4gf_config.")
        return
    contents = p4_print(config_path)
    print("\nglobal p4gf_config:\n\n{0}".format(re.sub(r'^#.*\n',r'',contents,0,re.MULTILINE)))

def print_repo_configs(repo):
    """Print a repo's config files."""

    # Strip comment lines for printed config.

    repo_name = git_to_repo(repo)
    config_path = P4GF_CONFIG_REPO.format(
        P4GF_DEPOT=P4GF_DEPOT, repo_name=repo_name)
    config_path2 = P4GF_CONFIG_REPO2.format(
        P4GF_DEPOT=P4GF_DEPOT, repo_name=repo_name)

    # Exists p4gf_config?
    r = p4_run(['files', '-e',  config_path])
    if 'code' in r[0] and r[0]['code'] == 'error':
        print("No p4gf_config for repo '{0}'".format(repo))
        return
    contents = p4_print(config_path)
    print("\n'{0}' p4gf_config:\n\n{1}".format(repo, re.sub(r'^#.*\n',r'',contents,0,re.MULTILINE)))

    # Exists p4gf_config2?
    r = p4_run(['files', '-e',  config_path2])
    if 'code' in r[0] and r[0]['code'] == 'error':
        return
    contents = p4_print(config_path2)
    print(
       "\n'{0}' p4gf_config2:\n\n{1}".format(repo, re.sub(r'^#.*\n',r'',contents,0,re.MULTILINE)))

def print_last_commit(repo , server_id):
    """Print the p4 describe for the last changelist in this repo."""

    key_name = P4GF_P4KEY_LAST_COPIED_CHANGE.format(repo_name=repo, server_id=server_id)
    r = p4_run(['keys', '-e', key_name])
    if not r:
        print("This repo does not appear to exist.")
        sys.exit(0)
    pattern = 'git-fusion-{repo_name}-(.*)-last-copied-changelist-number'.format(repo_name=repo)
    sid_regex = re.compile(pattern)
    for rr in r:
        if 'value' in rr and 'key' in rr:
            last_change = rr['value']
            kname       =   rr['key']
            m           = sid_regex.match(kname)
            if m:
                server_id = m.group(1)
                print("\n==>  server_id = {0}   ...  last_change ={1}".
                        format(server_id, last_change))
                d = check_output(['p4', '-p', P4PORT, '-u', P4USER, 'describe', '-s', last_change])
                if not isinstance(d, str):
                    d = decode(d)
                print(d)


def print_repo_status(repo, do_all):
    """Display the latest or all push push stastus."""
    arg = 'git-fusion-status-' + repo
    if do_all:
        arg = arg + '*'
    cmd = ['keys', '-e', arg]
    result = p4_run(cmd)
    for rr in result:
        print("{0}   {1}".format(rr['key'], rr['value']))


def usage():
    """Display full usage."""
    _usage="""
p4gf_commit_query.py --repo <repo> --convert <sha1>|<changenum> [--port NNNN] [--user <user>]
p4gf_commit_query.py --repo <repo> --last-commit [--server-id <server-id> ] [--port NNNN] [--user <user>]
p4gf_commit_query.py --repo <repo> --push_status [--port NNNN] [--user <user>]
p4gf_commit_query.py --repo <repo> --push_status-all [--port NNNN] [--user <user>]
p4gf_commit_query.py --repo <repo> --config [--port NNNN] [--user <user>]
p4gf_commit_query.py --config [--port NNNN] [--user <user>]
p4gf_commit_query.py --help|-h
"""
    print(_usage)

def main():
    """Tool to provide various queries into a Git Fusion Helix server."""
    # pylint: disable=too-many-branches, too-many-statements, unused-variable
    short_opt = 'hdu:p:'
    long_opt = NTR(['repo=', 'convert=', 'port=', 'user=',
                'push-status', 'push-status-all', 'config',  'last-commit', 'server-id=', 'help'])
    try:
        options, _ = getopt.getopt(sys.argv[1:], short_opt, long_opt)
        if not options:
            usage()
            sys.exit(0)
    except getopt.GetoptError as err:
        print(_("Command line options parse error: {error}").format(error=err))
        usage()
        sys.exit(P4FAIL)
    global P4PORT, P4USER, DEBUG
    push_status    = False
    push_status    = False
    commit_change  = None
    last_commit    = False
    server_id      = '*'
    repo = None
    config = False
    if 'P4PORT' in os.environ:
        P4PORT = os.environ['P4PORT']
    if 'P4USER' in os.environ:
        P4USER = os.environ['P4USER']
    if  PYTHON3:
        p4set = check_output(['p4', 'set', '-q']).decode('cp437').split('\n')
    else:
        p4set = check_output(['p4', 'set', '-q']).split('\n')
    setdict = dict([x.split('=') for x in p4set if len(x)])
    if 'P4PORT' in setdict:
        P4PORT = setdict['P4PORT']
    if 'P4USER' in setdict:
        P4USER = setdict['P4USER']
    for o, a in options:
        if o == '--repo':
            repo = a
        elif o == '--config':
            config = True
        elif o == '--convert':
            commit_change = a
        elif o == '--port' or o == '-p':
            P4PORT = a
        elif o == '--user' or o == '-u':
            P4USER = a
        elif o == '--push-status':
            push_status=True
        elif o == '--push-status-all':
            push_status_all=True
        elif o == '--last-commit':
            last_commit=True
        elif o == '--server-id':
            server_id = a
        elif o in ('-h', '--help'):
            usage()
            sys.exit(P4PASS)
        elif o == '-d':
            DEBUG = True
        else:
            assert False, 'unhandled option'

    if DEBUG:
        print ("repo={0}  commit_change={1} push_status={2}".format(
            repo, commit_change, push_status))
        print("P4PORT={0}  P4USER={1}".format(P4PORT, P4USER))

    if repo and commit_change:
        commit_sha1_list = convert_sha1_changenum(commit_change, repo)
        if commit_sha1_list:
            for cs in commit_sha1_list:
                print("    repo: {0}     sha1: {1}     change: {2}     branch_id: {3}".
                        format(cs[REPO],cs[SHA1],cs[CHANGE], cs[BRANCH]))
        else:
            print("repo: '{0}' No commit|change matches '{1}'".format(repo, commit_change))
        sys.exit(0)

    if config:
        if repo:
            print_repo_configs(repo)
        else:
            print_global_config()
        sys.exit(0)

    if repo and last_commit:
        print_last_commit(repo, server_id)
        sys.exit(0)

    if push_status:
        print_repo_status(repo, False)
        sys.exit(0)

    if push_status_all:
        print_repo_status(repo, True)
        sys.exit(0)

if __name__ == "__main__":
    if sys.hexversion < 0x02060000 or \
            (sys.hexversion > 0x03000000 and sys.hexversion < 0x03020000):
        print(_("Python 2.6+ or Python 3.2+ is required"))
        sys.exit(P4FAIL)
    main()
