#! /usr/bin/env python3.3
"""Populate a repo entirely from the object cache."""

import binascii
from collections import deque
import glob
import logging
import sys
import os
import re

import pygit2

import p4gf_git
import p4gf_object_type
import p4gf_path
import p4gf_proc
import p4gf_util
import p4gf_const

LOG = logging.getLogger(__name__)

COMMIT = 'commits'
UNCONVERTED_COMMIT = 'unconverted_commits'
UNCONVERTED_BLOB = 'unconverted_blobs'


def fast_reclone(ctx):
    """Try to do fast reclone from mirror.

    :return: branch heads and set of all added commits and trees.

    """
    # don't try this on a non-empty GF repo
    if not ctx.repo.is_empty:
        LOG.debug("fast_reclone: repo not empty")
        return None, None

    LOG.debug("fast_reclone: server_id=%s  repo=%s", ctx.server_id, ctx.config.repo_name)
    # write empty tree into repo, just in case it's referenced by any
    # commits we'll be copying
    result = p4gf_proc.popen_no_throw(['git', 'write-tree'])
    if result['ec']:
        LOG.debug('fast_reclone: failed to write empty tree: {}'.format(result))
        return None, None

    LOG.debug("fast-reclone commencing for %s", ctx.config.repo_name)
    # copy commits and trees into the repo
    branch_heads, commits = _get_objects(ctx)

    # There may be unconverted commits added to the object cache
    # by p4gf_gitmirror.py --fix
    _, unconverted_commits = _get_objects(ctx, otype=UNCONVERTED_COMMIT)
    if unconverted_commits:
        commits = commits | unconverted_commits
        _get_objects(ctx, otype=UNCONVERTED_BLOB)

    trees = _get_trees(ctx, commits)
    LOG.debug("fast-reclone copied %d commits and %d trees for %s",
              len(commits), len(trees), ctx.config.repo_name)
    return branch_heads, commits | trees


def check_fast_reclone(ctx, objects):
    """Run git fsck to ensure correctness, removing all objects if not.

    :param set objects: the SHA1 of each added commit and tree.
    :return: True if repo is ok, otherwise False.

    """
    result = p4gf_proc.popen_no_throw(['git', 'fsck'])
    LOG.debug("git-fsck returned: %s", result)
    if result['ec']:
        # We must remove only the commits and trees that were added
        # earlier, leaving all of the file blobs in place so that the
        # "slow" reclone can reuse them.
        for sha1 in objects:
            sha1_path = p4gf_util.sha1_to_git_objects_path(sha1)
            git_path = os.path.join(ctx.repo_dirs.GIT_DIR, sha1_path)
            os.unlink(git_path)
        # We have encountered problems were there is no P4 changelist
        # for a corresponding git commit. This condition prevents an
        # accurate reclone from P4 history as follows.
        # When a p4 changelist is missing, then the first commit after
        # the missing changelist will contain the sha1 for the commit
        # for the p4 changelist just before the missing change.
        # The resulting git repo has a different history
        # than that in an accurrate copy of the git repo (on another instance).
        #
        # The missing changelist problem can likely be overcome by
        # p4gf_gitmirror.py --fix <repo> which copies objects from a valid git repo on a some
        # other Git Fusion instance into the object cache 'unconverted' paths.
        # After which p4gf_fast_reclone._get_objects() will populate the
        # the new git repo with the --fix'ed objects.
        # With Luck - an accurate reclone from the object cache.
        # N.B. The P4 history would remain incorrect but harmless.
        msg = "Perforce:p4gf_fast_reclone failed on 'git fsck'\n{}".format(result)
        msg = msg + "\n    Will attempt clone from P4 changelist history."
        msg = msg + "\n    Try p4gf_gitmirror --fix <repo> and retry clone."
        LOG.error(msg)
        sys.stderr.write(msg)
        return False
    return True


COMMIT_REGEX = re.compile("/(?P<slashed_sha1>[^-]+)"
                          "-(?P<branch_id>[^,]+)"
                          ",(?P<change_num>\\d+)")

UNCONVERTED_COMMIT_REGEX = re.compile("/(?P<slashed_sha1>[^-]+)-,")
UNCONVERTED_BLOB_REGEX = re.compile("/(?P<slashed_sha1>.+)")


def _get_objects(ctx, otype=COMMIT):
    """Copy a repo's object cache objects from depot to git repo.

    :param str otype: object type to copy
                      one of 'commits', 'unconverted_commits', 'unconverted_blobs'

    Objects under 'unconverted_<commits|blobs>' paths will have been
    added by 'p4gf_gitmirror.py --fix <repo>, which also adds trees to the common trees path.
    These are commits and blobs for git commits which have no matching P4 changelists.
    fast_reclone() calls this function for commits and 'unconverted_commits', and conditionally
    for 'unconverted_blobs'.
    This self-repairs a fast_reclone after a gitmirror --fix.
    """
    # pylint: disable=too-many-branches, too-many-statements
    # sync the commit objects for the repo
    LOG.debug("_get_objects: server_id=%s repo=%s otype=%s",
              ctx.server_id, ctx.config.repo_name, otype)
    object_path = p4gf_object_type.commit_depot_path('*', '*', ctx.config.repo_name, '*')
    objects_root = os.path.join(ctx.repo_dirs.p4root,
                                'objects', 'repos', ctx.config.repo_name, 'commits')
    glob_path = os.path.join(objects_root, '*', '*', '*')
    oregex = COMMIT_REGEX

    if not otype == COMMIT:
        object_path = object_path.replace(r'/commits/', '/' + otype + '/')
        objects_root = os.path.join(
            ctx.repo_dirs.p4root, 'objects', 'repos', ctx.config.repo_name, otype)
        r = ctx.p4run('files', '-e', '-m1', object_path)
        if not r:
            LOG.debug('_get_objects: no %s for this repo.', otype)
            return None, None
        else:
            LOG.debug('_get_objects: some %s for this repo.', otype)
        LOG.debug2("_get_objects %s object_path %s %s", otype, object_path, objects_root)
        if otype == UNCONVERTED_COMMIT:
            oregex = UNCONVERTED_COMMIT_REGEX
            glob_path = os.path.join(objects_root, '*', '*', '*')
        elif otype == UNCONVERTED_BLOB:
            oregex = UNCONVERTED_BLOB_REGEX
            glob_path = os.path.join(objects_root, '*', '*', '*', '*', '*')

    LOG.debug2("_get_objects otype=%s\n    object_path=%s\n   objects_root=%s\n    glob_path=%s",
               otype, object_path, objects_root, glob_path)

    if p4gf_const.READ_ONLY:
        p4gf_util.print_depot_objects_to_object_root(
            ctx, object_path, objects_root=ctx.repo_dirs.p4root)
    else:
        ctx.p4run('sync', '-p', object_path)

    # Copy each object of 'otype' into the repo.
    # For COMMIT type, keep track of the last commit on
    # each branch, so the caller can set the branch references.
    # Return the set of commits (but [] for the BLOB case)
    branch_heads = {}
    commits = set()
    blobs = set()
    LOG.debug("_get_objects : glob path %s", glob_path)
    for path in glob.iglob(glob_path):
        # make sure the path is one of our expected object type
        LOG.debug("_get_objects :      path %s", path)
        m = oregex.search(path[len(objects_root):])
        if not m:
            LOG.debug('_get_objects: no match for %s', otype)
            continue
        # extract info encoded into the path
        sha1 = m.group('slashed_sha1').replace('/', '')
        # move the file into the git repo
        git_path = os.path.join(ctx.repo_dirs.GIT_DIR, p4gf_util.sha1_to_git_objects_path(sha1))
        if sha1 not in ctx.repo:
            p4gf_util.ensure_parent_dir(git_path)
            os.rename(path, git_path)
            try:
                # Ensure the object we just stuck in the repo is actually a
                # valid git object. If not, discard it and pretend we never
                # saw it. (see GF-3020 and GF-835)
                ctx.repo.read(sha1)
            except pygit2.GitError:
                LOG.warning("invalid git object %s, ignoring...", sha1)
                os.unlink(git_path)
                continue
        if otype == COMMIT:
            branch_id = m.group('branch_id')
            change_num = int(m.group('change_num'))
            # if this is a more recent commit on its branch, remember it
            if branch_id not in branch_heads or change_num > branch_heads[branch_id][1]:
                branch_heads[branch_id] = (sha1, change_num)
        if otype == UNCONVERTED_BLOB:
            blobs.add(sha1)
        else:
            commits.add(sha1)
        if not otype == COMMIT:
            LOG.debug2("_get_objects: unconverted object %s -> %s", path, git_path)
    # Log unconverted activities.
    if otype == UNCONVERTED_COMMIT:
        LOG.debug('_get_objects: %s  %s added to this repo.',
                  len(commits), otype)
    if otype == UNCONVERTED_BLOB:
        LOG.debug('_get_objects: %s  %s added to this repo.',
                  len(blobs), otype)
    return branch_heads, commits


def _get_trees(ctx, commits):
    """Copy the tree objects associated with the repo.

    :param ctx: Git Fusion context.
    :param commits: set of commit SHA1 values already copied.

    """
    trees = set()
    # Seed the queue with all of the trees from the known commits.
    # Need to ensure the desired commits are actually in the repo
    # (see GF-3020, GF-835).
    queue = deque(ctx.repo[sha1].tree_id.hex for sha1 in commits if sha1 in ctx.repo)
    while queue:
        sha1 = queue.popleft()
        LOG.debug2('considering tree %s', sha1)
        if sha1 not in ctx.repo:
            # Missing the tree object, print it into the object store.
            sha1_path = p4gf_util.sha1_to_git_objects_path(sha1)
            git_path = os.path.join(ctx.repo_dirs.GIT_DIR, sha1_path)
            p4gf_util.ensure_parent_dir(git_path)
            depot_path = p4gf_path.tree_p4_path(sha1)
            ctx.p4gfrun(['print', '-q', '-o', git_path, depot_path])
            LOG.debug2('printed tree %s into %s', sha1, git_path)
            if sha1 not in ctx.repo:
                LOG.error('failed to print %s', depot_path)
                # let the fsck fail later on and clean up
                break
        trees.add(sha1)
        # Add any child trees to the queue to process them later. Must
        # explicitly ignore the empty tree as the get_blob() function does
        # not handle that particular object very well.
        if sha1 != p4gf_const.EMPTY_TREE_SHA1:
            LOG.debug2('inspecting tree %s', sha1)
            tree_blob = p4gf_git.get_blob(sha1, ctx.repo)
            for subtree in _tree_scanner(tree_blob):
                queue.append(subtree)
    return trees


def _tree_scanner(blob):
    """Generator function that returns all tree SHA1 values in the given tree blob.

    :param blob: the part of the blob after the null-terminated header.
    :return: generator of tree SHA1 values.

    """
    # Mask of mode string for trees is 040000. Sometimes tree entries do
    # not have leading zero (stored as '40000') due to a git bug.
    idx = 0
    end = len(blob)
    while idx < end:
        nindex = blob.index(b'\x00', idx) + 21
        # Check entry mode, first non-zero digit is a 4
        if (blob[idx] == 48 and blob[idx + 1] == 52) or blob[idx] == 52:
            yield binascii.hexlify(blob[nindex - 20:nindex]).decode()
        idx = nindex
