#! /usr/bin/env python3.3
"""Serialize access to the server/repo specific data.

Allows either multiple readers, or a single writer, of a particular
repository. This should be used when accessing the repository itself, or
when accessing meta-data (e.g. keys) related to the repository.

"""

from contextlib import contextmanager
import fcntl
import json
import logging
import os
import signal
import time

import p4gf_const
import p4gf_ensure_dir
from p4gf_l10n import _
import p4gf_lock_common

LOG = logging.getLogger(__name__)
# How long to sleep between attempts to acquire the lock, in seconds.
_RETRY_PERIOD = 0.5
_LOCK_FILE = "lock"
_WRITE_FILE = "write"

# We can get away this shared mutable global state because we are
# essentially a single-threaded application.
_HAS_READ_LOCK = set()  # If repo name appears in set, read lock is owned.
_HAS_WRITE_LOCK = set()  # If repo name appears in set, write lock is owned.


class LockBusy(Exception):

    """Raised in the case of a non-blocking lock acquisition."""

    pass


@contextmanager
def read_lock(repo_name, group_id=None, username='none'):
    """Context manager to acquire and release shared read lock.

    :param repo_name: name of the repository being lock.
    :param group_id: pid for a writer with which we are cooperating, and hence
                     it is acceptable to additionally acquire a read lock.
    :param username: name of user acquiring the lock.

    Yields True if the reader had to wait for a writer to finish.

    """
    waited = acquire_read_lock(repo_name, group_id, username)
    wait_tracker = p4gf_lock_common.WaitTracker("read-lock", LOG)
    try:
        yield waited
    finally:
        release_read_lock(repo_name)
        wait_tracker.finished_waiting("read-lock held: {} for".format(repo_name))


@contextmanager
def write_lock(repo_name, blocking=True, group_id=None, username='none'):
    """Context manager to acquire and release exclusive write lock.

    :param repo_name: name of the repository being lock.
    :param blocking: if True, wait for readers to finish, otherwise raise LockBusy.
    :param group_id: pid for a writer with which we are cooperating, and hence
                     it is acceptable to append our own pid to the write lock.
    :param username: name of user acquiring the lock.

    """
    acquire_write_lock(repo_name, blocking, group_id, username)
    wait_tracker = p4gf_lock_common.WaitTracker("write-lock", LOG)
    try:
        yield
    finally:
        release_write_lock(repo_name)
        wait_tracker.finished_waiting("write-lock held: {} for".format(repo_name))


def _active_reader_count(repo_name, seconds_waited):
    """Determine the number of readers currently in progress.

    This should only be called when either the write lock acquisition
    process has begun (which prevents additional readers), or when the
    key-based lock is held by the current process.

    """
    entries = os.listdir(_locks_dir_name(repo_name))
    count = len(entries)
    # not exactly concise, but hopefully correct
    if _LOCK_FILE in entries:
        count -= 1
    if _WRITE_FILE in entries:
        count -= 1
    log_level = p4gf_lock_common.log_level(seconds_waited)
    if LOG.isEnabledFor(log_level):
        pids = ' '
        # build string of 'pid/user' for each read lock
        for e in entries:
            # read lock filenames are pids
            if e.isdigit():
                pids = pids + ' ' + e
                # read the lock file for the username
                read_fname = _read_lock_name(repo_name, e)
                content = _load_read_lock(read_fname)
                if content and 'user' in content:
                    pids = pids + '/' + content['user']
        if pids.isspace():
            pids = ''
        LOG.log(log_level, "active reader count %s: %s%s", count, repo_name, pids)
    return count


def detect_stale_readers(repo_name):
    """Return the list of process IDs of apparently stale read locks."""
    # No flock coordination, just scan the directory to avoid blocking.
    entries = os.listdir(_locks_dir_name(repo_name))
    if _LOCK_FILE in entries:
        entries.remove(_LOCK_FILE)
    if _WRITE_FILE in entries:
        entries.remove(_WRITE_FILE)
    return [pid for pid in entries if not p4gf_lock_common.live_fusion_process(pid)]


def detect_stale_writers(repo_name):
    """Examine the write lock owners to find stale owners."""
    # No flock coordination, just read the file to avoid blocking.
    write_fname = _write_lock_name(repo_name)
    content = _load_write_lock(write_fname)
    return [pid for pid in content.keys() if not p4gf_lock_common.live_fusion_process(pid)]


def prune_stale_readers(repo_name):
    """Find and remove any stale reader locks.

    :param repo_name: name of repository for which to find stale locks
    :return: True if one or more read locks were removed.

    """
    removed = False
    with _git_repo_lock(repo_name):
        for pid in detect_stale_readers(repo_name):
            LOG.warning("stale read-lock %s for %s removed", pid, repo_name)
            os.unlink(_read_lock_name(repo_name, pid))
            removed = True
    return removed


def prune_stale_writers(repo_name):
    """Find and remove any stale writer locks.

    :param repo_name: name of repository for which to find stale locks
    :return: True if one or more write locks were removed.

    """
    removed = False
    with _git_repo_lock(repo_name):
        write_fname = _write_lock_name(repo_name)
        content = _load_write_lock(write_fname)
        new_content = dict()
        for pid, value in content.items():
            if p4gf_lock_common.live_fusion_process(pid):
                new_content[pid] = value
            else:
                LOG.warning("stale write-lock %s for %s removed", pid, repo_name)
                removed = True
        if new_content:
            with open(write_fname, 'w') as fobj:
                json.dump(new_content, fobj)
        elif os.path.exists(write_fname):
            os.unlink(write_fname)
    return removed


def acquire_read_lock(repo_name, group_id=None, username='none'):
    """Acquire a shared read lock on the named repository.

    :param repo_name: name of repository for which to acquire lock
    :param group_id: pid for a writer with which we are cooperating, and hence
                     it is acceptable to additionally acquire a read lock.
    :param username: name of user acquiring the lock.
    :return: True if read request had waited on a write lock, False otherwise.

    """
    LOG.debug2("read-lock acquiring: %s", repo_name)
    write_fname = _write_lock_name(repo_name)
    read_fname = _read_lock_name(repo_name)
    waited = False
    label = _("acquiring read-lock for {repo_name}").format(repo_name=repo_name)
    wait_tracker = p4gf_lock_common.WaitTracker(label, LOG)
    while True:
        with _git_repo_lock(repo_name):
            # need to wait for any pending writer to finish
            if not os.path.exists(write_fname):
                _store_read_lock(read_fname, username)
                break
            # If the given group pid is contained in the write lock, then
            # add the PID of this process to the lock.
            if _check_write_lock_pid(write_fname, group_id):
                _store_read_lock(read_fname, username)
                LOG.debug("read-lock borrowed: %s by %s", repo_name, os.getpid())
                break
            # check if the writer process is still alive
            if _check_all_processes(write_fname):
                LOG.warning("stale write lock for %s removed", repo_name)
                os.unlink(write_fname)
                _store_read_lock(read_fname, username)
                break
        waited = True
        wait_tracker.been_waiting()
        time.sleep(_RETRY_PERIOD)

    wait_tracker.finished_waiting("read-lock acquired: {} after".format(repo_name))
    _HAS_READ_LOCK.add(repo_name)
    return waited


def _store_read_lock(read_fname, username):
    """Serialize the data to JSON and write to the read lock file."""
    content = json.dumps({'user': username})
    with open(read_fname, "w") as fobj:
        fobj.write(content)


def release_read_lock(repo_name):
    """Remove a shared read lock on the named repository.

    :param repo_name: name of repository for which to remove lock

    """
    with _git_repo_lock(repo_name):
        read_fname = _read_lock_name(repo_name)
        if os.path.exists(read_fname):
            os.unlink(read_fname)
    _HAS_READ_LOCK.discard(repo_name)
    LOG.debug("read-lock released: %s", repo_name)


def has_read_lock(repo_name):
    """Return True if this process owns the read lock.

    :param repo_name: name of repository for which to check the lock

    """
    return repo_name in _HAS_READ_LOCK


def acquire_write_lock(repo_name, blocking=True, group_id=None, username='none'):
    """Acquire an exclusive write lock on the named repository.

    :param repo_name: name of repository for which to acquire lock.
    :param blocking: if False, raise LockBusy if unable to get exclusive access.
    :param group_id: pid for a writer with which we are cooperating, and hence
                     it is acceptable to append our own pid to the write lock.
    :param username: name of user acquiring the lock.

    Will automatically allow a single reader if this process itself owns
    the read lock for the named repository. Otherwise waits for all other
    read locks to release before returning.

    """
    # indicate our interest in writing to the repository
    LOG.debug2("write-lock acquiring: %s, blocking=%s, group_id=%s", repo_name, blocking, group_id)
    write_fname = _write_lock_name(repo_name)
    label = _("acquiring write-lock for {repo_name}").format(repo_name=repo_name)
    wait_tracker = p4gf_lock_common.WaitTracker(label, LOG)
    self_pid = str(os.getpid())
    start_time = time.time()
    while True:
        with _git_repo_lock(repo_name):
            # If there is no other writer, then take the lock.
            if not os.path.exists(write_fname):
                _append_to_write_lock(write_fname, self_pid, username)
                break
            # If the given group pid is contained in the write lock, then
            # add the PID of this process to the lock.
            if _check_write_lock_pid(write_fname, group_id):
                _append_to_write_lock(write_fname, self_pid, username)
                LOG.debug("write-lock borrowed: %s by %s", repo_name, self_pid)
                break
            # check if the other writer process is still alive.
            if _check_all_processes(write_fname):
                LOG.warning("stale write-lock for %s removed", repo_name)
                os.unlink(write_fname)
                _append_to_write_lock(write_fname, self_pid, username)
                break
        if not blocking:
            raise LockBusy()
        wait_tracker.been_waiting()
        time.sleep(_RETRY_PERIOD)
    LOG.debug2("write-lock pending: %s", repo_name)
    # Wait for all of the currently active readers to finish, unless this
    # process itself is a reader, in which case we permit that one reader.
    allowed_readers = 1 if has_read_lock(repo_name) else 0
    seconds_elapsed = (time.time() - start_time) / 1000
    while _active_reader_count(repo_name, seconds_elapsed) > allowed_readers:
        if not blocking:
            release_write_lock(repo_name, no_log=True)
            LOG.debug("write-lock cancelled: %s", repo_name)
            raise LockBusy()
        prune_stale_readers(repo_name)
        wait_tracker.been_waiting()
        time.sleep(_RETRY_PERIOD)
    _HAS_WRITE_LOCK.add(repo_name)
    wait_tracker.finished_waiting("write-lock acquired: {} after".format(repo_name))


def _load_write_lock(write_fname):
    """Deserialize the write lock and return the data."""
    try:
        with open(write_fname, 'r') as fobj:
            content = json.load(fobj)
            if not isinstance(content, dict):
                raise ValueError()
            return content
    except FileNotFoundError:
        return dict()
    except ValueError:
        LOG.exception("encountered corrupt lock file %s", write_fname)
        return dict()


def _load_read_lock(read_fname):
    """Deserialize the read lock and return the data."""
    try:
        with open(read_fname, 'r') as fobj:
            content = json.load(fobj)
            if not isinstance(content, dict):
                raise ValueError()
            return content
    except FileNotFoundError:
        return dict()
    except ValueError:
        LOG.exception("encountered corrupt lock file %s", read_fname)
        return dict()


def _check_write_lock_pid(write_fname, pid):
    """Return True if the given pid is contained in the write lock."""
    content = _load_write_lock(write_fname)
    return str(pid) in content


def _append_to_write_lock(write_fname, pid, user):
    """Append the given information to the write lock."""
    content = _load_write_lock(write_fname)
    content[str(pid)] = {'user': user}
    with open(write_fname, 'w') as fobj:
        return fobj.write(json.dumps(content))


def release_write_lock(repo_name, no_log=False):
    """Remove the exclusive write lock on the named repository.

    Removes only the entry corresponding to the current process.
    If the file becomes empty, it will be removed entirely.

    :param repo_name: name of repository for which to remove lock
    :param bool no_log: if True, do not log a (misleading) message

    """
    write_fname = _write_lock_name(repo_name)
    with _git_repo_lock(repo_name):
        content = _load_write_lock(write_fname)
        content.pop(str(os.getpid()), None)
        if content:
            with open(write_fname, 'w') as fobj:
                json.dump(content, fobj)
        elif os.path.exists(write_fname):
            os.unlink(write_fname)
    _HAS_WRITE_LOCK.discard(repo_name)
    if not no_log:
        LOG.debug("write-lock released: %s", repo_name)


def has_write_lock(repo_name):
    """Return True if this process owns the write lock.

    :param repo_name: name of repository for which to check the lock

    """
    return repo_name in _HAS_WRITE_LOCK


def wholly_owned_write_lock(repo_name):
    """Indicate if the current process has the one and only write lock.

    :param repo_name: name of repository for which to check the lock

    """
    return len(_HAS_WRITE_LOCK) == 1 and repo_name in _HAS_WRITE_LOCK


def _read_lock_name(repo_name, pid=None):
    """Return the name of a read lock for the named repo and this process."""
    if pid is None:
        pid = os.getpid()
    return "{0}/locks/{1}/{2}".format(p4gf_const.P4GF_HOME, repo_name, pid)


def _write_lock_name(repo_name):
    """Return the name of the write lock for the named repo."""
    return "{0}/locks/{1}/{2}".format(p4gf_const.P4GF_HOME, repo_name, _WRITE_FILE)


def _locks_dir_name(repo_name):
    """Return the name of the directory which contains the lock files."""
    return "{0}/locks/{1}".format(p4gf_const.P4GF_HOME, repo_name)


@contextmanager
def _git_repo_lock(repo_name):
    """Acquire exclusive access to the locks for the named repository."""
    def handler(_signum, _frame):
        """Raise a LockBusy exception."""
        raise LockBusy("timeout on flock call for {}".format(repo_name))

    path = "{0}/locks/{1}/{2}".format(p4gf_const.P4GF_HOME, repo_name, _LOCK_FILE)
    parent_dir = os.path.dirname(path)
    if not os.path.exists(parent_dir):
        p4gf_ensure_dir.ensure_dir(parent_dir)
    LOG.debug2("git-lock acquiring: %s", repo_name)
    fd = os.open(path, os.O_CREAT | os.O_WRONLY)
    # Avoid blocking indefinitely on the flock call by using an alarm
    # signal to wake us up after a minute.
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(60)
    fcntl.flock(fd, fcntl.LOCK_EX)
    signal.alarm(0)
    os.write(fd, bytes("{}".format(os.getpid()), "utf-8"))
    wait_tracker = p4gf_lock_common.WaitTracker("git-lock", LOG)
    try:
        LOG.debug2("git-lock acquired: %s", repo_name)
        yield
    finally:
        os.close(fd)
    wait_tracker.finished_waiting("git-lock held: {} for".format(repo_name))


def _check_all_processes(fname):
    """Read the PIDs from the named file and verify if any are alive.

    :param str fname: path of file containing process IDs.
    :return: True if all processes are no longer alive, False otherwise.

    """
    content = _load_write_lock(fname)
    for pid in content.keys():
        if p4gf_lock_common.live_fusion_process(pid):
            LOG.debug("found fresh lock process %s", pid)
            return False
        elif LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("found stale lock process %s", pid)
    return True
