#! /usr/bin/env python3.3
"""Manages the object cache of commits and trees."""

from collections import defaultdict
from contextlib import ExitStack
import functools
import logging
import os
from collections import namedtuple
import re
import sys

import p4gf_env_config    # pylint: disable=unused-import
import p4gf_branch
import p4gf_config
import p4gf_const
import p4gf_context
import p4gf_create_p4
from p4gf_depot_branch import DepotBranchInfo
from p4gf_desc_info import DescInfo
from p4gf_fastimport_mark import Mark
from p4gf_fastimport_marklist import MarkList
import p4gf_git
import p4gf_git_repo_lock
from p4gf_l10n import _, NTR
from p4gf_lfs_file_spec import lfs_depot_path
import p4gf_lock
import p4gf_log
from p4gf_object_type import ObjectType
import p4gf_object_type_util
from p4gf_p2g_rev_range import RevRange
import p4gf_p4cache
from p4gf_p4changelist import P4Changelist
import p4gf_p4key as P4Key
from p4gf_p4file import P4File
import p4gf_p4msg
import p4gf_p4msgid
import p4gf_path
import p4gf_proc
from p4gf_profiler import Timer
import p4gf_progress_reporter as ProgressReporter
import p4gf_pygit2
import p4gf_translate
import p4gf_util

from P4 import OutputHandler, P4Exception

try:
    from enum import Enum
except ImportError:
    class Enum:
        """Gee golly I wish we had Python 3.4 Enum."""

        def __init__(self):
            """Initialize an instance of Enum."""
            pass

Sha1ChangeBranch = namedtuple('Sha1ChangeBranch', ['sha1', 'change_num', 'branch_id'])
ChangeBranch = namedtuple('ChangeBranch', ['change_num', 'branch_id'])

# cannot use __name__ since it will often be "__main__"
LOG = logging.getLogger("p4gf_gitmirror")
MAX_COMMITS = 10000  # should the user configure 0, ignore it and use this.
COMMIT_CHANGE_RE = re.compile(r'^ ' + p4gf_const.P4GF_DESC_KEY_CHANGE + r': (?P<change_num>\d+)$',
                              re.MULTILINE)


def _get_max_commits_per_submit():
    """Read the int(value) from the git_to_perforce section of the config.

    :return: Default if value is not set.
    :raises RuntimeError: if config value is malformed.

    """
    try:
        max_commits = p4gf_config.GlobalConfig.get(p4gf_config.SECTION_GIT_TO_PERFORCE,
                                                   p4gf_config.KEY_MIRROR_MAX_COMMITS_PER_SUBMIT)
        max_commits = int(max_commits)
        if not max_commits or max_commits == 0:
            max_commits = MAX_COMMITS
        return max_commits
    except ValueError:
        raise RuntimeError(_('global p4gf_config: {section}/{key}  is not a valid natural number.'
                             ' Please contact your administrator.')
                           .format(section=p4gf_config.SECTION_GIT_TO_PERFORCE,
                                   key=p4gf_config.KEY_MIRROR_MAX_COMMITS_PER_SUBMIT))

#
# Monkey-patch Context class, lazily adding a mirror property. Do this here
# rather than in p4gf_context.py to avoid an import cycle.
#


def __get_mirror(self):
    """Return the context's mirror, creating one if necessary."""
    # pylint: disable=protected-access
    if not hasattr(self, '_mirror'):
        self._mirror = GitMirror(self.config.repo_name)
    return self._mirror


def __set_mirror(self, value):
    """Replace the context's mirror value."""
    self._mirror = value    # pylint: disable=protected-access
p4gf_context.Context.mirror = property(__get_mirror, __set_mirror)


def __repr__with_mirror(self):
    return str(self) + "\n" + repr(self.mirror)
p4gf_context.__repr__ = __repr__with_mirror


_BITE_SIZE = 1000       # How many files to pass in a single 'p4 xxx' operation.
# If this is changed, also change mirror_boundary.t test so it submits this
# exact number of commits.

DESCRIPTION = _("Git Fusion '{repo}' copied to Git.")


class FilterAddFstatHandler(OutputHandler):
    """OutputHandler for p4 fstat, builds list of files that don't already exist."""

    def __init__(self):
        """Init the handler."""
        OutputHandler.__init__(self)
        self.files = []

    def outputMessage(self, m):
        """outputMessage call expected for any files not already added.

        Otherwise logs an error.

        """
        try:
            if m.msgid == p4gf_p4msgid.MsgDm_ExFILE:
                self.files.append(m.dict['argc'])
        except Exception:  # pylint: disable=broad-except
            LOG.exception("outputMessage")
        return OutputHandler.REPORT


class MirrorStatus(Enum):
    """Mirror status."""

    NONE = 0
    STARTED = 1
    COMPLETED = 2
    PENDING = 3

    @staticmethod
    def from_str(s):
        """Convert string to integer."""
        return int(s)

    @staticmethod
    def from_p4(ctx):
        """Construct a MirrorStatus for the given context."""
        name = p4gf_const.P4GF_P4KEY_MIRROR_STATUS.format(
            repo_name=ctx.config.repo_name, server_id=ctx.server_id)
        value = P4Key.get(ctx, name)
        return MirrorStatus.from_str(value)

    @staticmethod
    def update(ctx, value):
        """Save the mirror status value to Perforce."""
        name = p4gf_const.P4GF_P4KEY_MIRROR_STATUS.format(
            repo_name=ctx.config.repo_name, server_id=ctx.server_id)
        P4Key.set(ctx, name, str(value))


class MirrorObjects:
    """A collection of commit and tree ObjectType instances."""

    def __init__(self):
        """Init object."""
        self.pending_commits = {}   # commit+branches we need to add+submit
        self.pending_trees = set()  # trees we need to add+submit
        self.last_tree = None       # commit tree for last commit added
        self.pending_blobs = set()  # blobs we need to add+submit
        # switch to set when called with the --fix option
        self.add_unconverted = False

    def add_commit(self, sha1, change_num, ctx, branch_id):
        """Queue commit for adding along with trees it references."""
        if sha1 not in ctx.repo:
            LOG.warning('commit %s for change %s on branch %s not found in repo, cannot mirror',
                        sha1, change_num, branch_id)
            return

        key = sha1 + ',' + branch_id if branch_id else sha1
        assert key not in self.pending_commits
        commit = ObjectType.create_commit(sha1=sha1,
                                          repo_name=ctx.config.repo_name,
                                          change_num=change_num,
                                          branch_id=branch_id)
        self.pending_commits[key] = commit
        self._add_commits(ctx)
        with Timer(EXTRACT_OBJECTS):
            # as commits are added, build set of referenced trees
            if not self.last_tree:
                self.last_tree = _get_snapshot_trees(sha1, self.pending_trees, ctx.repo)
            else:
                self.last_tree = _get_delta_trees(self.last_tree, sha1, self.pending_trees,
                                                  ctx.repo)

    def flush(self, ctx):
        """Add any pending commits and tree objects to the numbered change."""
        # If there are no new commits, there will be no new trees.
        # But there could be commits but no trees if all trees were previously
        # referenced by other commits.
        self._add_commits(ctx, force=True)
        self._add_trees(ctx)
        self._add_blobs(ctx)

    def clear(self):
        """Clear any pending commits and trees.

        Remember any added commits and trees.

        """
        self.pending_commits.clear()
        self.pending_trees.clear()
        self.pending_blobs.clear()
        self.last_tree = None

    def _add_commits(self, ctx, force=False):
        """Add any pending commit objects.

        P4 add any pending commits, if there are _BITE_SIZE commits pending.
        If force is True, p4 add any pending commits, no matter how few.

        """
        if not self.pending_commits:
            return
        if not force and len(self.pending_commits) < _BITE_SIZE:
            return
        LOG.debug("_add_commit: pending_commits = %s", len(self.pending_commits))
        _try_twice(functools.partial(_add_commits_to_p4, ctx, self.pending_commits))
        self.pending_commits.clear()

    def _add_trees(self, ctx):
        """Add any pending tree objects."""
        if not self.pending_trees:
            return
        _try_twice(functools.partial(_add_trees_to_p4, ctx, self.pending_trees))
        self.pending_trees.clear()
        self.last_tree = None

    def _add_blobs(self, ctx):
        """Add any pending blob objects."""
        if not self.pending_blobs:
            return
        _try_twice(functools.partial(_add_blobs_to_p4, ctx, self.pending_blobs,
                   self.add_unconverted))
        self.pending_blobs.clear()


# timer/counter names
OVERALL = NTR('GitMirror Overall')
EXTRACT_OBJECTS = NTR('extract objects')
P4_FSTAT = NTR('p4 fstat')
P4_ADD = NTR('p4 add')
P4_SUBMIT = NTR('p4 submit')


class GitMirror:
    """Handle git things that get mirrored in Perforce."""

    def __init__(self, repo_name):
        """Object init."""
        self.objects = MirrorObjects()
        self.repo_name = repo_name

        # List of branches that we write to config2:
        # - lightweight branches, whether or not they have a
        #   git-branch-name. Includes any deleted lightweight
        #   branches.
        # - branches based on a stream: we record the stream's
        #   original view to config2, even though the branch's
        #   main definition remains in config not config2.
        self._branch_list_config_2 = None

        self._depot_branch_info_list = None

        # List of depot files already written to local disk by
        # ChangelistDataFile.write().
        # Most are new (require 'p4 add'), some will be updates
        # to existing (require 'p4 sync -k' + 'p4 edit')
        self._changelist_data_file_list = None

        # Marks list data from fast-import or fast-export
        self._marks = None
        self._mark_list = None
        self._mark_to_branch_id = None

        # If True, invoke ObjectType.update_indexes()
        self._update_indexes = False

    def _add_objects_to_p4(self, ctx, missing_commits=None):
        """Submit Git commit and tree objects associated with the provided marks."""
        log = LOG.getChild("add_objects_to_p4")
        branch_dict = ctx.branch_dict()
        max_commits_per_submit = _get_max_commits_per_submit()
        log.debug2("max_commits_to_submit %s", max_commits_per_submit)
        sha1_change_branch = []
        if missing_commits:
            log.debug2("adding missing commits either deleted or previously not mirrored: %s",
                       missing_commits)
            sha1_change_branch.extend(missing_commits)
        try:
            log.debug3("mark-to-branch: %s", self._mark_to_branch_id)
            with ProgressReporter.Indeterminate():
                desc = DESCRIPTION.format(repo=ctx.config.repo_name)
                with Timer(OVERALL):
                    for mark_line in self._marks:
                        mark = Mark.from_line(mark_line)
                        log.debug3('%s from "%s"', mark, mark_line)
                        mark_num = mark.mark
                        if self._mark_list:
                            change_num = self._mark_list.mark_to_cl(mark_num)
                        else:
                            change_num = mark_num
                        sha1 = mark.sha1
                        branch_id = mark.branch
                        if (not branch_id) and self._mark_to_branch_id:
                            branch_id = self._mark_to_branch_id.get(mark_num)
                            log.debug3('new branch_id %s', branch_id)
                        if (branch_id in branch_dict and branch_dict[branch_id].start_at_change and
                                int(change_num) < branch_dict[branch_id].start_at_change):
                            log.debug2("SKIP %s/%s < start_at_change=%s", branch_id, change_num,
                                       branch_dict[branch_id].start_at_change)
                            continue
                        sha1_change_branch.append(Sha1ChangeBranch(
                            sha1=sha1, change_num=change_num, branch_id=branch_id))

                        if len(sha1_change_branch) >= max_commits_per_submit:
                            self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)
                            sha1_change_branch = []

                    # Add the last commits
                    if len(sha1_change_branch):
                        self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)

        finally:
            # Let my references go!
            self.objects.clear()
            self._marks = None
            self._mark_list = None
            self._mark_to_branch_id = None

    def add_missing_to_p4(self, ctx, converted_commits):
        """Submit commit and tree objects associated with the given commits.

        The provided mapping of commit SHA1 values and their corresponding
        changes is fairly complete, and the job of this function is to
        submit only those objects that are missing from the object cache.
        After all, it performs that check anywayy, no sense doing it twice.

        :param ctx: context for the repo of interest.
        :param converted_commits: map of SHA1 to a list of ChangeBranch tuples.

        """
        assert ctx.repo_lock.has_lock, 'mirroring requires repo lock'
        log = LOG.getChild("add_missing_to_p4")
        branch_dict = ctx.branch_dict()
        max_commits_per_submit = _get_max_commits_per_submit()
        log.debug2("max_commits_to_submit %s", max_commits_per_submit)
        sha1_change_branch = []
        try:
            desc = DESCRIPTION.format(repo=ctx.config.repo_name)
            for sha1, changes in converted_commits.items():
                for cb in changes:
                    branch_id = cb.branch_id
                    change_num = cb.change_num
                    if (branch_id in branch_dict and branch_dict[branch_id].start_at_change and
                            int(change_num) < branch_dict[branch_id].start_at_change):
                        log.debug2("SKIP %s/%s < start_at_change=%s", branch_id, change_num,
                                   branch_dict[branch_id].start_at_change)
                        continue
                    sha1_change_branch.append(Sha1ChangeBranch(
                        sha1=sha1, change_num=change_num, branch_id=branch_id))
                    if len(sha1_change_branch) >= max_commits_per_submit:
                        self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)
                        sha1_change_branch = []
            # add the left-over commits
            if len(sha1_change_branch):
                self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)
        finally:
            self.objects.clear()
            sha1_change_branch = []

    def add_unconverted_to_p4(self, ctx, unconverted_commits):
        """Submit commit and tree objects associated with the given unconverted commits.

        :param ctx: context for the repo of interest.
        :param unconverted_commits: SHA1

        """
        assert ctx.repo_lock.has_lock, 'mirroring requires repo lock'
        log = LOG.getChild("add_unconverted_to_p4")

        self.objects.add_unconverted = True
        max_commits_per_submit = _get_max_commits_per_submit()
        log.debug2("max_commits_to_submit %s", max_commits_per_submit)
        sha1_change_branch = []
        try:
            desc = DESCRIPTION.format(repo=ctx.config.repo_name)
            for sha1 in unconverted_commits:
                self._add_blobs_from_unconverted_commit(sha1)
                sha1_change_branch.append(Sha1ChangeBranch(
                    sha1=sha1, change_num='', branch_id=''))
                if len(sha1_change_branch) >= max_commits_per_submit:
                    self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)
                    sha1_change_branch = []
            # add the left-over commits
            if len(sha1_change_branch):
                self._add_commits_in_numbered_changelist(sha1_change_branch, ctx, desc)
        finally:
            self.objects.clear()
            sha1_change_branch = []
            self.objects.add_unconverted = False

    def _add_commits_in_numbered_changelist(self, sha1_change_branch, ctx, desc):
        """Using a new NumberChangelist, submit the list of commits and trees."""
        with p4gf_util.NumberedChangelist(gfctx=ctx, description=desc) as nc:
            for scb in sha1_change_branch:
                index_key_name = p4gf_const.P4GF_P4KEY_INDEX_OT.format(
                                     repo_name=ctx.config.repo_name,
                                     change_num=scb.change_num, branch_id=scb.branch_id)
                # Add commit object.
                # But, do not add if there already exists an index_branch key for this
                # repo/branch/changenum. Otherwise we would rewrite history when performing
                # reclone. The duplicated commit object with different sha1, same branch/changenum
                # would be used to set the branch HEAD.
                #
                index_key_value = P4Key.get(ctx.p4gf, index_key_name)
                # If the key value is not set, we have a new commit.
                # If the key value == scb.sha1, we may be replacing a missing commit object.
                # If the key value != scb sha1, we assume that this scb.sha1 is incorrect
                # and skip.
                if index_key_value == '0' or index_key_value == scb.sha1:
                    self.objects.add_commit(sha1=scb.sha1, change_num=scb.change_num, ctx=ctx,
                                            branch_id=scb.branch_id)
                else:
                    LOG.debug("skip mirror of commit %s - already exists in cache", scb)
            # add the remaining objects to P4
            self.objects.flush(ctx)
            opened = ctx.p4gfrun('opened', '-m1')
            if opened:
                ProgressReporter.increment(_('Submitting new Git commit objects to Perforce'))
                with Timer(P4_SUBMIT):
                    r = nc.submit_with_retry()
                if r and self._update_indexes:
                    ObjectType.update_indexes(ctx, r)
            else:
                ProgressReporter.write(_('No new Git objects to submit to Perforce'))
                LOG.debug("ignoring empty change list...")

    def _add_blobs_from_unconverted_commit(self, sha1):
        """Derive the new blobs in this commit and add to the pending_blobs."""
        # use git diff-tree to find all blobs in uncommitted commit
        cmd = ['git', 'diff-tree', '-r', '-m', '--no-commit-id', sha1]
        po = p4gf_proc.popen_no_throw(cmd)['out']
        for line in po.splitlines():
            m = ALL_ENT_RE.match(line)
            if m:
                if m.group(1) in ['100644', '100664', '100755', '120000']:
                    self.add_pending_blob(m.group(2))
                    LOG.debug2("for unconverted sha=%s adding blob=%s", sha1, m.group(1))

    def add_pending_blob(self, sha1):
        """Add given SHA1 to pending blob list."""
        self.objects.pending_blobs.add(sha1)

    def _update_branches(self, ctx):
        """Add/update files for new/changed branches."""
        # Update the config files for any new or changed branches. We must
        # have exclusive access when updating said files.
        assert ctx.repo_lock.has_lock, 'mirroring requires repo lock'
        with ExitStack() as stack:
            stack.enter_context(Timer(OVERALL))
            # Add branch_info files for new depot branches. If we created
            # any new depot branches to house lightweight branches, record
            # a branch_info file for each new depot branch.
            result = []
            branches = ctx.depot_branch_info_index().by_id.values()
            LOG.debug3("_add_depot_branch_infos() branches to consider %s", branches)
            for dbi in branches:
                if dbi.needs_p4add or dbi.needs_p4edit:
                    result.append(dbi)
            LOG.debug3("_add_depot_branch_infos() branches to add/edit %s", result)
            self._depot_branch_info_list = result
            # Record added branches in p4gf_config2 file. If we defined any
            # new lightweight branches, record those mappings in the
            # p4gf_config2 file. If we have any stream-based branches,
            # record their initial views in p4gf_config2 as well.
            self._branch_list_config_2 = [b for b in ctx.branch_dict().values()
                                          if b.is_worthy_of_config2()]
            ctx.write_config1()
            self._add_config2_branch_defs_to_p4(ctx)
            desc = DESCRIPTION.format(repo=ctx.config.repo_name)
            nc = stack.enter_context(p4gf_util.NumberedChangelist(gfctx=ctx, description=desc))
            depot_branch_infos_added = self._add_depot_branch_infos_to_p4(ctx)
            cldfs_added = self._add_cldfs_to_p4(ctx)
            if depot_branch_infos_added or cldfs_added:
                ProgressReporter.increment(_('Submitting branch updates to Perforce'))
                stack.enter_context(Timer(P4_SUBMIT))
                nc.submit()

    def _add_depot_branch_infos_to_p4(self, ctx):
        """Add branch-info files to Perforce for new depot branches.

        If we created any new depot branches, 'p4 add' their branch-info
        file to Perforce. Does not submit.

        If we edited any existing depot branches, 'p4 edit' them.

        Return number of branch-info files added or edited.

        """
        if not self._depot_branch_info_list:
            LOG.debug2('_add_depot_branch_infos_to_p4() nothing to add')
            return 0

        add_path_list = []
        edit_path_list = []
        for dbi in self._depot_branch_info_list:
            local_path = _write_depot_branch_info_local_file(ctx, dbi)
            if dbi.needs_p4add:
                add_path_list.append(local_path)
            else:
                edit_path_list.append(local_path)
            # Mark as written to P4
            dbi.needs_p4add = False
            dbi.needs_p4edit = False

        with Timer(P4_ADD):
            success_list = _p4_add_in_bites(ctx, add_path_list)
        success_list.extend(_p4_sync_k_edit(ctx, edit_path_list))
        LOG.debug3("_add_depot_branch_infos_to_p4() adding %s", success_list)
        self._depot_branch_info_list = None
        return len(success_list)

    def _add_config2_branch_defs_to_p4(self, ctx):
        """Update p4gf_config2 file with new branches.

        If we defined any new named+lightweight branches, update (or write
        the first revision of) this repo's p4gf_config2 file with all the
        currently defined named+lightweight branches.

        """
        # Nothing to write? Maybe we have just deleted the remaining refs.
        if self._branch_list_config_2:
            for b in p4gf_branch.ordered(self._branch_list_config_2):
                LOG.debug("add branch to config2 %s", b)
                b.add_to_config(ctx.repo_config.repo_config2)

        return ctx.repo_config.write_repo2_if(ctx.p4gf)

    def _add_cldfs_to_p4(self, ctx):
        """Open ChangelistDataFiles for add or edit.

        If we have any ChangelistDataFile local paths in
        changelist_data_file_list, 'p4 add' or 'p4 sync -k' + 'p4 edit' them
        now.

        Return True if we added/edited at least one file, False if not.

        """
        if not self._changelist_data_file_list:
            return False

        # Rather than run 'p4 opened' and then 'p4 sync' + 'p4
        # edit -k' on any files we failed to 'p4 add', we can
        # save a lot of thinking by just blindly 'p4 add'ing and
        # 'p4 edit'ing all files. Yeah it's stupid, yeah it will
        # pollute logs with a lot of warnings.
        with Timer(P4_ADD), ctx.p4gf.at_exception_level(ctx.p4gf.RAISE_NONE):
            ctx.p4gfrun('sync', '-k', self._changelist_data_file_list)
            ctx.p4gfrun('edit', '-k', self._changelist_data_file_list)
            ctx.p4gfrun('add',        self._changelist_data_file_list)

        self._changelist_data_file_list = None
        return True

    def set_changelist_data_file_list(self, cldf_list):
        """Remember a list of local file paths to add later."""
        self._changelist_data_file_list = cldf_list

    def set_marks_list_data(self, marks, mark_list=None, mark_to_branch_id=None):
        """Set the marks list data for commit/tree caching."""
        self._marks = marks
        self._mark_list = mark_list
        self._mark_to_branch_id = mark_to_branch_id

    def do_phase_1(self, ctx, missing_commits=None, skip_log=False, update_indexes=False):
        """Perform phase 1 of the mirroring work.

        :param ctx: Context instance.
        :param missing_commits: list of Sha1ChangeBranch instances (optional).
        :param bool skip_log: True to skip writing intent log (default False);
                              this is only useful when restoring from the log.
        :param bool update_indexes: True to update commit/change mapping keys
                                    (basically for pull only).

        """
        assert ctx.repo_lock.has_lock, 'mirroring requires repo lock'
        LOG.debug('start of phase 1 mirroring for %s', self.repo_name)
        if not skip_log:
            _write_intent_log(self.repo_name, self, 1, missing_commits=missing_commits)
        MirrorStatus.update(ctx, MirrorStatus.STARTED)
        LOG.info('copying objects and metadata data for %s', self.repo_name)
        self._update_indexes = update_indexes
        self._add_objects_to_p4(ctx, missing_commits)
        _delete_intent_log(self.repo_name)
        LOG.info('phase 1 mirroring complete for %s', self.repo_name)

    def do_phase_2(self, ctx, lfs_files=None, skip_log=False):
        """Perform phase 2 of the mirroring work.

        :param ctx: Context instance.
        :param lfs_files: map of sha256 to blob depot path (optional).
        :param bool skip_log: True to skip writing intent log (default False);
                              this is only useful when restoring from the log.

        """
        assert ctx.repo_lock.has_lock, 'mirroring requires repo lock'
        filename = _intent_log_filename(ctx.config.repo_name)
        if os.path.exists(filename):
            # let the mirror recovery pick up after this
            LOG.warning('phase 1 mirroring failed, will skip phase 2')
            return
        LOG.debug('start of phase 2 mirroring for %s', self.repo_name)
        if not skip_log:
            _write_intent_log(self.repo_name, self, 2, lfs_files=lfs_files)
        self._update_branches(ctx)
        if lfs_files:
            desc = DESCRIPTION.format(repo=ctx.config.repo_name)
            with p4gf_util.NumberedChangelist(gfctx=ctx, description=desc) as nc:
                for sha256, depot_path in lfs_files.items():
                    ctx.p4gfrun('integ', depot_path, lfs_depot_path(ctx, sha256))
                with Timer(P4_SUBMIT):
                    nc.submit_with_retry()
        MirrorStatus.update(ctx, MirrorStatus.COMPLETED)
        _delete_intent_log(self.repo_name)
        LOG.info('phase 2 mirroring complete : mirroring done for %s', self.repo_name)

    def __str__(self):
        """Return string representation of this Git mirror."""
        return 'GitMirror for {0}'.format(self.repo_name)

    @staticmethod
    def from_dict(data, repo):
        """Translate the JSON compatible dict into a GitMirror instance.

        :param dict data: data deserialized from JSON.
        :param str repo: name of repository.

        """
        mirror = GitMirror(repo)
        mirror.from_dict_instance(data)
        return mirror

    def from_dict_instance(self, data):
        """Read the GitMirror properties from the JSON compatible dict.

        :param dict data: data deserialized from JSON.

        """
        # must convert the JSON list() back to a set()
        self.objects.pending_blobs = set(data['pending_blobs'])
        self._branch_list_config_2 = data['branch_list']
        branch_infos = data['branch_infos']
        if branch_infos:
            branch_infos = [DepotBranchInfo.from_dict(dbi) for dbi in branch_infos]
            self._depot_branch_info_list = branch_infos
        self._changelist_data_file_list = data['changelist_data_file_list']
        self._marks = data['marks']
        mark_list = data.get('mark_list')
        if mark_list:
            self._mark_list = MarkList.from_dict(data['mark_list'])
        mark_to_branch_id = data.get('mark_to_branch_id')
        if mark_to_branch_id:
            # force the keys to be integers, as expected in _add_objects_to_p4()
            # (would be nice to know how str keys got in there...)
            self._mark_to_branch_id = {int(k): v for k, v in mark_to_branch_id.items()}

    def to_dict(self):
        """Serialize this GitMirror to a JSON compatible dict.

        :return: dict of GitMirror properties.

        """
        result = dict()
        # must convert the set() to a list() for JSON
        result['pending_blobs'] = list(self.objects.pending_blobs)
        result['branch_list'] = self._branch_list_config_2
        if self._depot_branch_info_list:
            result['branch_infos'] = [dbi.to_dict() for dbi in self._depot_branch_info_list]
        else:
            result['branch_infos'] = None
        result['changelist_data_file_list'] = self._changelist_data_file_list
        result['marks'] = self._marks
        if self._mark_list:
            result['mark_list'] = self._mark_list.to_dict()
        result['mark_to_branch_id'] = self._mark_to_branch_id
        return result


# -- module-level ------------------------------------------------------------


def pending_mirror(ctx, reset=False):
    """Indicate that a mirror-worthy operation is about to begin.

    :param bool reset: if True, sets the mirror status to none (0).

    Does nothing in READ_ONLY mode.

    """
    if not p4gf_const.READ_ONLY:
        LOG.debug('set mirror status pending for %s', ctx.config.repo_name)
        value = MirrorStatus.NONE if reset else MirrorStatus.PENDING
        MirrorStatus.update(ctx, value)


def recover_failed_mirror(ctx):
    """If necessary, recover from a previously failed mirror operation.

    :param ctx: context for the repo of interest.

    """
    server_ids = P4Key.get_all(ctx, p4gf_const.P4GF_P4KEY_SERVER_ID + '*')
    known_servers = [k.replace(p4gf_const.P4GF_P4KEY_SERVER_ID, '') for k in server_ids.keys()]
    LOG.debug2('recover_failed_mirror known servers: %s', known_servers)

    keys_pattern = p4gf_const.P4GF_P4KEY_MIRROR_STATUS.format(
        repo_name=ctx.config.repo_name, server_id="*")
    key_values = P4Key.get_all(ctx, keys_pattern)
    LOG.debug2('recover_failed_mirror status keys: %s', key_values)
    status_by_server = dict()
    for key_name, key_value in key_values.items():
        server_id = key_name[len(keys_pattern)-1:]
        if server_id not in known_servers:
            # Must be a repository name with a shared prefix and a hyphen,
            # e.g. "foobar" and "foobar-main".
            LOG.debug('recover_failed_mirror ignoring unknown server %s', server_id)
            continue
        status = MirrorStatus.from_str(key_value)
        status_by_server[server_id] = status
    LOG.debug2('recover_failed_mirror status parsed: %s', status_by_server)

    # Any status that indicates "started" is one that failed to finish
    # since by the time we get here, it should have completed (or it never
    # started in the first place).
    mss = MirrorStatus.STARTED
    msp = MirrorStatus.PENDING
    failed_set = {k for k, v in status_by_server.items() if v == mss or v == msp}
    LOG.debug2('recover_failed_mirror failed set: %s', failed_set)
    if ctx.server_id in failed_set:
        filename = _intent_log_filename(ctx.config.repo_name)
        if os.path.exists(filename):
            assert ctx.repo_lock.has_lock, 'mirror operations require repo lock'
            _restore_from_intent_log(ctx, filename)
        else:
            LOG.error('mirror status not completed, yet missing intent log for %s',
                      ctx.config.repo_name)
            LOG.error('try using `p4gf_gitmirror.py --fix %s` to recover', ctx.config.repo_name)
            raise RuntimeError(_('mirror failure not recoverable; contact admin and request'
                                 ' that they run `p4gf_gitmirror.py --fix {}`').format(
                                 ctx.config.repo_name))
    elif failed_set:
        server_id = failed_set.pop()
        LOG.error('mirror failure on instance %s; use pull to recover', server_id)
        # Some other instance had a mirror failure, we cannot fix it
        # here, report the problem and return immeidately.
        frm = _('This Git Fusion instance is currently blocked by a Git Fusion mirroring '
                'failure on instance {}. Please pull from that server to engage auto-recovery.')
        raise RuntimeError(frm.format(server_id))


def _write_intent_log(repo, mirror, phase, missing_commits=None, lfs_files=None):
    """Write the provided data to the mirroring intent log.

    :param mirror: instance of GitMirror to serialize.
    :param int phase: the mirroring phase, either 1 or 2
    :param missing_commits: list of Sha1ChangeBranch instances (optional).
    :param lfs_files: map of sha256 to blob depot path (optional).

    """
    filename = _intent_log_filename(repo)
    data = mirror.to_dict()
    data['phase'] = phase
    data['missing_commits'] = missing_commits
    data['lfs_files'] = lfs_files
    p4gf_util.write_dict_to_file(data, filename)


def _restore_from_intent_log(ctx, intent_log):
    """Recover any pending work from an intent log and submit to Perforce."""
    LOG.warn('previous mirror failed, will attempt to recover from log %s', intent_log)
    data = p4gf_util.read_dict_from_file(intent_log)
    LOG.debug3('mirror intent log contents: %s', data)

    def make_scb(mc):
        """Convert a plain list to a Sha1ChangeBranch tuple."""
        return Sha1ChangeBranch(sha1=mc[0], change_num=mc[1], branch_id=mc[2])

    phase = data.pop('phase')
    missing_commits = data.pop('missing_commits', None)
    if missing_commits:
        missing_commits = [make_scb(mc) for mc in missing_commits]
    lfs_files = data.pop('lfs_files', None)
    mirror = GitMirror.from_dict(data, ctx.config.repo_name)
    LOG.debug('successfully loaded intent log for %s', ctx.config.repo_name)
    # If the mirror had failed in phase 1, start with that. In either case,
    # we know we will be doing phase 2, since a failure in phase 1 implies
    # that phase 2 also failed.
    #
    # In both cases, skip writing a new intent log, since it is an intent
    # log from which we are recovering.
    if phase < 2:
        mirror.do_phase_1(ctx, missing_commits, skip_log=True)
    mirror.do_phase_2(ctx, lfs_files, skip_log=True)


def _delete_intent_log(repo):
    """Remove the mirroring intent log, if it exists."""
    filename = _intent_log_filename(repo)
    if os.path.exists(filename):
        os.unlink(filename)


def _intent_log_filename(repo):
    """Generate the name of the intent log file for the given repo.

    :param str repo: name of repository.
    :return: path and name for intent log file.

    """
    file_name = NTR('mirror-data-{repo}.json').format(repo=repo)
    return os.path.join(p4gf_const.P4GF_HOME, file_name)


def _write_depot_branch_info_local_file(ctx, dbi):
    """Write a DepotBranchInfo to a file in the perforce client workspace.

    Create a new local file to house the given DepotBranchInfo's spec,
    fill it, and return the local path to it.

    Return the path of the client workspace file suitable for  use with
    p4 add.

    """
    with Timer(EXTRACT_OBJECTS):
        config = dbi.to_config()
        depot_path = dbi.to_config_depot_path()
        local_path = p4gf_util.depot_to_local_path(depot_path, ctx.p4gf, ctx.client_spec_gf)
        p4gf_util.ensure_dir(p4gf_util.parent_dir(local_path))
        p4gf_util.make_writable(local_path)
        with open(local_path, 'w') as f:
            config.write(f)
        p4gf_config.clean_up_parser(config)
    return local_path


def _p4_sync_k_edit(ctx, path_list):
    """'p4 sync -k' then 'p4 edit' the paths.

    Return list of depotFile successfully opened.

    """
    if not path_list:
        return []
    ctx.p4gfrun('sync', '-k', path_list)
    ctx.p4gfrun('edit', '-k', path_list)
    l = p4gf_p4msg.find_msgid(ctx.p4gf, p4gf_p4msgid.MsgDm_OpenSuccess)
    success_list = [x['depotFile'] for x in l]
    return success_list


def _p4_add_in_bites(ctx, path_list):
    """'p4 add' all the files in path_list.

    Return list of depotFile successfully opened.

    """
    success_list = []
    for bite in _spoon_feed(path_list):
        r = ctx.p4gfrun('add', bite)
        success_list.extend(
            [x['depotFile'] for x in r
             if isinstance(x, dict) and 'action' in x and x['action'] == 'add'])

    LOG.debug("_p4_add_in_bites() want=%s success=%s", len(path_list), len(success_list))
    return success_list


def _add_commits_to_p4(ctx, commits):
    """Run p4 add to create mirror files in .git-fusion."""
    log = LOG.getChild('add_commits_to_p4')
    log.debug("considering %s commits to be added", len(commits))

    # build list of objects to add, extracting them from git
    LOG.debug3('_add_commits_to_p4() commits: %s', commits)
    add_files = [_extract_commit_to_p4(ctx, go) for go in commits.values()]
    add_files = _optimize_objects_to_add_to_p4(ctx, add_files)

    if not len(add_files):
        # Avoid a blank line in output by printing something
        ProgressReporter.write(_('No Git objects to submit to Perforce'))
        log.debug("found nothing new to add")
    else:
        _add_objects_to_p4(ctx, add_files)
        ObjectType.reset_cache()


def _extract_commit_to_p4(ctx, go):
    """Extract a commit to the git-fusion perforce client workspace.

    Return the path of the client workspace file suitable for use with
    p4 add.

    """
    ProgressReporter.increment(_('Adding new Git commit objects to Perforce...'))

    # get client path for .git-fusion file
    dst = os.path.join(ctx.gitlocalroot, go.to_p4_client_path())

    if os.path.exists(dst):
        LOG.debug2("reusing existing commit: %s", dst)
    else:
        with Timer(EXTRACT_OBJECTS):
            # Extract the object and write to file
            p4gf_git.write_git_object_from_sha1(ctx.repo, go.sha1, dst)
            LOG.debug2("adding new object: %s", dst)
    return dst


# line is: mode SP type SP sha TAB path
# we only want the sha from lines with type "tree"
TREE_REGEX = re.compile(r'^[0-7]{6} tree ([0-9a-fA-F]{40})\t.*')
ALL_ENT_RE = re.compile(r'^:[0-7]{6} ([0-7]{6}) [0-9a-fA-F]{40} ([0-9a-fA-F]{40}) .*')


def _get_snapshot_trees(commit, trees, view_repo):
    """Get all tree objects for a given commit.

    commit: SHA1 of commit
    trees: set of all trees seen so far
    view_repo: pygit2 repo

    Each tree found is added to the set to be mirrored.

    """
    # get top level tree and make sure we haven't already processed it
    commit_tree = p4gf_pygit2.object_to_sha1(view_repo.get(commit).tree)
    if commit_tree in trees:
        return commit_tree

    # ls-tree doesn't return the top level tree, so add it here
    trees.add(commit_tree)
    LOG.debug2("adding (total=%s) commit tree %s", len(trees), commit_tree)

    # use git ls-tree to find all subtrees in this commit tree
    po = p4gf_proc.popen_no_throw(['git', 'ls-tree', '-rt', commit_tree])['out']
    for line in po.splitlines():
        m = TREE_REGEX.match(line)
        if m:
            trees.add(m.group(1))
            LOG.debug2("adding (total=%s) subtree %s", len(trees), m.group(1))
    return commit_tree

# line is: :mode1 SP mode2 SP sha1 SP sha2 SP action TAB path
# we want sha2 from lines where mode2 indicates a dir
TREE_ENT_RE = re.compile(r'^:[0-7]{6} 04[0-7]{4} [0-9a-fA-F]{40} ([0-9a-fA-F]{40}) .*')


def _get_delta_trees(commit_tree1, commit2, trees, view_repo):
    """Get all tree objects new in one commit vs another commit.

    commit_tree1: SHA1 of first commit's tree
    commit2: SHA1 of second commit
    trees: set of all trees seen so far
    view_repo: pygit2 repo

    Each tree found is added to the set to be mirrored.

    """
    # get top level tree and make sure we haven't already processed it
    commit_tree2 = p4gf_pygit2.object_to_sha1(view_repo.get(commit2).tree)
    if commit_tree2 in trees:
        return commit_tree2

    # diff-tree doesn't return the top level tree, so add it here
    trees.add(commit_tree2)
    LOG.debug2("adding (total=%s) commit tree %s", len(trees), commit_tree2)

    # use git diff-tree to find all subtrees not in previous commit tree
    po = p4gf_proc.popen_no_throw(['git', 'diff-tree', '-t', commit_tree1, commit_tree2])['out']
    for line in po.splitlines():
        m = TREE_ENT_RE.match(line)
        if m:
            trees.add(m.group(1))
            LOG.debug2("adding (total=%s) subtree %s", len(trees), m.group(1))
    return commit_tree2


def _add_trees_to_p4(ctx, trees):
    """Run p4 add to create mirror files in .git-fusion."""
    LOG.debug("adding %s trees to .git-fusion..., change=%s", len(trees), ctx.numbered_change_gf)
    LOG.debug2("trees to add: %s", trees)

    # build list of trees to add, extracting them from git
    add_files = []
    for sha1 in list(trees):
        relative_path = 'trees/' + p4gf_path.slashify_sha1(sha1)
        add_files.append(_extract_object_to_p4(ctx, sha1, relative_path))
    add_files = _optimize_objects_to_add_to_p4(ctx, add_files)
    if not len(add_files):
        return
    _add_objects_to_p4(ctx, add_files)


def _add_blobs_to_p4(ctx, blobs, add_unconverted):
    """Run p4 add to create mirror files in .git-fusion."""
    LOG.debug("adding %s blobs to .git-fusion..., change=%s", len(blobs), ctx.numbered_change_gf)
    LOG.debug2("blobs to add: %s", blobs)
    LOG.debug("add_unconverted = %s", add_unconverted)

    # build list of blobs to add, extracting them from git
    add_files = []
    for sha1 in list(blobs):
        if add_unconverted:
            relative_path = 'repos/{}/unconverted_blobs/{}'.format(
                    ctx.config.repo_name, p4gf_path.slashify_blob_sha1(sha1))
            LOG.debug("unconverted_blog path %s", relative_path)
        else:
            relative_path = 'blobs/' + p4gf_path.slashify_blob_sha1(sha1)
        add_files.append(_extract_object_to_p4(ctx, sha1, relative_path))
    add_files = _optimize_objects_to_add_to_p4(ctx, add_files)
    if not len(add_files):
        return
    _add_objects_to_p4(ctx, add_files)


def _extract_object_to_p4(ctx, sha1, local_path):
    """Extract an object into the git-fusion perforce client workspace.

    Return the path of the client workspace file suitable for use with p4 add.

    :param ctx: context for the repo of interest.
    :param str sha1: SHA1 of the object to extract.
    :param str local_path: trailing path for local file.

    """
    # get client path for .git-fusion file
    dst = os.path.join(ctx.gitlocalroot, 'objects', local_path)
    # An object is likely to already exist, in which case we don't need or
    # want to try to recreate it. We'll just use the existing one.
    if os.path.exists(dst):
        LOG.debug("reusing existing git tree object: %s", dst)
        return dst
    with Timer(EXTRACT_OBJECTS):
        # Extract object from repo and write to dst
        p4gf_git.write_git_object_from_sha1(ctx.repo, sha1, dst)
    return dst


def _optimize_objects_to_add_to_p4(ctx, add_files):
    """If many files to add, filter out those which are already added.

    Only do this if the number of files is large enough to justify the cost
    of the fstat.

    """
    original_count = len(add_files)
    enough_files_to_use_fstat = 100
    if original_count < enough_files_to_use_fstat:
        return add_files
    with ExitStack() as stack:
        stack.enter_context(Timer(P4_FSTAT))
        LOG.debug("using fstat to optimize add")
        handler = FilterAddFstatHandler()
        stack.enter_context(ctx.p4gf.using_handler(handler))
        stack.enter_context(ctx.p4gf.at_exception_level(ctx.p4gf.RAISE_NONE))
        for bite in _spoon_feed(add_files):
            ctx.p4gf.run("fstat", bite)
        add_files = handler.files
        LOG.debug("%s files removed from add list", original_count - len(add_files))
        return add_files


def _add_objects_to_p4(ctx, add_files):
    """'p4 add' Git tree and commit objects to Perforce. Does not submit."""
    with Timer(P4_ADD):
        treecount = 0
        commitcount = 0
        for bite in _spoon_feed(add_files):
            result = ctx.p4gfrun('add', '-t', 'binary+F', bite)
            for r in result:
                if isinstance(r, dict) and r["action"] != 'add' or \
                   isinstance(r, str) and r.find("currently opened for add") < 0:
                    # file already exists in depot, perhaps?
                    LOG.debug(r)
                elif isinstance(r, dict):
                    if p4gf_object_type_util.OBJPATH_TREE_REGEX.search(r["depotFile"]):
                        treecount += 1
                    else:
                        commitcount += 1
    LOG.debug("Added %s commits and %s trees", commitcount, treecount)


def _spoon_feed(items, bite_size=_BITE_SIZE):
    """Break a list of items in bite size chunks.

    When running p4 commands (e.g. add, fstat) on many files, it works
    better to break it up into several commands with smaller lists of files.

    """
    while items:
        bite = items[:bite_size]
        items = items[bite_size:]
        yield bite


def _try_twice(func):
    """Call func and retry if there's a P4Exception raised.

    At most one retry will be made.

    """
    try:
        return func()
    except P4Exception:
        return func()


def _discover_converted_changes(ctx, known_commits):
    """Build a set of changes that were converted to commits.

    This relies on the presense of the export header, without which we
    cannot know which commits originated from a changelist.

    :return: dict of change numbers (int) to commit SHA1 (str)

    """
    converted_changes = dict()
    for sha1 in known_commits:
        commit = ctx.repo[sha1]
        if p4gf_const.P4GF_EXPORT_HEADER in commit.message:
            m = COMMIT_CHANGE_RE.search(commit.message)
            if m:
                # P4Changelist uses an int for the change number, so do the
                # same here for easier comparison later.
                change_num = int(m.group('change_num'))
                converted_changes[change_num] = sha1
    return converted_changes


def _discover_converted_commits(ctx, converted_changes):
    """Build a set of all commits that have corresponding changes."""
    # A mapping of commit SHA1 to a list of ChangeBranch tuples that
    # correspond to that commit, based on changelist descriptions.
    changes = defaultdict(list)

    def append(branch_id, change):
        """Add the commit SHA1 to the set of converted commits."""
        di = DescInfo.from_text(change.description)
        # Not all changelists originated from git, after all. We must also
        # ignore ghost changes, otherwise we get false negatives (i.e.
        # missing commits that do not really exist).
        #
        # Alternatively, if we already know of a change that was converted
        # to a commit, we can now determine the branch for that change.
        if di and not di.is_ghost():
            cb = ChangeBranch(change_num=change.change, branch_id=branch_id)
            changes[di.sha1].append(cb)
        elif change.change in converted_changes:
            cb = ChangeBranch(change_num=change.change, branch_id=branch_id)
            sha1 = converted_changes[change.change]
            changes[sha1].append(cb)

    range_as_string = RevRange.from_start_stop(ctx, "@1", "#head").as_range_string()
    all_branches = [b[1] for b in sorted(ctx.branch_dict().items())]
    for b in all_branches:
        appender = functools.partial(append, b.branch_id)
        with ctx.switched_to_branch(b):
            path_range = ctx.client_view_path() + range_as_string
            P4Changelist.get_changelists(ctx.p4, path_range, appender)
    return changes


def _discover_missing_commits(ctx, known_commits):
    """Discover those commit objects which are missing from the cache.

    :param ctx: context for the repo of interest.
    :param known_commits: list of known commit SHA1s.

    :return: tuple: (dict: {commit_sha1 => [ChangeBranch]),
                     set: {commit_sha1} not converted to changelists).

    """
    # Determine all of the changelists that were converted to commits.
    converted_changes = _discover_converted_changes(ctx, known_commits)
    # Start with all of the commits that we know were converted to p4.
    converted_commits = _discover_converted_commits(ctx, converted_changes)

    # Determine those commits currently in git that were apparently not
    # translated to p4. This will include any commits that originated in
    # p4, and hence may not really be missing.
    unconverted_commits = set(known_commits) - set(converted_commits.keys())

    def files_callback(p4file):
        """Process the file."""
        nonlocal converted_commits
        nonlocal unconverted_commits
        m = p4gf_object_type_util.OBJPATH_COMMIT_REGEX.search(p4file.depot_path)
        if m:
            sha1 = m.group(NTR('slashed_sha1')).replace('/', '')
            if sha1 in converted_commits:
                changes = converted_commits[sha1]
                # P4Changelist.change is an int, so convert this one, too
                change_num = int(m.group(NTR('change_num')))
                branch_id = m.group(NTR('branch_id'))
                cb = ChangeBranch(change_num=change_num, branch_id=branch_id)
                if cb in changes:
                    changes.remove(cb)
            unconverted_commits.discard(sha1)

    # Now discover what git-originating commits are missing from the cache.
    # Whatever is left in the converted_commits set is indeed missing.
    commits_path = NTR('{0}/repos/{1}/commits/...').format(
        p4gf_const.objects_root(), ctx.config.repo_name)
    P4File.get_files(ctx.p4gf, commits_path, files_callback)

    # Attempt to filter out the incorrectly marked p4-originating commits
    # by checking the commit message for our export header. Since we are
    # modifying the set while iterating it, wrap it in a list first.
    for candidate in list(unconverted_commits):
        commit = ctx.repo[candidate]
        if p4gf_const.P4GF_EXPORT_HEADER in commit.message:
            unconverted_commits.discard(candidate)

    return (converted_commits, unconverted_commits)


def _discover_missing_trees(ctx, known_commits):
    """Discover those tree objects which are missing from the cache.

    :param ctx: context for the repo of interest.
    :param known_commits: list of known commit SHA1s.

    :return: set of missing trees.

    """
    # Start by assuming that they are all missing. Whatever is left in this
    # set really is missing from our cache.
    missing_trees = p4gf_git.all_git_trees(ctx.repo, known_commits)
    # Compute the ordered set of tree SHA1 two-character prefixes. This is
    # used to fetch the trees from the object cache in manageable batches,
    # without fetching each one individually (too many requests), nor
    # fetching all of the tree objects in existence (expensive request).
    for prefix in set(sha1[:2] for sha1 in missing_trees):
        tree_path = NTR('{0}/trees/{1}/...').format(p4gf_const.objects_root(), prefix)
        for rr in ctx.p4gfrun('files', '-e', tree_path):
            if 'depotFile' in rr:
                match = p4gf_object_type_util.OBJPATH_TREE_REGEX.search(rr['depotFile'])
                if match:
                    sha1 = match.group(NTR('slashed_sha1')).replace('/', '')
                    missing_trees.discard(sha1)
    return missing_trees


def _prune_deleted_commits(ctx, missing_commits):
    """Remove those commits that are marked as deleted in the depot."""
    really_missing_commits = defaultdict(list)
    for sha1, changes in missing_commits.items():
        sha1_path = p4gf_path.slashify_sha1(sha1)
        for cb in changes:
            depot_path = NTR('{0}/repos/{1}/commits/{2}-{3},{4}').format(
                p4gf_const.objects_root(), ctx.config.repo_name, sha1_path,
                cb.branch_id, cb.change_num)
            r = ctx.p4gf.run('fstat', '-T', 'depotFile', '-F', 'headAction=delete', depot_path)
            if len(r) == 0:
                really_missing_commits[sha1].append(cb)
    return really_missing_commits


def check_missing_objects(ctx, verbose=False, ignore_deleted=False):
    """Check for any missing cache objects for the context's repo.

    :param ctx: context for the repo of interest.
    :param bool verbose: True to print individual missing objects.
    :param bool ignore_deleted: True to ignore deleted objects

    :return: number of missing commits and trees.
    :rtype: 2-tuple of ints

    """
    known_commits = p4gf_git.all_git_commits(ctx.repo)
    missing_commits, unconverted_commits = _discover_missing_commits(ctx, known_commits)
    if ignore_deleted:
        missing_commits = _prune_deleted_commits(ctx, missing_commits)
    missing_commit_count = 0
    for sha1, changes in missing_commits.items():
        if verbose:
            for cb in changes:
                print(_("missing commit {sha1}-{branch},{change}").format(
                    sha1=sha1, branch=cb.branch_id, change=cb.change_num))
        missing_commit_count += len(changes)
    if verbose:
        for sha1 in unconverted_commits:
            print(_("unconverted commit {sha1}").format(sha1=sha1))
    missing_commit_count += len(unconverted_commits)
    missing_trees = _discover_missing_trees(ctx, known_commits)
    if verbose:
        for tree in missing_trees:
            print(_("missing tree {tree}").format(tree=tree))
    return (missing_commit_count, len(missing_trees))


def _add_missing_objects(ctx):
    """Ensure all commit and tree objects are stored in the object cache."""
    known_commits = p4gf_git.all_git_commits(ctx.repo)
    missing_commits, unconverted_commits = _discover_missing_commits(ctx, known_commits)
    for sha1 in unconverted_commits:
        # We cannot associate this commit with a changelist so report it
        print("WARNING: commit not converted to changelist {}. "
              "Adding to 'unconverted' cache.".format(sha1))
    if unconverted_commits:
        # Add the git commit and its blobs to the object cache 'unconverted' paths.
        # Add its trees to the usual objects/trees/...
        ctx.mirror.add_unconverted_to_p4(ctx, unconverted_commits)

    ctx.mirror.add_missing_to_p4(ctx, missing_commits)

    missing_trees = _discover_missing_trees(ctx, known_commits)
    desc = DESCRIPTION.format(repo=ctx.config.repo_name)
    with p4gf_util.NumberedChangelist(gfctx=ctx, description=desc) as nc:
        _add_trees_to_p4(ctx, missing_trees)
        opened = ctx.p4gfrun('opened', '-m1')
        if opened:
            nc.submit_with_retry()

    # Clear the mirror status for this repo now that we're done.
    MirrorStatus.update(ctx, MirrorStatus.COMPLETED)


def _process_repo(repo, fix, verbose, ignore_deleted):
    """Process the named repository.

    :param str repo: git name of the repository
    :param bool fix: True to attempt to fix missing objects
    :param bool verbose: True to print details
    :param bool ignore_deleted: True to ignore deleted objects

    """
    repo_name = p4gf_translate.TranslateReponame.git_to_repo(repo)
    print(_("Processing repository {repo_name}...").format(repo_name=repo_name))
    try:
        ctx = p4gf_context.create_context(repo_name)
        with ExitStack() as stack:
            stack.enter_context(ctx)
            if not os.path.exists(ctx.repo_dirs.GIT_WORK_TREE):
                print(_('not local to this server'))
                return
            os.chdir(ctx.repo_dirs.GIT_WORK_TREE)
            if not ctx.add_copied_from_perforce:
                print('WARNING: enable-add-copied-from-perforce is set to false,'
                      ' results may be inaccurate')
            ctx.repo_lock = p4gf_lock.RepoLock(ctx.p4gf, repo_name, blocking=False)
            stack.enter_context(ctx.repo_lock)
            # To find and fix the missing objects, we will need a client
            # pool and write access to this instance.
            stack.enter_context(p4gf_git_repo_lock.write_lock(repo_name))
            if fix:
                _add_missing_objects(ctx)
            else:
                (commits, trees) = check_missing_objects(ctx, verbose, ignore_deleted)
                print(_("missing {commits} commits and {trees} trees")
                      .format(commits=commits, trees=trees))
    except p4gf_config.ConfigLoadError:
        print(_("error loading config file"))
    except p4gf_lock.LockBusy:
        print(_("unable to acquire lock"))
    except RuntimeError as err:
        LOG.exception(err)
        # truncate long error messages
        print(str(err).splitlines()[0])


def main():
    """Check for missing cache objects, with the option to add them."""
    p4gf_util.has_server_id_or_exit()
    desc = _("Check for missing cache objects.")
    epilog = _("Note that at this time, missing commits cannot be fixed.")
    parser = p4gf_util.create_arg_parser(desc=desc, epilog=epilog)
    parser.add_argument(NTR("repo"), metavar=NTR("R"), nargs="*",
                        help=_("name of the repositories to be checked"))
    parser.add_argument("-a", "--all", action="store_true",
                        help=_("process all known Git Fusion repositories"))
    parser.add_argument("-F", "--fix", action="store_true",
                        help=_("find missing trees and submit to Perforce"))
    parser.add_argument("-v", "--verbose", action="store_true",
                        help=_("print details of missing objects"))
    parser.add_argument("--ignore-deleted", action="store_true",
                        help=_("ignore deleted commit objects"))
    args = parser.parse_args()

    # Check that either --all, or "repo" was specified.
    if not args.all and len(args.repo) == 0:
        sys.stderr.write(_("Missing view names; try adding --all option.\n"))
        sys.exit(2)
    if args.all and len(args.repo) > 0:
        sys.stderr.write(_("Ambiguous arguments. Choose --all or a view name.\n"))
        sys.exit(2)

    p4gf_proc.init()
    with p4gf_create_p4.Closer():
        p4 = p4gf_create_p4.create_p4_temp_client()
        if not p4:
            sys.exit(2)
        # Sanity check the connection (e.g. user logged in?) before proceeding.
        try:
            p4gf_p4cache.fetch_client(p4)
        except P4Exception as e:
            sys.stderr.write(_("P4 exception occurred: {exception}\n").format(exception=e))
            sys.exit(1)
        if args.all:
            repos = p4gf_util.repo_config_list(p4)
            if len(repos) == 0:
                print(_("No Git Fusion repositories found, nothing to do."))
                sys.exit(0)
        else:
            repos = args.repo
        p4gf_branch.init_case_handling(p4)
        p4gf_create_p4.p4_disconnect(p4)

        for repo in repos:
            _process_repo(repo, args.fix, args.verbose, args.ignore_deleted)


if __name__ == "__main__":
    p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
