#! /usr/bin/env python3.3
"""Collection of lock related functions."""

import errno
import logging
import os
import re
import time

import p4gf_log
import p4gf_proc

LOG = logging.getLogger(__name__)
RE_FUSION_PROC = re.compile(r'p4gf_.*\.py')

# How long (in seconds) before logging as a warning that we have been
# waiting. Prior to this time, logging is at debug level.
_NOWARN_PERIOD = 240.


def pid_exists(pid):
    """Check if a process by the given pid exists."""
    try:
        os.kill(int(pid), 0)
    except ValueError:
        LOG.error("encountered non-integer PID value: %s", pid)
        return False
    except OSError as err:
        if err.errno == errno.ESRCH:
            return False
        elif err.errno == errno.EPERM:
            return True
        raise
    return True


def live_fusion_process(pid):
    """Determine if the given process is a live Git Fusion process.

    First checks if the process is alive at all, and if so, then uses the
    ps command to determine if that process appears to be a Git Fusion
    process (i.e. contains p4gf_*.py). If the ps command exits with a non-
    zero exit code, simply check if the process is alive or not.

    :param pid: process identifier
    :return: True if process is alive, False otherwise

    """
    if not pid_exists(pid):
        return False
    # Get the command and arguments, since our script name is an argument
    # to the Python interpreter, and the command is typically 'python'.
    # The arguments to ps happen to work for both 'linux' and 'darwin'.
    cmd = ['ps', '-o', 'command', '-p', str(pid)]
    LOG.info('ps command: %s', cmd)
    result = p4gf_proc.popen_no_throw(cmd)
    LOG.info('ps output for %s: %s, %s; %s', pid, result['ec'], result['out'], result['err'])
    if result['ec'] != 0:
        # maybe the process _just_ died this very moment
        return pid_exists(pid)
    # Does the process appear to be a Git Fusion process?
    if RE_FUSION_PROC.search(result['out']):
        return True
    return False


def log_level(seconds, default=logging.DEBUG):
    """Return logging level based on comparing seconds against _NOWARN_PERIOD."""
    return logging.WARNING if seconds >= _NOWARN_PERIOD else default


class WaitTracker(p4gf_log.LongWaitReporter):

    """Keeps track of the duration of a waiting process, logging occassionally."""

    def __init__(self, label, logger, wait_time=60):
        """Construct a new instance with the given label and logger.

        :param label:     included in the log message.
        :param logger:    the logging instance with which to report.
        :param wait_time: how often (in seconds) to report wait.

        """
        p4gf_log.LongWaitReporter.__init__(self, label, logger, wait_time)
        self.start_time = time.time()

    def finished_waiting(self, msg):
        """Log a message at the appropriate log level."""
        seconds = (time.time() - self.start_time) / 1000
        self.logger.log(log_level(seconds), "%s %.3f s", msg, seconds)
