#! /usr/bin/env python3.3
"""Cache of results fetched from Perforce."""

import logging

import p4gf_const
from p4gf_l10n import NTR

LOG = logging.getLogger(__name__)

# Run 'p4 info' un-tagged to get human-friendly server info labels.
InfoResults = None
# But un-tagged doesn't return 'unicode' so run tagged to get that.
InfoResultsTagged = None


class ClientCacheDict(dict):
    """Subclass of dict for caching client specs."""

    def __init__(self):
        """Initialize a ClientCacheDict instance."""
        super(ClientCacheDict, self).__init__()

    def __getitem__(self, key):
        """Retrieve (a copy of) the cached client."""
        value = super(ClientCacheDict, self).__getitem__(key)
        # Return a shallow copy to avoid the caller tainting the cached
        # client spec and thus diverging from what is stored in Perforce.
        return dict(value)

    def __setitem__(self, key, value):
        """Store the key and value in the mapping."""
        # Store a shallow copy to avoid the caller tainting the cached
        # client spec and thus diverging from what is stored in Perforce.
        replicant = dict(value)
        super(ClientCacheDict, self).__setitem__(key, replicant)

# Cache of client specifications to avoid retrieving clients more than is
# really necessary. Especially helpful if a customer is running a form-in
# trigger on client specs.
ClientCache = ClientCacheDict()


def fetch_client(p4, client_name=None):
    """Retrieve the named client.

    :param p4: P4API instance.
    :param str client_name: name of client, defaults to p4.client

    """
    name = client_name if client_name else p4.client
    if name in ClientCache:
        spec = ClientCache[name]
        if 'Stream' in spec and 'View' not in spec:
            # Update the stale cached client which had a stream but no view
            # mapping assigned by retrieving the authoritative version.
            spec = p4.fetch_client(name, allow_client=True)
            ClientCache[name] = spec
            LOG.debug3('cache stale %s', spec)
        else:
            LOG.debug3('cache hit %s', spec)
    else:
        spec = p4.fetch_client(name, allow_client=True)
        ClientCache[name] = spec
        LOG.debug3('cache miss %s', spec)
    return spec


def save_client(p4, client_name, client_spec):
    """Save the named client.

    :param p4: P4API instance.
    :param str client_name: name of client to save.
    :param dict client_spec: specification of client.

    """
    saved_client = p4.client
    p4.client = client_name
    p4.save_client(client_spec, allow_client=True)
    LOG.debug3('saved %s', client_spec)
    p4.client = saved_client
    ClientCache[client_name] = client_spec


def set_client_stream(p4, client_name, stream_name):
    """Change an existing client to use a stream rather than a view.

    :param p4: P4API instance.
    :param str client_name: name of client to modify.
    :param str stream_name: name of stream to set as "Stream" of client.

    """
    p4.run('client', '-f', '-s', '-S', stream_name, client_name, allow_client=True)
    LOG.debug('set stream %s for client %s', stream_name, client_name)
    # Force the client spec to be fetched again to get the view.
    ClientCache.pop(client_name, None)


def delete_client(p4, client_name):
    """Delete the named client without raising errors.

    :param p4: P4API instance.
    :param str client_name: name of client to remove.

    """
    with p4.at_exception_level(p4.RAISE_NONE):
        p4.delete_client('-f', client_name, allow_client=True)
    LOG.debug('delete_client removed %s', client_name)
    ClientCache.pop(client_name, None)


def create_temp_client(p4, client_name, view_lines, options=None, root=None):
    """Create a temporary client for the given p4 connection.

    Creates the temporary client using the -x option. This client cannot
    shelve changes. See @465851 for details. Only one temporary client per
    connection will be auto-deleted.

    :param p4: P4API object.
    :param str client_name: name of client to create.
    :param list view_lines: list of strings for view mapping.
    :param str options: client options, defaults to p4gf_const.CLIENT_OPTIONS.
    :param str root: client root, defaults to p4gf_const.P4GF_HOME

    """
    client = p4.fetch_client(client_name, allow_client=True)
    client['Owner'] = p4gf_const.P4GF_USER
    client['LineEnd'] = NTR('unix')
    client['View'] = view_lines
    client['Options'] = options if options else p4gf_const.CLIENT_OPTIONS
    client['Root'] = root if root else p4gf_const.P4GF_HOME
    # The -x option is a deep undoc feature that signals to p4d that this
    # is a temporary client, which will be automatically deleted upon
    # disconnect. Requires passing the client specification using -i flag.
    p4.client = client_name
    p4.save_client(client, '-x', allow_client=True)
    LOG.debug3('temp %s', client)
    ClientCache[client_name] = client


def fetch_info(p4, tagged=True):
    """Fetch the 'info' from the Perforce server.

    :param p4: P4API instance.
    :param bool tagged: if True, return dict, else return str list.

    For tagged output result is a dict.
    For un-tagged output result is a list of str.
    """
    global InfoResults, InfoResultsTagged
    if tagged:
        if InfoResultsTagged is None:
            with p4.while_tagged(True):
                InfoResultsTagged = _first_dict(p4.run('info', '-s'))
        return InfoResultsTagged
    if InfoResults is None:
        with p4.while_tagged(False):
            InfoResults = p4.run('info', '-s')
    return InfoResults


# Copied from p4gf_util to avoid a cyclic import.
def _first_dict(result_list):
    """Return the first dict result in a p4 result list."""
    for e in result_list:
        if isinstance(e, dict):
            return e
    return None
