#! /usr/bin/env python3.3
"""P4Changelist class."""

import logging
import re

from P4 import OutputHandler

from p4gf_desc_info import DescInfo
from p4gf_l10n import NTR
from p4gf_p4file import P4File
import p4gf_const
import p4gf_util

LOG = logging.getLogger(__name__)
# Note that this regex only matches certain Git Fusion client names, and
# may very well not match older patterns (e.g. git-fusion--temp-N-...).
# That seems to be okay, and attempting to widen the match (e.g. by using
# P4GF_CLIENT_PREFIX) causes the rerepo to produce incorrect dates.
P4GF_CLIENT_RE = re.compile(
    r'^git-fusion-.*-[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')


class ChangesHandler(OutputHandler):
    """OutputHandler for p4 changes, passes changelists to callback function."""

    def __init__(self, callback):
        """Construct instance of ChangesHandler."""
        OutputHandler.__init__(self)
        self.callback = callback
        self.count = 0

    def outputStat(self, h):
        """Grab clientFile from fstat output."""
        try:
            change = P4Changelist.create_using_changes(h)
            self.callback(change)
            self.count = self.count + 1
        except Exception:  # pylint: disable=broad-except
            LOG.exception("outputStat")
        return OutputHandler.HANDLED


class P4Changelist:
    """A changelist, as reported by p4 describe.

    Run p4 describe of a changelist and filter the files reported
    against a specified root path, e.g. //depot/main/p4/
    """

    def __init__(self):
        """Construct instance of P4Changelist."""
        self.change = None
        self.description = None
        self.user = None
        self.time = None
        self.path = None
        self.files = []   # P4Files in this changelist
        self.identity = None

    @staticmethod
    def create_using_describe(p4, change, depot_root):
        """Create a P4Changelist by running p4 describe."""
        result = p4.run("describe", "-s", str(change))
        cl = P4Changelist()
        vardict = p4gf_util.first_dict_with_key(result, 'change')
        cl.change = int(vardict["change"])
        cl.description = vardict["desc"]
        cl.user = vardict["user"]
        cl.time = vardict["time"]
        cl.identity = vardict.get('changeIdentity', None)
        P4Changelist.clean_duplicate_imported_from_git(cl, vardict)

        if 'path' in vardict:
            cl.path = vardict['path']
        for i in range(len(vardict["depotFile"])):
            p4file = P4File.create_from_describe(vardict, i)
            # filter out files not under our root right now
            if not p4file.depot_path.startswith(depot_root):
                continue
            cl.files.append(p4file)

        return cl

    @staticmethod
    def create_using_changes(vardict):
        """Create a P4Changelist from the output of p4 changes."""
        cl = P4Changelist()
        cl.change = int(vardict["change"])
        cl.description = vardict["desc"]
        cl.user = vardict["user"]
        cl.time = vardict["time"]
        cl.identity = vardict.get('changeIdentity', None)
        if 'path' in vardict:
            cl.path = vardict['path']
        P4Changelist.clean_duplicate_imported_from_git(cl, vardict)
        return cl

    @staticmethod
    def create_using_change(p4, change):
        """p4 change -o nnn."""
        result = p4.run('change', '-o', str(change))
        cl = P4Changelist()
        vardict = p4gf_util.first_dict_with_key(result, 'Change')
        cl.change = int(vardict["Change"])
        cl.description = vardict["Description"]
        cl.user = vardict["User"]
        cl.time = vardict["Date"]   # string "2012/01/31 23:59:59"
        cl.identity = vardict.get('Identity', None)
        P4Changelist.clean_duplicate_imported_from_git(cl, vardict)
        return cl

    @staticmethod
    def get_changelists(p4, path, callback):
        """Run p4 changes to get a list of changes.

        Call callback with each found changelist.

        p4: initialized P4 object
        path: path + revision specifier, e.g. //depot/main/p4/...@1,#head
        callback: function taking P4Changelist
        """
        cmd = NTR(["changes", "-l"])
        # replace: p4 changes //depot/...@N,#head
        # with:    p4 changes -e N //depot/...
        m = re.match(r'(.+)@(\d+),#head', path)
        if m:
            cmd.append("-e")
            cmd.append(m.group(2))
            cmd.append(m.group(1))
        else:
            cmd.append(path)
        LOG.debug(cmd)
        handler = ChangesHandler(callback)
        with p4.using_handler(handler):
            p4.run(cmd)
        return handler.count

    @staticmethod
    def get_changelists_from_viewlist(p4, viewlist, client_name_path, callback):
        """Run p4 changes to get a list of changes.

        Call callback with each found changelist.

        p4: initialized P4 object
        path: path + revision specifier, e.g. //depot/main/p4/...@1,#head
        callback: function taking P4Changelist
        """
        LOG.debug('get_changelist_from_viewlist: client_name_path %s viewlist %s',
                  client_name_path, viewlist)
        cmd = NTR(["changes", "-l"])
        m = re.match(r'(.+)@(\d+),#head', client_name_path)
        if m:
            cmd.append("-e")
            cmd.append(m.group(2))
        cmd.extend(viewlist)
        LOG.debug(cmd)
        handler = ChangesHandler(callback)
        with p4.using_handler(handler):
            p4.run(cmd)
        return handler.count

    @staticmethod
    def create_changelist_list_as_dict(p4, path, _limit=None):
        """Run p4 changes to get a list of changes.

        Return the result as a dict indexed by changelist number (as string).

        Returns a dict["change"] ==> P4Changelist

        p4: initialized P4 object
        path: path + revision specifier, e.g. //depot/main/p4/...@1,#head
        """
        changes = {}

        def append(change):
            """Append a change to the list."""
            changes[change.change] = change

        P4Changelist.get_changelists(p4, path, append)
        if LOG.isEnabledFor(logging.DEBUG):
            cll = []
            for c in changes.keys():
                if len(cll) == 10:
                    cll.append(NTR('...{} total').format(len(changes)))
                    break
                cll.append(str(c))
            LOG.debug("create_changelist_list_as_dict returning\n%s", ' '.join(cll))
        return changes

    @staticmethod
    def clean_duplicate_imported_from_git(cl, vardict):
        """Remove 'Imported from Git' block if not created by Git Fusion.

        It may occur that users copy/integrate changes pushed from git into
        Perforce. The 'Imported from Git' data block in those new
        changelist descriptions will produce incorrect history - since the
        parent-changes list is not pertinent to the history of this new
        change describing the target files. Remove the the 'Import from
        Git' block if the change's client was not a Git Fusion client.

        """
        # If the change's client was a git-fusion temp client do not strip
        # P4GF_IMPORT_HEADER from the changelist description.
        if 'client' not in vardict:
            return
        client = vardict['client']
        if P4GF_CLIENT_RE.match(client):
            return
        desc, suffix = DescInfo.split_imported(cl.description)
        if suffix is None:
            return
        # The Import block should have 'parent-changes'. If not do not strip.
        if not (p4gf_const.P4GF_DESC_KEY_PARENT_CHANGES in suffix or
                p4gf_const.P4GF_DESC_KEY_PARENTS in suffix):    # legacy
            return

        # Client is not a git-fusion client and the the description
        # contains the P4GF_IMPORT_HEADER remove it.
        LOG.debug("removed 'Imported from Git' from change=%s client=%s", cl.change, client)
        LOG.debug2('clean_duplicate_imported_from_git before <<%s>>', cl.description)
        cl.description = desc
        LOG.debug2('clean_duplicate_imported_from_git after <<%s>>', cl.description)

    def __str__(self):
        """Return string representation of self."""
        return "change {0} with {1} files".format(self.change, len(self.files))

    def __repr__(self):
        """Return detailed string representation of self."""
        files = [repr(p4file) for p4file in self.files]
        result = "\n".join(["change: " + str(self.change),
                            "description: " + self.description,
                            "user: " + self.user,
                            "time: " + self.time,
                            "files:",
                            ] + files)
        return result
