#! /usr/bin/env python3.3
"""Translate the incoming Git commits to Perforce changes.

This hook is invoked by git-receive-pack on the remote repository, which
happens when a git push is done on a local repository. It executes on the
remote repository once after all the refs have been updated.

This file must be copied or symlinked into .git/hooks/post-receive

"""

from collections import OrderedDict
import functools
import json
import logging
import os
import sys

import p4gf_env_config  # pylint:disable=unused-import
import p4gf_atomic_lock
import p4gf_branch
from p4gf_branch_id import Assigner
import p4gf_config
import p4gf_const
import p4gf_copy_to_p4
import p4gf_create_p4
from p4gf_fast_push import FastPush
import p4gf_fastexport_marks
import p4gf_git_repo_lock
from p4gf_git_swarm import GSReviewCollection
from p4gf_l10n import _
from p4gf_lfs_row import LFSRow
import p4gf_lock
import p4gf_log
from p4gf_new_depot_branch import NDBCollection
from p4gf_receive_hook import ReceiveHook, _packet_filename, _is_gitref_in_gf,\
    _DONT_CARE, PreReceiveTupleLists
from p4gf_prl_file import PRLFile
import p4gf_path
import p4gf_proc
from p4gf_profiler import with_timer, Timer, Report
from p4gf_push_limits import PushLimits
import p4gf_tag
import p4gf_tempfile
import p4gf_util
import p4gf_version_3
import p4gf_lfs_cache_prune

# cannot use __name__ since it will often be "__main__"
LOG = logging.getLogger("p4gf_post_receive_hook")


class CopyOnlyHook(ReceiveHook):

    """Perform the copy-to-p4 work (assumes preflight happened)."""

    def __init__(self, label, repo_name):
        """Initialize the copy-only hook object."""
        # we'll get prl from the packet later on
        ReceiveHook.__init__(self, label, repo_name, prl=None)
        self.assigner = None
        self.export_data = None
        self.all_marks = p4gf_fastexport_marks.Marks()
        self.gsreview_coll = None
        self.ndb_coll = None
        self.fast_push = None
        self.lfs_row_list = None
        self.g2p = None

    def load_packet(self):
        """Load the packet file into data structures."""
        self.prl, self.assigner, extras = read_packet(self.context)
        if extras.get("fast_push"):
            fast_push = FastPush.from_post_receive(self.context)
            if fast_push:
                self.fast_push = fast_push
                self.prl = fast_push.prl()
                self.ndb_coll = fast_push.ndb_coll()
                return
        if 'fast-export' in extras:
            self.export_data = {p4gf_branch.BranchRef(k): v for (k, v) in
                                extras.pop('fast-export').items()}
            for prt in self.prl.set_heads:
                ref = p4gf_branch.BranchRef(prt.ref)
                marks = self.export_data[ref]['marks']
                self.all_marks.add(ref, marks)
        if 'gsreview' in extras:
            self.gsreview_coll = GSReviewCollection.from_dict(self.context, extras.pop('gsreview'))
        if 'ndb' in extras:
            self.ndb_coll = NDBCollection.from_dict(self.context, extras.pop('ndb'))
        if extras.get("lfs_row_list"):
            self.lfs_row_list = [LFSRow.from_dict(d)
                                 for d in extras.get("lfs_row_list")]

    def do_it(self):
        """Perform all of the setup, processing, and clean up.

        :rtype: int
        :return: status code for the process upon exit.

        """
        group_id = os.environ[p4gf_const.P4GF_FORK_PUSH]
        # Take a write lock, in cooperation with the existing write lock
        # held by the foreground process.
        with p4gf_git_repo_lock.write_lock(self.repo_name, group_id=group_id):
            # Delegate to the base class...
            ec = ReceiveHook.do_it(self)
            # N.B. at this point the p4key and atomic locks are gone
            p4gf_tag.process_tags(self.context, self.prl.tags())
            LOG.debug('after() performing review post-push processing')
            with Timer('swarm post-copy'):
                if self.gsreview_coll:
                    self.gsreview_coll.post_push()
            with Timer('depot branch post-copy'):
                if self.ndb_coll:
                    self.ndb_coll.post_push(self.context)
        p4gf_lfs_cache_prune.prune_lfs_file_cache()
        return ec

    def process(self):
        """Perform the copy to Perforce work.

        :return: status code from process_unsafe() function.
        :rtype: int

        """
        try:
            self.process_unsafe()
        finally:
            # Make sure the atomic lock is removed while we are holding
            # the write lock, lest we allow another foreground process
            # to acquire the write lock but fail (noisily) to acquire
            # the atomic lock. Must also have an active context.
            p4gf_atomic_lock.lock_update_repo_reviews(
                self.context, action=p4gf_atomic_lock.REMOVE)

    def process_unsafe(self):
        """Perform the copy to Perforce work, possibly raising an exception.

        :return: status code
        :rtype: int

        """
        self.load_packet()
        ctx = self.context

        # update the space usage values
        PushLimits(ctx).pre_copy()

        if self.fast_push:
            try:
                self.fast_push.post_receive()
                PRLFile(ctx.config.repo_name).delete()
            except Exception:  # pylint: disable=broad-except
                ctx.record_push_failed_p4key(sys.exc_info()[1])
                # N.B. post-receive hook exit codes are ignored by git, but
                # let's return a non-zero anyway.
                return os.EX_SOFTWARE
        else:
            self.g2p = p4gf_copy_to_p4.G2P(ctx, self.assigner, self.gsreview_coll)
            if self.lfs_row_list:
                self.g2p.lfs_row_list = self.lfs_row_list
            self.g2p.pre_all_heads()
            if self.assigner:
                self._copy_heads()
            _delete_heads(ctx, self.prl)
            ctx.mirror.do_phase_2(ctx)   # pylint:disable=no-member
            # if _copy_heads fails with Exception
            # then this delete of the PRLFILE will be skipped
            # so the next request will apply the retained PRLFile
            PRLFile(ctx.config.repo_name).delete()

        # Update the total disk usage for the repo.
        with Timer('push limits'):
            PushLimits(ctx).post_copy()

        self.context.record_push_success_p4key()
        if not self.fast_push:
            _delete_packet(self.context.config.repo_name)

        # N.B. post-receive hook exit codes are ignored by git.
        return 0

    def _copy_heads(self):
        """For each of the heads being pushed, copy their commits to Perforce."""
        ctx = self.context
        gsreview = self.gsreview_coll
        try:
            branch_dict = ctx.branch_dict()
            for prt in self.prl.set_heads:
                ref = p4gf_branch.BranchRef(prt.ref)
                LOG.debug("copy: current branch_dict:\n%s", p4gf_util.dict_to_log(branch_dict))
                LOG.info("copy %s", prt)
                ref_is_review = gsreview and gsreview.ref_in_review_list(ref)
                LOG.debug("prt.ref %s is_review %s ctx.swarm_reviews %s",
                          prt.ref, ref_is_review, ctx.swarm_reviews)
                commits = self.export_data[ref]['commits']
                self.g2p.copy(prt, commits, self.all_marks)
        except:
            # If the push fails to process successfully, be sure to
            # revert the disk space key values.
            PushLimits.push_failed(ctx)
            raise


@with_timer('delete refs')
def _delete_heads(ctx, prl):
    """For each of the heads being deleted, remove the branch definition from p4gf_config[2]."""
    heads = prl.del_heads
    if not heads:
        return
    branch_dict = ctx.branch_dict()

    for prt in heads:
        LOG.info("delete %s", prt)
        branch = _is_gitref_in_gf(p4gf_branch.BranchRef(prt.ref),
                                  branch_dict, is_lightweight=_DONT_CARE)
        if not branch:
            # Branch is not known to Git Fusion branch, so we've nothing
            # more to do. But branch might still be in the Git repo. Let
            # Git sort that out, report its own Git errors if necessary.
            continue

        if branch.more_equal:
            raise RuntimeError(_(
                "Cannot delete {branch}. First branch defined in p4gf_config cannot be deleted.")
                .format(branch=branch.git_branch_name))
        LOG.info("branch %s marked deleted", branch.git_branch_name)
        branch.deleted = True
        with ctx.switched_to_branch(branch):
            branch.deleted_at_change = p4gf_util.head_change_as_string(ctx, submitted=True)

def delete_head(ctx, prt, lw_only=_DONT_CARE ):
    """For one prt,  remove the branch definition from p4gf_config[2].
       Return True/False is branch was found and deleted"""

    LOG.info("delete_head %s", prt)
    branch_dict = ctx.branch_dict()
    branch = _is_gitref_in_gf(p4gf_branch.BranchRef(prt.ref),
                              branch_dict, is_lightweight=lw_only)
    if not branch:
        # Branch is not known to Git Fusion branch, so we've nothing
        # more to do. But branch might still be in the Git repo. Let
        # Git sort that out, report its own Git errors if necessary.
        LOG.debug("delete_head - branch not in p4gf_config[2] %s", prt.ref)
        return False

    if branch.more_equal:
        raise RuntimeError(_(
            "Cannot delete {branch}. First branch defined in p4gf_config cannot be deleted.")
            .format(branch=branch.git_branch_name))
    LOG.info("delete_head branch %s marked deleted", branch.git_branch_name)
    branch.deleted = True
    with ctx.switched_to_branch(branch):
        branch.deleted_at_change = p4gf_util.head_change_as_string(ctx, submitted=True)
        LOG.info("delete_head branch %s deleted_at_change", branch.deleted_at_change)
    return True

def read_packet(ctx):
    """Read the serialized Git Fusion data from a JSON encoded file.

    :type ctx: :class:`p4gf_context.Context`
    :param ctx: Git Fusion context

    :return: tuple (PreReceiveTupleLists, Assigner, extras dict) for normal push,
             (None, None, package) for fast push,
             or (None, None, None) if no packet file.

    """
    file_name = _packet_filename(ctx.config.repo_name)
    if os.path.exists(file_name):
        with open(file_name) as fobj:
            package = json.load(fobj, object_pairs_hook=OrderedDict)
            if package.get("fast_push"):
                return (None, None, package)
            # read the pre-receive tuples
            prl = PreReceiveTupleLists.from_dict(package.pop('prl'))
            # read the branch dictionary (likely modified by assigner in preflight)
            branch_dict = p4gf_branch.from_dict(package.pop('branch_dict'), ctx.config.p4client)
            ctx.reset_branch_dict(branch_dict)
            # read the branch assignment data
            assign_dict = package.pop('assigner')
            if assign_dict:
                assigner = Assigner.from_dict(assign_dict, branch_dict, prl.set_heads, ctx)
            else:
                assigner = None
            # read the configuration data
            config = p4gf_config.from_dict(package.pop('config'))
            config2 = None
            if 'config2' in package:
                config2 = p4gf_config.from_dict(package.pop('config2'))
            ctx.repo_config.set_repo_config(config, None)
            ctx.repo_config.set_repo_config2(config2, None)
            return (prl, assigner, package)
    return (None, None, None)


def _delete_packet(repo):
    """Remove the JSON packet file upon completion of a successful push.

    :type repo: str
    :param repo: name of the repo

    """
    file_name = _packet_filename(repo)
    if os.path.exists(file_name):
        LOG.debug('removed packet file %s', file_name)
        os.unlink(file_name)
    else:
        LOG.warning('packet file %s missing', file_name)


def forked_execed_main():
    """The main that runs after double-fork + exec."""
    LOG.debug('forked_execed_main() beginning copy-to-p4')
    try:
        ec = main_main()
        LOG.debug('forked_execed_main() returned %s', ec)
        return ec
    finally:
        # The @atexit.register does not seem to fire for double-forked
        # Python processes, so call the profiler report explicitly.
        Report()
        LOG.debug('forked_execed_main() completed')


def forked_main():
    """The main invoked by post-receive hook in a double-forked process.

    To work around OS X fork unsafe issues, do an exec now.
    """
    os.execvp(__file__, sys.argv + ['--do-it-now'])


def main_main():
    """Common bit of logic for all post-receive processing, forked or otherwise.

    :return: exit status code.

    """
    p4gf_branch.init_case_handling()
    repo_name = p4gf_path.cwd_to_repo_name()
    receive_hook = CopyOnlyHook('post-receive', repo_name)
    receive_hook.setup()
    ec = receive_hook.do_it()
    receive_hook.teardown()
    p4gf_tempfile.prune_old_files()
    return ec


def main():
    """Do the post-receive work."""
    for h in ['-?', '-h', '--help']:
        if h in sys.argv:
            print(_('Git Fusion post-receive hook.'))
            return 2
    p4gf_version_3.print_and_exit_if_argv()

    # If P4GF_FORK_PUSH is not set, then this is not the genuine push
    # payload and simply a preliminary request made by the HTTP client.
    # In the case of SSH, it will always be set.
    if p4gf_const.P4GF_FORK_PUSH not in os.environ:
        return 0

    # If this is not an HTTP environment, and the setting to process the
    # push in the foreground process is set, then invoke our "main" without
    # the acrobatics of double-fork+exec. This has the benefit that any
    # push failures are reported to the client in real-time, rather than
    # seemingly going through and failing in the background.
    if p4gf_const.P4GF_FOREGROUND_PUSH in os.environ and 'REMOTE_ADDR' not in os.environ:
        return main_main()

    # Indicate that the lock is about to be acquired by the upcoming
    # background process; the main server process will wait until the lock
    # acquisition is completed by the background process.
    LOG.debug('main() setting up forked process')
    with p4gf_create_p4.Closer():
        p4 = p4gf_create_p4.create_p4_temp_client()
        repo_name = p4gf_path.cwd_to_repo_name()
        p4gf_log.configure_for_repo(repo_name)
        group_id = os.environ[p4gf_const.P4GF_FORK_PUSH]
        with p4gf_lock.RepoLock(p4, repo_name, group_id=group_id) as lock:
            lock.set_acquire_pending()

    # Spawn a process to do the work that pre-receive hook could not do
    # before git updated the references. This is an attempt to prevent
    # any timing issues with respect to git.
    LOG.debug('main() starting processing in forked process')
    func = functools.partial(forked_main)
    p4gf_proc.double_fork(func)
    LOG.debug('main() forked process initiated')

    return 0


if __name__ == "__main__":
    if '--do-it-now' in sys.argv:
        p4gf_log.run_with_exception_logger(forked_execed_main, write_to_stderr=True)
    else:
        p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
