#! /usr/bin/env python3.3
"""Execute a preflight-push hook for each push, if so configured.

The hooks are specified in the p4gf_config file, and can be one of "none",
"pass <msg>", "fail <msg>", or a list of commands to run. If the exit code
of the command(s) is zero, the push is approved, otherwise it is rejected.
The commands are given the pre-receive tuple list, with old and new SHA1
values, and the ref name, via standard input.

This is similar to the preflight-commit config setting, except that it is
run once per push, rather than once per commit.

"""
import logging
import re
import shlex
import sys

import p4gf_config
import p4gf_const
from p4gf_l10n import _, NTR
import p4gf_path
import p4gf_preflight_common
import p4gf_proc
import p4gf_protect

LOG = logging.getLogger(__name__)
ACTION_NONE = NTR('none')
ACTION_PASS = NTR('pass')
ACTION_FAIL = NTR('fail')
ACTION_RUN = NTR('run')


class PreflightPush:

    """Executes an action for a push, during preflight."""

    def __init__(self):
        """Initialize the PreflightPush instance."""
        self.action = None
        self.msg = None
        self.cmds = None

    def __str__(self):
        """Return a str of this object."""
        if self.action is None:
            return 'None'
        elif self.action == ACTION_NONE:
            return 'none'
        elif self.action in [ACTION_PASS, ACTION_FAIL]:
            return '{}: {}'.format(self.action, self.msg)
        else:
            # format list of lists
            cmdstr = "\n".join(str(c) for c in self.cmds)
            return '{0}: {1}'.format(self.action, cmdstr)

    def is_callable(self):
        """Return True if this hook can be invoked."""
        return ACTION_NONE != self.action

    @staticmethod
    def from_context(ctx):
        """Search for a preflight-push configuration in the current config."""
        hook = PreflightPush()
        value = ctx.repo_config.get(p4gf_config.SECTION_GIT_TO_PERFORCE,
                                    p4gf_config.KEY_PREFLIGHT_PUSH)
        value_lower = str(value).strip().lower()
        if re.search(r'none\b', value_lower) or not value_lower:
            hook.action = ACTION_NONE
            LOG.debug('from_context() none')
            return hook
        elif re.search(r'pass\b', value_lower):
            hook.action = ACTION_PASS
            hook.msg = value[4:].strip()
            LOG.debug('from_context() pass: %s', hook.msg)
            return hook
        elif re.search(r'fail\b', value_lower):
            hook.action = ACTION_FAIL
            hook.msg = value[4:].strip()
            LOG.debug('from_context() fail: %s', hook.msg)
            return hook
        else:
            hook.action = ACTION_RUN
            hook.cmds = []
            for line in value.splitlines():
                hook.cmds.append(PreflightCommand.from_line(line))
            LOG.debug('from_context() cmd: %s', hook.cmds)
            return hook

    def __call__(self, ctx, prt):
        """If preflight hook configured, invoke it (or PASS/FAIL it).

        :param ctx: Git Fusion context.
        :param prt: pre-receive tuple (old SHA1, new SHA1, ref name).

        If fail, raise exception detailing why.

        Route hook's stdout and stderr to our stderr.
        """
        if LOG.isEnabledFor(logging.DEBUG3):
            LOG.debug3('call() %s %s', prt, self)
        if self.action is ACTION_NONE:
            return
        elif self.action is ACTION_PASS:
            if self.msg:
                sys.stderr.write(self.msg + '\n')
            LOG.debug('pass %s', prt)
            return
        elif self.action is ACTION_FAIL:
            raise_rejection(prt, self.msg)
        else:
            cmd_line_vars = calc_cmd_line_vars(ctx, prt)
            for cmd in self.cmds:
                cmd.run(cmd_line_vars, prt)


class PreflightCommand:

    """PreflightCommand represents a command to run."""

    def __init__(self, cmd):
        """Initialize an instance of PreflightCommand."""
        self.cmd = cmd

    @staticmethod
    def from_line(line):
        """Create a PreflightCommand from the given configuration value."""
        return PreflightCommand(shlex.split(line.strip()))

    def run(self, cmd_line_vars, prt):
        """Run the command for the given push."""
        cmd = [substitute_cmd_line_vars(cmd_line_vars, word) for word in self.cmd]
        LOG.debug3('run cmd: %s', cmd)
        LOG.debug3('run prt: %s', prt)
        d = p4gf_proc.popen_no_throw(cmd, prt.to_line().encode())
        LOG.debug3('run result: %s', d)
        msg = p4gf_path.join_non_empty('\n', d['out'], d['err'])
        if d['ec']:
            raise_rejection(prt, msg)
        sys.stderr.write(msg)

    def __repr__(self):
        """Return a debug representation of this."""
        return "PreflightCommand[cmd={}]".format(self.cmd)

    def __str__(self):
        """Return a string representation of this."""
        return ' '.join(self.cmd)


def raise_rejection(prt, msg):
    """The push was rejected by the preflight hook, raise an exception."""
    frm = _('preflight-push rejected: {prt} {msg}')
    raise p4gf_preflight_common.PreflightException(frm.format(prt=prt, msg=msg))


def substitute_cmd_line_vars(cmd_line_vars, word):
    """Perform p4 trigger-like variable substitution."""
    result = word
    for k, v in cmd_line_vars.items():
        result = result.replace(k, v)
    return result


def calc_cmd_line_vars(ctx, prt):
    """Return our mapping of command-line variable substitutions.

    DANGER: Little Bobby Tables! Sanitize your shell inputs!

    These become strings in a command. Watch out for any input that a user
    can control such as repo name or jobs list.

    """
    client_host = p4gf_protect.get_remote_client_addr()
    git_action = 'update'
    if prt.old_sha1 == p4gf_const.NULL_COMMIT_SHA1:
        git_action = 'new'
    elif prt.new_sha1 == p4gf_const.NULL_COMMIT_SHA1:
        git_action = 'delete'
    user = ctx.foruser if ctx.foruser else ctx.authenticated_p4user
    r = {
        '%repo%':       shlex.quote(ctx.config.repo_name),
        '%user%':       shlex.quote(user),
        '%ref%':        shlex.quote(prt.ref),
        '%client%':     ctx.p4.client,
        '%clienthost%': client_host if client_host else '',
        '%command%':    'user-submit',
        '%quote%':      '"',
        '%serverport%': ctx.p4.port,
        '%formname%':   NTR('new'),
        '%formtype%':   NTR('change'),
        '%git-action%': git_action
    }
    return r
