#! /usr/bin/env python3.3
"""Report the estimated size for a newly created P4 based Git Fusion repo."""

import locale
import logging
import os
import re
import shutil
import sqlite3
import sys

import p4gf_env_config  # pylint: disable=unused-import
import p4gf_context
import p4gf_branch
import p4gf_create_p4
from p4gf_l10n import _, NTR
import p4gf_log
import p4gf_const
import p4gf_config
import p4gf_p4cache
import p4gf_p4spec
import p4gf_tempfile
import p4gf_util

DESCRIPTION = _("""Report the estimated size for a newly created P4 based Git Fusion repo.

Requires a p4gf_config input data file.
This may be run against any Git Fusion p4d , as long as the Git Fusion code base
in properly installed locally. No Git Fusion instance need be created.
The serverid is spoofed only for the purpuse of creating temp client names.
If run in the Git Fusion user environment, it will use the P4PORT set therein,
honoring the P4GF_ENV configuration.

""")

# cannot use __name__ since it will often be "__main__"
LOG = logging.getLogger("p4gf_repo_size")
TREE_SIZE = 500
COMMIT_SIZE = 300
SHA1_SIZE = 40
SPOOFED_SERVER_ID = 'temp-server-id'


def commonprefix(l):
    """Path-component aware version of os.path.commonprefix.

    This unlike the os.path.commonprefix version
    always returns path prefixes as it compares path component wise.
    """
    cp = []
    ls = [p.split('/') for p in l]
    ml = min(len(p) for p in ls)
    for i in range(ml):
        s = set(p[i] for p in ls)
        if len(s) != 1:
            break
        cp.append(s.pop())
    return '/'.join(cp)


def user_has_admin(p4):
    """Return True if this user has admin permissions somewhere."""
    r = p4.run(['protects'])
    for rr in r:
        if rr['perm'] == 'admin' or rr['perm'] == 'super':
            return True
    return False


def _get_all_parent_paths(depotpath):
    """Return an array of each descending parent to this path.

    Assume the path end with a file - not a directory.
    """
    # Expect paths to start with '//' and so remove it.
    parents = []
    if depotpath.startswith('//'):
        # remove the leading '//depot/'
        parts = re.sub(r'//[^/]+/', '', depotpath).split('/')
    else:
        parts = depotpath.split('/')
    for i in range(1, len(parts)):
        parents.append('/'.join(parts[:i]))
    return parents


class RepoSize():
    """"Estimate the git and P4 disk size of cloning a new repo."""

    def __init__(self, config_path, p4port, p4user, _locale, _start_change, need_server_id):
        """Initialize a RepoSize instance."""
        p4gf_const.READ_ONLY = True  # READ_ONLY will permit ctx to cleanup temporary clients
        p4gf_const.P4GF_USER = p4user
        self.config_path = config_path
        self.p4port = p4port
        self.p4user = p4user
        self.locale = _locale
        self.repo_name = None
        self.repo_config = None
        self.start_change = _start_change
        self.changes = []
        self.fstats = []
        self.git_data = 0
        self.p4_data = 0
        self.git_file_data = 0  # use P4 digest to prevent increase from duplicated file data
        self.p4_file_data = 0
        self.file_rev_count = 0
        self.commit_count = 0
        self.commit_data = 0
        self.tree_count = 0
        self.tree_data = 0
        self.key_count = 0
        self.key_data = 0
        self.sql_db = None
        self.tempdir = None
        self.db_file_name = None
        self.ctx = None
        self.is_empty = False
        self.longest_shared_view = ''
        if need_server_id:
            self.tempdir = p4gf_tempfile.new_temp_dir()
            p4gf_const.P4GF_HOME = self.tempdir
            server_id_path = p4gf_util.server_id_file_path()
            with open(server_id_path, 'w') as sip:
                sip.write(SPOOFED_SERVER_ID)
                sip.write('\n')

        self._create_database()
        locale.setlocale(locale.LC_ALL, self.locale)

    def _create_database(self):
        """Use sqlite3 maintain separate lists of unique MD5 and tree paths."""
        # pylint: disable=line-too-long
        db_file = p4gf_tempfile.new_temp_file(prefix='repo_size_', suffix='.db', delete=False)
        self.sql_db = sqlite3.connect(database=db_file.name, isolation_level="EXCLUSIVE")
        self.db_file_name = db_file.name
        self.sql_db.execute("PRAGMA synchronous = OFF")
        self.sql_db.execute("CREATE TABLE md5  (key TEXT PRIMARY KEY)")
        self.sql_db.execute("CREATE TABLE changes  (key TEXT, branch_id Text, "
                            " PRIMARY KEY(key, branch_id))")
        self.sql_db.execute("CREATE TABLE tree (key TEXT, change TEXT, PRIMARY KEY(key, change))")

    def _add_md5_if_not_present(self, md5):
        """Add the md5 key if not already present.

        Return True if added - else False.
        """
        cursor = self.sql_db.execute("SELECT key FROM md5 WHERE key=?", (md5,))
        row = cursor.fetchone()
        if row is None:
            self.sql_db.execute("INSERT INTO md5 VALUES(?)", (md5,))
            return True
        else:
            return False

    def _add_change_if_not_present(self, change, branch_id):
        """Add the change key if not already present.

        Return True if added - else False.
        """
        cursor = self.sql_db.execute("SELECT key FROM changes WHERE key=? AND branch_id=?",
                                     (change, branch_id))
        row = cursor.fetchone()
        if row is None:
            self.sql_db.execute("INSERT INTO changes VALUES(?, ?)", (change, branch_id))
            return True
        else:
            return False

    def _get_row_count(self, tablename):
        """Return row count."""
        query = "SELECT Count(*) from {}".format(tablename)
        count = self.sql_db.execute(query)
        return count

    def _add_tree_if_not_present(self, dirpath, changenum):
        """Add the dirpath key if not already present.

        Return True if added - else False.
        """
        cursor = self.sql_db.execute(
                "SELECT key FROM tree WHERE key=? AND change=?", (dirpath, changenum))
        row = cursor.fetchone()
        if row is None:
            self.sql_db.execute("INSERT INTO tree VALUES(?, ?)", (dirpath, changenum))
            return True
        else:
            return False

    def _get_longest_shared_view(self, viewlines):
        """Determine the longest shared common path for a branches viewlines.

        Do not add trees for subpaths higher than this shared common path.
        """
        common_roots = []
        for view in viewlines:
            if view.startswith('//'):
                # remove the leading '//depot/'
                view = re.sub(r'//[^/]+/', '', view)
            # remove the trailing /xxx - where xxx may be a file or mask
            front = view[0:view.rfind('/')]
            common_roots.append(front)
        if 'DEBUG' in os.environ:
            print("_get_longest_shared_view common_roots = {}".format(common_roots))
        if len(common_roots) == 1:
            # if there is only one view, fabricate another bogus view adding one char
            # this will allow the following call to extract the common prefix on a single view
            common_roots.append(common_roots[0] + 'z')
        self.longest_shared_view = commonprefix(common_roots)
        if 'DEBUG' in os.environ:
            print("_get_longest_shared_view longest_shared_view = {}".format(
                self.longest_shared_view))

    def unlock_delete_temp_clients(self):
        """Unlock the temp clients so the user may consequently delete them."""
        pattern = p4gf_const.P4GF_REPO_TEMP_CLIENT.format(server_id=p4gf_util.get_server_id(),
                                                          repo_name=self.ctx.config.repo_name,
                                                          uuid='*')
        clients = self.ctx.p4gfrun('clients', '-e', pattern)
        options = re.sub('locked', 'unlocked', p4gf_const.CLIENT_OPTIONS)
        for client in clients:
            client_name = client['client']
            p4gf_p4spec.ensure_spec_values(
                self.ctx.p4gf, spec_type='client', spec_id=client_name,
                values={'Options': options})
            p4gf_p4cache.delete_client(self.ctx.p4gf, client_name)

    def cleanup(self):
        """Clean up the context and clientpool."""
        try:
            self.unlock_delete_temp_clients()
            self.ctx.disconnect()
            if self.tempdir:
                shutil.rmtree(self.tempdir)
            if self.db_file_name:
                os.unlink(self.db_file_name)
        except Exception as e:  # pylint: disable=W0703
            print(_('error on disconnect {exception}').format(exception=e))

    def estimate_size(self):
        """Perforce the repo size estimate."""
        p4 = p4gf_create_p4.create_p4_temp_client(port=self.p4port, user=self.p4user)
        p4gf_branch.init_case_handling(p4)
        self.repo_name = 'estimate_repo_size_' + p4gf_util.uuid(p4)
        self.repo_config = p4gf_config.RepoConfig.from_local_files(
                  self.repo_name, p4, self.config_path, None)
        ctx = p4gf_context.create_context(self.repo_name)
        self.ctx = ctx
        ctx.p4gf = p4
        with ctx.exclusive_access():
            ctx.repo_config = self.repo_config
            ctx.config.p4user = self.p4user
            ctx.config.p4port = self.p4port
            ctx.connect()
            # Perform the actual size estimation now that everything is
            # properly set up.
            if ctx.union_view_empty():
                self.is_empty = True
            else:
                self.estimate_files()
            self.cleanup()

    def count_trees_in_path(self, depotpath, changenum):
        """Count the trees - list of descending dirs - to a path.

        Use a sqlite3 single table to de-dupe trees.
        """
        trees = _get_all_parent_paths(depotpath)
        longest_shared_view_len = len(self.longest_shared_view)
        for t in trees:
            if len(t) > longest_shared_view_len:
                if self._add_tree_if_not_present(t, changenum):
                    self.tree_count += 1
                    self.tree_data += TREE_SIZE

    def estimate_files(self):
        """For each change, add to the commit, tree, key, file and size counts."""
        for b in self.ctx.branch_dict().values():
            branch_rev_count = 0
            with self.ctx.switched_to_branch(b):
                _range = '@{},#head'.format(self.start_change)
                path_range = self.ctx.client_view_path() + _range
                self._get_longest_shared_view(b.view_p4map.lhs())
                r = self.ctx.p4run('fstat', '-Olf',  path_range)
                branch_revs = len(r)
                branch_rev_count = 0
                for rr in r:
                    if os.isatty(sys.stdout.fileno()) and 'NOPROGRESS' not in os.environ:
                        sys.stdout.write(
                            _("\r...  {num_revs:12} file revisions out of "
                              "{branch_revs:12} for branch '{branch_name}' "
                              "total commits = {commits}")
                            .format(num_revs=branch_rev_count,
                                    branch_revs=branch_revs,
                                    branch_name=b.git_branch_name,
                                    commits=self.commit_count))
                    change_num = rr['headChange']
                    depot_file = rr['depotFile']
                    file_size = None
                    if 'fileSize' in rr:
                        file_size = rr['fileSize']
                    digest = None
                    if 'digest' in rr:
                        digest = rr['digest']

                    if self._add_change_if_not_present(change_num, b.branch_id):
                        self.commit_count += 1
                        self.commit_data += COMMIT_SIZE
                        # Keys: one each of these per commit
                        # git-fusion-index-branch-{repo_name},{change_num},{branch_id}'
                        # git-fusion-{repo_name}-rev-sha1-{change_num}
                        self.key_count += 2
                        self.key_data += SHA1_SIZE  # size of git-fusion-index-branch... value

                    if file_size:
                        self.p4_file_data += int(file_size)
                    if digest and file_size and self._add_md5_if_not_present(digest):
                        self.git_file_data += int(file_size)
                    self.file_rev_count += 1
                    branch_rev_count += 1
                    self.count_trees_in_path(depot_file, change_num)
                    # git-fusion-{repo_name}-rev-sha1-{change_num}
                    # each file path and its sha1 is added to this key value
                    # accumulate size of this key value
                    self.key_data += len(depot_file) + SHA1_SIZE

        if os.isatty(sys.stdout.fileno()) and 'NOPROGRESS' not in os.environ:
            # clear the progress line
            sys.stdout.write("\r {0:100s}".format(' '))

        self.git_data = self.commit_data + self.tree_data + self.git_file_data
        self.p4_data = self.commit_data + self.tree_data + self.key_data
        sys.stdout.write('\r {0:80}'.format(' '))  # clean the progress line

    def get_changes(self):
        """Get the a list of all changes from all branches."""
        for b in self.ctx.branch_dict().values():
            with self.ctx.switched_to_branch(b):
                path_range = self.ctx.client_view_path() + '@1,#head'
                r = self.ctx.p4run('changes', path_range)
                for rr in r:
                    self.changes.append(rr['change'])

    def report(self):
        """Print the results."""
        if self.is_empty:
            print(_("Repo views in {path} are empty.".format(path=self.config_path)))
        else:
            onemeg_f = float(1024*1024)
            print('\n'.join([
                _('\nRepo config: {path}').format(
                    path=self.config_path),
                _('File revision count:               {rev_count:>20}').format(
                    rev_count=locale.format("%d", self.file_rev_count, grouping=True)),
                _('\ngit clone size:'),
                _('  Commit count:                      {commit_count:>20}').format(
                    commit_count=locale.format("%d", self.commit_count, grouping=True)),
                _('  Commit objects size*:              {commit_data: >#019.3f}M').format(
                    commit_data=float(self.commit_data)/onemeg_f),
                _('  Tree count:                        {tree_count:>20}').format(
                    tree_count=locale.format("%d", self.tree_count, grouping=True)),
                _('  Tree objects size*:                {tree_data: >#019.3f}M').format(
                    tree_data=float(self.tree_data)/onemeg_f),
                _('  Git total uncompressed file data:  {git_file_data: >#019.3f}M').format(
                    git_file_data=float(self.git_file_data)/onemeg_f),
                _('  Git total data:                    {git_data: >#019.3f}M').format(
                    git_data=float(self.git_data)/onemeg_f),
                _('\nP4 data added:'),
                _('  Commit count:                      {commit_count:>20}').format(
                    commit_count=locale.format("%d", self.commit_count, grouping=True)),
                _('  Commit objects size*:              {commit_data: >#019.3f}M').format(
                    commit_data=float(self.commit_data)/onemeg_f),
                _('  Tree count:                        {tree_count:>20}').format(
                    tree_count=locale.format("%d", self.tree_count, grouping=True)),
                _('  Tree objects size*:                {tree_data: >#019.3f}M').format(
                    tree_data=float(self.tree_data)/onemeg_f),
                _('  P4 key count:                      {key_count:>20}').format(
                    key_count=locale.format("%d", self.key_count, grouping=True)),
                _('  P4 key data:                       {key_data: >#019.3f}M').format(
                    key_data=float(self.key_data)/onemeg_f),
                _('  P4 total data:                     {p4_data: >#019.3f}M').format(
                    p4_data=float(self.p4_data)/onemeg_f),
                _('                                           '
                  '* using average commit size = {commit_size}').format(commit_size=COMMIT_SIZE),
                _('                                           '
                  '* using average tree  size = {tree_size}').format(tree_size=TREE_SIZE)
                ]))


def main():
    """Parse the command-line arguments and report on repo metrics."""
    # pylint: disable=too-many-branches
    desc = _(DESCRIPTION)
    parser = p4gf_util.create_arg_parser(desc=desc)
    parser.add_argument(NTR('--config'),  metavar=NTR('config'),
                        help=_('Path to Git Fusion p4gf_config file (required)'), required=True)
    parser.add_argument('-u', '--p4user', metavar='p4user',
                        help=_('Perforce user'))
    parser.add_argument('-p', '--p4port', metavar='p4port',
                        help=_('Perforce server'))
    parser.add_argument('--start', metavar='start',
                        help=_('Starting changelist (default=1)'), default='1')
    parser.add_argument('--locale', metavar='locale', default='en_US.UTF-8',
                        help=_('system locale setting'))
    args = parser.parse_args()

    need_serverid = False
    try:
        p4gf_util.get_server_id()
    except:  # pylint: disable=W0702
        need_serverid = True

    # If connect args not passed, check that the environment is set.
    if not args.p4port:
        if 'P4PORT' not in os.environ and 'P4GF_ENV' not in os.environ:
            print(
              _("Neither --p4port is an argument nor are P4GF_ENV and P4PORT in the environment."))
            sys.exit(0)
        if 'P4PORT' in os.environ:
            args.p4port = os.environ['P4PORT']
    else:
        # Set the requested port for Git Fusion's environment
        os.environ['P4PORT'] = args.p4port

    if not args.p4user:
        if 'P4USER' not in os.environ:
            print(_("Neither --p4user is an argument nor is P4USER in the environment."))
            sys.exit(0)
        else:
            args.p4user = os.environ['P4USER']
    else:
        # Set the requested user for Git Fusion's environment
        os.environ['P4USER'] = args.p4user

    try:
        if int(args.start) < 1:
            raise ValueError
    except ValueError:
        print(
          _("--start changenum must be numeric and > '0'."))
        sys.exit(0)

    repo_size = RepoSize(args.config, args.p4port, args.p4user,
                         args.locale, args.start, need_serverid)
    repo_size.estimate_size()
    repo_size.report()


if __name__ == "__main__":
    p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
