#! /usr/bin/env python3.3
"""Report on the status of Git Fusion repositories.

Detecting stale p4key lock owners only works when the client was able to
disconnect properly. This is not the case if the process was terminated
abruptly. We can, of course, check for dead processes on the same server.

Detecting stale atomic locks (user reviews) is nigh impossible since
there may be multiple repositories sharing a view path, hence there would
be false positives.

Detecting missing changes requires that all Git commits which originated
from Perforce changes must have the "Copied from Perforce" text in the
commit message. Otherwise, we cannot be sure where it came from.

"""

from contextlib import ExitStack
import json
import os
import sys

import pygit2
from P4 import P4Exception

import p4gf_branch
from p4gf_desc_info import DescInfo
import p4gf_config
import p4gf_const
import p4gf_context
import p4gf_create_p4
import p4gf_git
import p4gf_git_repo_lock
import p4gf_gitmirror
from p4gf_l10n import _, NTR
import p4gf_lock
import p4gf_log
from p4gf_p2g_rev_range import RevRange
from p4gf_p4changelist import P4Changelist
import p4gf_p4key as P4Key
import p4gf_p4cache
import p4gf_proc
import p4gf_repo_dirs
import p4gf_translate
import p4gf_util


class RepoStatus:

    """Provides methods for detecting problems in a repository."""

    def __init__(self, p4, repo_name, server_id):
        """Initialize a RepoStatus instance."""
        self.p4 = p4
        self.repo_name = repo_name
        self.server_id = server_id
        self._repo = None
        self._repo_dirs = None
        self._repo_config = None
        self._branch_dict = None

    @property
    def repo(self):
        """Retrieve the repository property."""
        if self._repo is None:
            self._repo = pygit2.Repository(self.repo_dirs.GIT_DIR)
        return self._repo

    @property
    def repo_dirs(self):
        """Retrieve the repo dirs object."""
        if self._repo_dirs is None:
            self._repo_dirs = p4gf_repo_dirs.from_p4gf_dir(p4gf_const.P4GF_HOME, self.repo_name)
        return self._repo_dirs

    @property
    def repo_config(self):
        """Retrieve the repoistory configuration property."""
        if self._repo_config is None:
            self._repo_config = p4gf_config.RepoConfig.from_depot_file(
                self.repo_name, self.p4, create_if_missing=True)
        return self._repo_config

    @property
    def branch_dict(self):
        """Retrieve the branch dictionary property."""
        if self._branch_dict is None:
            self._branch_dict = p4gf_branch.config_to_branch_dict(
                config=self.repo_config, p4=self.p4, rhs_client_name=self.p4.client)
        return self._branch_dict

    def check_status_key(self):
        """Report on the repo push status key."""
        status_key_name = P4Key.calc_repo_status_p4key_name(self.repo_name)
        status_key = P4Key.get(self.p4, status_key_name)
        if status_key == '0':
            # no status yet
            return
        if 'completed successfully' not in status_key:
            print('Push status appears incomplete: {}'.format(status_key))

    def check_push_start_key(self):
        """Report on the repo push start key value."""
        start_key_name = P4Key.calc_push_start_p4key_name(self.repo_name)
        start_key = P4Key.get(self.p4, start_key_name)
        if start_key != '0':
            print('Push start key presence indicates push did not complete (yet).')

    def check_last_copied_change(self):
        """Compare the last-copied-change key to the git-fusion-index-last- keys."""
        highest = dict()
        highest_branch_id = None
        for branch_id in self.branch_dict.keys():
            # Look for the last known change on that branch.
            lcn_key_name = p4gf_const.P4GF_P4KEY_INDEX_LCN_ON_BRANCH.format(
                repo_name=self.repo_name, branch_id=branch_id)
            lcn_key = P4Key.get(self.p4, lcn_key_name)
            if ',' in lcn_key:
                change_num, _sha1 = lcn_key.split(',')
                if branch_id in highest:
                    if int(change_num) > highest[branch_id]:
                        highest[branch_id] = int(change_num)
                else:
                    highest[branch_id] = int(change_num)
                if not highest_branch_id or int(change_num) > highest[highest_branch_id]:
                    highest_branch_id = branch_id
        if highest_branch_id:
            highest_change = highest[highest_branch_id]
            last_copied_key_name = P4Key.calc_last_copied_change_p4key_name(
                self.repo_name, self.server_id)
            last_copied_key = P4Key.get(self.p4, last_copied_key_name)
            if int(last_copied_key) != highest_change:
                il = 'git-fusion-index-last-<repo>,<branch>'
                lccn = 'git-fusion-<repo>-<server>-last-copied-changelist-number'
                print('{} appears out of sync with {}'.format(il, lccn))

    def check_stale_file_locks(self):
        """Detect stale file locks and report any findings."""
        for pid in p4gf_git_repo_lock.detect_stale_readers(self.repo_name):
            print('Stale read lock pid {}'.format(pid))
        for pid in p4gf_git_repo_lock.detect_stale_writers(self.repo_name):
            print('Stale write lock pid {}'.format(pid))

    def check_stale_p4key_owners(self):
        """Detect the stale p4key lock owners and report any findings."""
        lock_key_name = p4gf_const.P4GF_P4KEY_LOCK_VIEW_OWNERS.format(repo_name=self.repo_name)
        lock_key = P4Key.get(self.p4, lock_key_name)
        if lock_key == '0':
            return
        print('Lock owner key is non-zero, checking for stale owners...')
        content = json.loads(lock_key)
        (stale_owners, _fresh_owners) = p4gf_lock.detect_stale_owners(self.p4, content)
        if stale_owners:
            for owner in stale_owners:
                print('Stale p4key lock pid {}, started at {}'.format(
                    owner['process_id'], owner['start_time']))
            print('>> Run p4gf_lock_status.py for further details.')
        else:
            print('No apparent stale lock owners, check other Git Fusion instances')

    def check_git_refs(self):
        """Compare the git references to the changes in Perforce."""
        for branch_id, branch in self.branch_dict.items():
            # We can only examine named, live git branches...
            if branch.git_branch_name and not branch.deleted:
                # Look for the last known change on that branch.
                lcn_key_name = p4gf_const.P4GF_P4KEY_INDEX_LCN_ON_BRANCH.format(
                    repo_name=self.repo_name, branch_id=branch_id)
                lcn_key = P4Key.get(self.p4, lcn_key_name)
                if ',' in lcn_key:
                    _change_num, sha1 = lcn_key.split(',')
                    # compare with the tip of that branch in the repo
                    git_branch = self.repo.lookup_branch(branch.git_branch_name)
                    if git_branch:
                        git_obj = git_branch.peel()
                        if git_obj.id.hex != sha1:
                            print('Git branch {} is out of sync with Perforce'.format(
                                branch.git_branch_name))
                    else:
                        print('Git branch {} appears to be missing from repo'.format(
                            branch.git_branch_name))
                else:
                    print('git-fusion-index-last-{},{} key appears to be unset'.format(
                        self.repo_name, branch_id))

    def check_index_keys(self):
        """Check the git-fusion-index-branch keys for invalid values."""
        def key_callback(key):
            """Invoked for each key dict from p4 keys."""
            if key['value'] not in self.repo:
                print('Invalid SHA1 for {}'.format(key['key']))
            return False

        key_pattern = p4gf_const.P4GF_P4KEY_INDEX_OT.format(
            repo_name=self.repo_name, change_num='*', branch_id='*')
        P4Key.query_keys(self.p4, key_pattern, key_callback)

    def check_git_branches(self):
        """Compare the git branches to those in the p4gf_config files."""
        git_branches = set(self.repo.listall_branches())
        for branch in self.branch_dict.values():
            if branch.git_branch_name:
                git_branches.discard(branch.git_branch_name)
        for branch in git_branches:
            print('p4gf_config/2 is missing branch named "{}"'.format(branch))

    @staticmethod
    def check_mirror(ctx):
        """Report the number of missing commits and trees in the object cache."""
        (missing_commits, missing_trees) = p4gf_gitmirror.check_missing_objects(ctx)
        if missing_commits:
            print('Missing {} commit(s)'.format(missing_commits))
        if missing_trees:
            print('Missing {} tree(s)'.format(missing_trees))
        if missing_commits or missing_trees:
            print('>> Run p4gf_gitmirror.py --fix to try to correct this.')

    @staticmethod
    def check_missing_changes(ctx):
        """Report any commits for which there is no corresponding changelist."""
        commits_in_git = p4gf_git.all_git_commits(ctx.repo)
        commits_in_p4 = _discover_converted_commits(ctx)
        unconverted_commits = commits_in_git - commits_in_p4
        for candidate in unconverted_commits:
            commit = ctx.repo[candidate]
            if p4gf_const.P4GF_EXPORT_HEADER not in commit.message:
                print('Missing changelist for commit {}'.format(candidate))

    def check_git_fsck(self):
        """Run git-fsck and ensure it reports no problems."""
        if not os.path.exists(self.repo_dirs.GIT_WORK_TREE):
            return
        os.chdir(self.repo_dirs.GIT_WORK_TREE)
        result = p4gf_proc.popen_no_throw(['git', 'fsck'])
        if result['ec']:
            print("git-fsck reported trouble:\n{}".format(result['out']))


def check_repo(p4, repo_name, server_id, ctx=None):
    """Report the status of the named repository.

    :param p4: P4 object
    :param repo_name: name of repository to check.
    :param server_id: identifier of the current instance.
    :param ctx: Git Fusion context; if None, performs only quick checks.

    """
    checker = RepoStatus(p4, repo_name, server_id)
    checker.check_status_key()
    checker.check_push_start_key()
    checker.check_last_copied_change()
    checker.check_stale_p4key_owners()
    checker.check_stale_file_locks()
    checker.check_git_refs()
    checker.check_index_keys()
    checker.check_git_branches()
    checker.check_git_fsck()
    if ctx:
        # Client pool access (which is used for switching branch views)
        # requires a read lock.
        with p4gf_git_repo_lock.read_lock(repo_name):
            # The context is provided when we want to check those things
            # that are more expensive (i.e. missing cache objects) than
            # just a few key comparisons. Things like missing objects are
            # best checked when we have the lock, thus ensuring we are not
            # witnessing an incomplete mirroring operation, and the like.
            RepoStatus.check_mirror(ctx)
            if not ctx.add_copied_from_perforce:
                print('WARNING: enable-add-copied-from-perforce is set to false,'
                      ' results may be inaccurate')
            RepoStatus.check_missing_changes(ctx)


def _discover_converted_commits(ctx):
    """Build a set of all commits that have corresponding changes.

    :param ctx: context for the repo of interest.
    :return: set of commit SHA1 values that have been converted.

    """
    # The set of commits that form the basis of converted changes.
    changes = set()

    def append(change):
        """Add the commit SHA1 to the set of converted commits."""
        di = DescInfo.from_text(change.description)
        # Not all changelists originated from git, after all.
        if di:
            changes.add(di.sha1)

    range_as_string = RevRange.from_start_stop(ctx, "@1", "#head").as_range_string()
    all_branches = [b[1] for b in sorted(ctx.branch_dict().items())]
    for b in all_branches:
        with ctx.switched_to_branch(b):
            path_range = ctx.client_view_path() + range_as_string
            P4Changelist.get_changelists(ctx.p4, path_range, append)
    return changes


def main():
    """Parse the command-line arguments and report on repositories."""
    desc = _("Report on the status of Git Fusion repositories.")
    parser = p4gf_util.create_arg_parser(desc=desc)
    parser.add_argument(NTR("repo"), metavar=NTR("R"), nargs="*",
                        help=_("name of the repositories to be checked"))
    parser.add_argument("-a", "--all", action="store_true",
                        help=_("process all known Git Fusion repositories"))
    parser.add_argument("-F", "--full", action="store_true",
                        help=_("perform extensive (and time costly) checks"))
    args = parser.parse_args()

    # Check that either --all, or "repo" was specified.
    if not args.all and len(args.repo) == 0:
        sys.stderr.write(_("Missing view names; try adding --all option.\n"))
        sys.exit(2)
    if args.all and len(args.repo) > 0:
        sys.stderr.write(_("Ambiguous arguments. Choose --all or a repo name.\n"))
        sys.exit(2)

    server_id = p4gf_util.has_server_id_or_exit()
    p4gf_proc.init()
    with p4gf_create_p4.Closer():
        p4 = p4gf_create_p4.create_p4_temp_client()
        if not p4:
            sys.exit(1)
        # Sanity check the connection (e.g. user logged in?) before proceeding.
        try:
            p4gf_p4cache.fetch_client(p4)
        except P4Exception as e:
            sys.stderr.write(_("P4 exception occurred: {exception}\n").format(exception=e))
            sys.exit(1)
        p4gf_branch.init_case_handling(p4)
        if args.all:
            repos = p4gf_util.repo_config_list(p4)
            if len(repos) == 0:
                print(_("No Git Fusion repositories found, nothing to do."))
                sys.exit(0)
        else:
            repos = args.repo

        for repo in repos:
            repo_name = p4gf_translate.TranslateReponame.git_to_repo(repo)
            print(_("Processing repository {0}...").format(repo_name))
            if args.full:
                # Full checks require a context and the p4key lock.
                try:
                    ctx = p4gf_context.create_context(repo_name)
                    with ExitStack() as stack:
                        stack.enter_context(ctx)
                        ctx.repo_lock = p4gf_lock.RepoLock(ctx.p4gf, repo_name, blocking=False)
                        stack.enter_context(ctx.repo_lock)
                        check_repo(p4, repo_name, server_id, ctx)
                except p4gf_config.ConfigLoadError:
                    print(_("error loading config file"))
                except p4gf_lock.LockBusy:
                    print(_("unable to acquire lock"))
            else:
                # Perform a quick check, without a context.
                check_repo(p4, repo_name, server_id)


if __name__ == "__main__":
    p4gf_log.run_with_exception_logger(main, write_to_stderr=True)
