#! /usr/bin/env python3.3
"""Common functions for all P4GF server implementations."""

from collections import defaultdict, namedtuple
from contextlib import contextmanager, ExitStack
import functools
import json
import logging
import os
import re
import signal
import sys
import time

import p4gf_atomic_lock
from P4 import Map, P4Exception
import p4gf_branch
import p4gf_config
import p4gf_const
import p4gf_context
import p4gf_copy_p2g
import p4gf_create_p4
from p4gf_desc_info import DescInfo
import p4gf_mem_gc
import p4gf_git
import p4gf_gitmirror
import p4gf_git_repo_lock
from p4gf_git_swarm import GSReviewCollection
import p4gf_group
import p4gf_init_host
import p4gf_imports
import p4gf_init
import p4gf_init_repo
import p4gf_p4key as P4Key
from p4gf_l10n import _, NTR
import p4gf_lock
import p4gf_log
import p4gf_p4msg
import p4gf_p4msgid
import p4gf_p4spec
from p4gf_prl_file import PRLFile
import p4gf_proc
from p4gf_profiler import Timer, with_timer
import p4gf_protect
from p4gf_repolist import RepoList
import p4gf_translate
import p4gf_util
from p4gf_util import CommandError
import p4gf_version_3
import p4gf_read_permission
import p4gf_lock_status

LOG = logging.getLogger(__name__)

CMD_GIT_UPLOAD_PACK = "git-upload-pack"  # aka fetch/pull/clone
CMD_GIT_RECEIVE_PACK = "git-receive-pack"  # aka push
CMD_LFS_OBJECTS = "objects"  # aka Git LFS pre-push metadata

COMMAND_TO_PERM = {
    CMD_GIT_UPLOAD_PACK:  p4gf_group.PERM_PULL,
    CMD_GIT_RECEIVE_PACK: p4gf_group.PERM_PUSH,
    # All LFS requests must be treated as "pull" until we can determine
    # exactly where in the depot (repo and branch) the files are going.
    CMD_LFS_OBJECTS:      p4gf_group.PERM_PULL
}


class ServerCommonException(Exception):

    """base class for exceptions that don't require logging."""

    pass


class BadRequestException(ServerCommonException):

    """bad args or similar."""

    pass


class PerforceConnectionFailed(ServerCommonException):

    """trouble with p4 connection."""

    pass


class SpecialCommandException(ServerCommonException):

    """requested repo is actually a special command."""

    pass


class RepoNotFoundException(ServerCommonException):

    """requested repo does not exist."""

    pass


class RepoInitFailedException(ServerCommonException):

    """problem initializing a repo."""

    pass


class MissingSubmoduleImportUrlException(ServerCommonException):

    """repo has Stream imports but no configured ssh-url."""

    pass


class ReadOnlyInstanceException(ServerCommonException):

    """Git Fusion instance is configured to reject pushes."""

    pass


class TerminatingException(ServerCommonException):

    """A terminating signal was received during pull/push processing."""

    pass


class ExceptionAuditLogger:

    """Print errors to standard channels, then propagate."""

    def __init__(self):
        """Initialize the logger."""
        pass

    def __enter__(self):
        """On enter do nothing."""
        return self

    def __exit__(self, exc_type, exc_value, _traceback):
        """On exit, log exceptions in some cases."""
        # Skip calls to exit().
        if exc_type == SystemExit:
            return False

        # Skip logging of certain known exception types
        if isinstance(exc_value, ServerCommonException):
            return False

        if exc_value:
            msg = "{}".format(exc_value)
            # Improve the readability of the error message for the client (GF-2311).
            msg = msg.replace('\\t', '\t').replace('\\n', '\n')
            sys.stderr.write(msg + '\n')
            if hasattr(exc_value, "usage") and exc_value.usage:
                print(exc_value.usage)

        return False  # False = do not squelch. Propagate


class Server(object):

    """base class for Git Fusion servers."""

    def __init__(self):
        """Init the Server object."""
        self.p4 = None
        self.user = None
        self.foruser = None
        self.repo_name_git = None
        self.command = None
        self.git_caller = None
        self.skip_perms_check = False
        self.poll_only = False
        self.repo_name = None
        self.repo_perm = None
        self.git_dir = None
        self._repo_config = None
        self._should_remove_atomic_lock = False
        self._server_id = None

    def before(self):
        """override to do setup before process."""
        pass

    def after(self):
        """override to do cleanup after process."""
        pass

    def push_received(self):
        """Return True if the push payload is being received.

        Some protocols, such as HTTP, receive multiple requests during a
        push operation, only one of which is the actual push payload. We
        need to know when that is the case, and the server implementations
        must override this method to provide that information.

        """
        # pylint:disable=no-self-use
        return False

    def record_access(self):
        """Record the access of the repository in the audit log."""
        pass

    def record_error(self, msg):
        """Record the given error message to the audit log.

        :param str msg: the error message to record.

        """
        pass

    @property
    def server_id(self):
        """Return the server identifier for this instance."""
        if self._server_id is None:
            self._server_id = p4gf_util.has_server_id_or_exit(log=LOG)
        return self._server_id

    @property
    def repo_config(self):
        """Fetch or create the repo configuration file, if not already loaded."""
        if self._repo_config is None:
            self._repo_config = p4gf_config.RepoConfig.from_depot_file(
                self.repo_name, self.p4, create_if_missing=True)
        return self._repo_config

    @property
    def effective_user(self):
        """Return the name of the effective user (foruser, if set, else user)."""
        return self.foruser if self.foruser else self.user

    @with_timer('Setup')
    def _setup(self):
        """do setup; no lock required."""
        LOG.debug(p4gf_log.memory_usage())
        LOG.info("Begin process for repo=%s on server_id=%s READ_ONLY instance=%s",
                 os.environ[p4gf_const.P4GF_LOG_REPO_NAME], self.server_id, p4gf_const.READ_ONLY)

        p4gf_util.reset_git_enviro()
        #
        # Initialize the external process launcher early, before allocating
        # lots of memory, and just after all other conditions have been
        # checked.
        #
        # Do this _after_ changing to the git working tree, as it seems to
        # be rather difficult to correct this later using the 'cwd'
        # argument to the subprocess.Popen() constructor, at least on
        # Linux systems.
        #
        # Also do this before creating any P4 connections, as that seems to
        # effect whether temporary clients are automatically deleted or not.
        #
        p4gf_proc.init()
        if p4gf_const.READ_ONLY:   # no client required for READ_ONLY
            self.p4 = p4gf_create_p4.create_p4(warn_no_client=False)
        else:
            self.p4 = p4gf_create_p4.create_p4_temp_client()
        if not self.p4:
            raise PerforceConnectionFailed()
        LOG.debug("connected to P4: %s", self.p4)

        p4gf_group.invalidate_groups_i()
        p4gf_branch.init_case_handling(self.p4)
        self._get_repo_name_and_foruser()

        self._check_readiness()
        self._check_reviews_users_login()
        self._check_lock_perm()
        self._check_protects()
        self.check_special_command()

        self._check_valid_command()
        self._check_gf_depot_permissions()
        self.check_user_exists()
        self._check_perms()
        self._init_system()

        self.check_lfs_enabled()
        self._write_motd()

    def _init_system(self):
        """Initialize the Git Fusion system."""
        # Create Git Fusion server depot, user, config. NOPs if already created.
        if p4gf_const.READ_ONLY:
            # Global init may not be performed on a READ_ONLY instance
            if not p4gf_init.is_global_init_completed(self.p4):
                raise ReadOnlyInstanceException(
                        _("Cannot initialize Git Fusion with read-only instance."))
        else:
            p4gf_init.init(self.p4)

    @with_timer('Server process')
    def process(self):
        """Process the request."""
        exit_code = 1
        with ExitStack() as stack:
            stack.enter_context(ExceptionAuditLogger())
            stack.enter_context(p4gf_create_p4.Closer())
            stack.enter_context(run_before_after(self))
            p4gf_log.set_repo_name(self.repo_name_git)
            stack.enter_context(gc_debug())
            stack.enter_context(log_start_end(self))
            self._setup()

            ctx = p4gf_context.create_context(self.repo_name)
            ctx.p4gf = self.p4
            ctx.foruser = self.foruser

            def _cleanup_atomic_lock(exc_type, _exc_value, _exc_traceback):
                """Clear the atomic lock if there is an error."""
                # Sometimes the context and/or config are not initialized
                # yet, in which case there is nothing to be done here.
                if exc_type and ctx.p4 and ctx.repo_config:
                    LOG.warn('request processing failed, removing atomic lock')
                    self._should_remove_atomic_lock = True
                    self._maybe_remove_atomic_lock(ctx)

            # Add a cleanup handler in case the context experiences an
            # error while disconnecting, which can happen in the case of
            # failing to release the lock.
            if not p4gf_const.READ_ONLY:
                stack.push(_cleanup_atomic_lock)
            stack.enter_context(ctx)
            ctx.log_context()
            self._init_host(ctx)
            stack.enter_context(NoSuchRepoCleaner(self.repo_name))
            # N.B. this temporarily takes both exclusive locks, if needed
            # N.B. for the LFS server, this function does nothing
            self._init_repo(ctx)

            # Change into the git working directory. Not all git commands
            # react well to the --work-tree option.
            self.git_dir = ctx.repo_dirs.GIT_DIR
            os.chdir(ctx.repo_dirs.GIT_WORK_TREE)
            stack.enter_context(raise_on_sigterm())

            try:
                exit_code = self.process_request(ctx)
            except TerminatingException:
                # The design of the "raise on sigterm" handler is to raise
                # an exception when SIGTERM is received, which should cause
                # all of the "process" functions to clean up and remove any
                # locks. We then quietly exit to avoid causing grief with
                # the T4 tests running in ElectricCommander.
                LOG.warning('terminating signal received, exiting quietly')

        return exit_code

    def process_request(self, ctx):
        """Handle the incoming request, now that everything is set up.

        :param ctx: Git Fusion context.
        :return: exit status code (usually zero).

        """
        try:
            rollback_prl(ctx, blocking=False)
        except p4gf_git_repo_lock.LockBusy:
            # This will happen often, do not fail fast, just log and move on.
            LOG.info("repo %s busy at this time, no rollback attempted", self.repo_name)

        if CMD_GIT_UPLOAD_PACK in self.command or self.poll_only:
            # This is a pull request.
            update_only_on_poll = self.repo_config.getboolean(
                p4gf_config.SECTION_PERFORCE_TO_GIT,
                p4gf_config.KEY_UPDATE_ONLY_ON_POLL)
            if update_only_on_poll and not self.poll_only:
                # The antithesis of polling: no update at all, serve
                # whatever is available in the repo right now.
                exit_code = self._call_git(ctx)
            else:
                # Normal fetch or a poll.
                exit_code = self._process_upload(ctx)
        else:
            # This is a push request.
            if p4gf_const.READ_ONLY:
                raise ReadOnlyInstanceException(_('Push to read-only instance prohibited'))
            exit_code = self._process_receive(ctx)

        return exit_code

    def _process_upload(self, ctx):
        """Service the git-upload-pack (fetch) request.

        This handles both the read-only and "writable" modes. There is, in
        fact, nothing special to be done here for read-only mode, as that
        is largely handled in the p4-to-git translation code.

        :param ctx: Git Fusion context.
        :return: exit code of the git command, or 0 if poll_only is True.

        """
        try:
            LOG.debug("_process_upload: p2g commencing for %s", self.repo_name)
            ctx.checkpoint("server_common:acquire_lock")
            # Acquire exclusive access so we can perform p4-to-git and
            # object mirroring safely and ensure data consistency. Avoid
            # blocking since that likely means that another process is
            # already updating the repo.
            with ExitStack() as stack:
                if p4gf_const.READ_ONLY:
                    LOG.info("_process_upload: READ_ONLY : taking only git write_lock for %s",
                             self.repo_name)
                    stack.enter_context(p4gf_git_repo_lock.write_lock(self.repo_name))
                else:
                    LOG.info("_process_upload:  taking repo lock and git write_lock for %s",
                             self.repo_name)
                    stack.enter_context(Timer('with Lock'))
                    stack.enter_context(ctx.exclusive_access(
                        blocking=False, username=self.effective_user))
                self._copy_p2g(ctx)
                # Clean up client pool while we have the write lock.
                ctx.client_pool_cleanup()
        except p4gf_lock.LockBusy:
            # Oh well, this fetch will likely be missing some recent
            # changes. At least we can respond quickly, without waiting on
            # other processes.
            LOG.info("skipping p2g due to contention on %s", self.repo_name)
        except p4gf_imports.ReadOnlyException as e:
            raise ReadOnlyInstanceException(str(e))
        except p4gf_init_repo.InitRepoReadOnly as e:
            raise ReadOnlyInstanceException(str(e))
        finally:
            ctx.checkpoint("server_common:releasing_lock")
            ctx.disconnect()
        if self.poll_only:
            # Record the event of accessing the repo for the audit log.
            self.record_access()
            code = os.EX_OK
        else:
            LOG.debug("delegating to git for %s", self.repo_name)
            code = self._call_git(ctx)
        return code

    def _process_receive(self, ctx):
        """Service the git-receive-pack (push) request.

        :param ctx: Git Fusion context.
        :return: exit code of the git command.

        """
        def cleanup(repo_lock, msg):
            """Clean up after error."""
            # Do not wait to release the lock, just do it immediately. We
            # cannot make any assumptions about what git-receive-pack and
            # our hooks are doing right now, we can only exit as quickly
            # and cleanly as possible.
            LOG.debug2('_process_receive.cleanup() %s', msg)
            if repo_lock:
                repo_lock.ignore_pending_acquire = True
            # When something bad happens, remove the atomic lock.
            self._should_remove_atomic_lock = True
            self._maybe_remove_atomic_lock(ctx)
            # Record the failure in the push status key.
            ctx.record_push_failed_p4key(msg)

        try:
            with ExitStack() as stack:
                # Get both write locks, waiting as long as necessary.
                stack.enter_context(ctx.exclusive_access(username=self.effective_user))
                self._should_remove_atomic_lock = True
                # We now have the exclusive write lock and the p4key lock and
                # can perform the p4-to-git and git-to-p4 translations.
                stack.enter_context(Timer('with Lock'))
                ctx.checkpoint("server_common:acquire_lock")
                # If p4gf_config is configured for non-ff push,
                # ensure the git repo is so configured
                if ctx.lw_non_ff_enable or ctx.fp_non_ff_enable:
                    p4gf_git.set_denynonfastforwards(ctx.repo, False)
                # Save as a Key the current last_copied change prior to this copy_p2g.
                # If this is a non-FF FP push, and is so configured, we still
                # will reject if there are untranslated changes in P4.
                # This key will be used in the pre-receive hook.
                _push_id = self._increment_push_counter(ctx)
                pre_p2g_key_name = P4Key.calc_repo_pre_p2g_last_copied_p4key_name(
                    self.repo_name, _push_id)
                P4Key.set(ctx.p4gf, pre_p2g_key_name, ctx.read_last_copied_change())

                # First _copy_p2g without the atomic locks - so most of the work
                # is performed without locking out P4 users.

                self._copy_p2g(ctx)
                # Second _copy_p2g with the atomic locks - will often copy
                # nothing. The atomic lock ensures that no other changes
                # come in during this next p4-to-git or the push
                p4gf_atomic_lock.lock_update_repo_reviews(ctx, action=p4gf_atomic_lock.ADD)
                self._copy_p2g(ctx)
                # Ensure the atomic lock is removed if the p4key lock is
                # released without successfully transferring it to the
                # background process (hence the two callbacks).
                release_cb = functools.partial(self._maybe_remove_atomic_lock, ctx)
                ctx.repo_lock.set_lock_release_cb(release_cb)
                ctx.repo_lock.set_transfer_complete_cb(self._lock_transfer_cb)
                # The git-to-p4 translation will happen in post-receive.
                _set_pre_receive_flag(ctx)
                stack.enter_context(Timer('call git'))
                ec = self._call_git(ctx)
                if ec:
                    msg = _('git-receive-pack returned {error_code}').format(error_code=ec)
                    LOG.error('{} for {}'.format(msg, self.repo_name))
                    cleanup(ctx.repo_lock, msg)
                    sys.stderr.write(msg + "\n")
                elif _detect_pre_receive_flag(ctx):
                    # Treat the "nop" as a success, but record in the log exactly
                    # what happened for easier debugging when a push seems to fail.
                    LOG.info('push {} had nothing to do (pre-receive not invoked)'.format(
                        ctx.push_id))
                    ctx.record_push_success_p4key()
                ctx.checkpoint("server_common:releasing_lock")
                # Do not clutter P4 with this one-time use push associated Key.
                # It was used only by pre-receive-hook for rejecting some FP nonFF pushes.
                if  pre_p2g_key_name:
                    P4Key.delete(ctx.p4gf, pre_p2g_key_name)
                return ec
        except:  # pylint:disable=bare-except
            LOG.exception('receive failed')
            cleanup(ctx.repo_lock, sys.exc_info()[1])
            raise

    def _get_repo_name_and_foruser(self):
        """Extract foruser from url if present; get translated repo name."""
        foruser_patt = re.compile('@foruser=([^ @]+)')
        m = foruser_patt.search(self.repo_name_git)
        if m:
            self.foruser = m.group(1)
            repo_name_git = self.repo_name_git[:m.start(0)] + self.repo_name_git[m.end(0):]
            LOG.debug('foruser: %s', self.foruser)
        else:
            repo_name_git = self.repo_name_git
        # translate '/' ':' ' ' .. etc .. for internal repo_name
        # and figure out if repo is  'repo' or 'repo.git'
        self.repo_name = p4gf_translate.TranslateReponame.url_to_repo(repo_name_git, self.p4)
        p4gf_log.configure_for_repo(self.repo_name)
        LOG.debug("public repo_name: %s   internal repo_name: %s",
                  repo_name_git, self.repo_name)

    def _check_readiness(self):
        """Check that P4GF is ready for accepting connections from clients."""
        prevent_sessions = P4Key.get_all(self.p4, p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS + '*')
        trigger_version = P4Key.get(self.p4, p4gf_const.P4GF_P4KEY_TRIGGER_VERSION)

        # Check if the "prevent further access" p4key has been set, and raise an
        # error if the p4key is anything other than zero.
        if prevent_sessions and \
                not p4gf_group.user_in_group(self.p4, self.user, p4gf_const.P4GF_GROUP_ADMIN):
            every_instance = prevent_sessions.get(p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_ALL, '0')
            key_name = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_INSTANCE + self.server_id
            our_instance = prevent_sessions.get(key_name, '0')
            if every_instance != '0' or our_instance != '0':
                raise RuntimeError(_('Git Fusion is shutting down. Please contact your admin.'))

            # cannot test invalid repo names, such as special commands
            contains_of_evil = [c for c in self.repo_name if c in ['@', '#', '$', '*']]
            if not contains_of_evil:
                # Check whether this repo is prevented on all instances
                key_name = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_REPO + self.repo_name
                our_repo = prevent_sessions.get(key_name, '0')
                if our_repo != '0':
                    emsg = _("Access to repo '{}' is temporarily blocked by Git Fusion. "
                             "Please contact your admin.").format(self.repo_name)
                    raise RuntimeError(emsg)

                # Check if this repo is prevented on this instance
                key_name = p4gf_const.P4GF_P4KEY_PREVENT_SESSIONS_INSTANCE_REPO.format(
                    server_id=self.server_id, repo_name=self.repo_name)
                instance_repo_key = prevent_sessions.get(key_name, '0')
                if instance_repo_key != '0':
                    emsg = _("Access to repo '{}' on '{}' is temporarily blocked by Git Fusion. "
                             "Please contact your admin.").format(self.repo_name, self.server_id)
                    raise RuntimeError(emsg)

        # Check that GF submit trigger is installed and has a compatible version.
        trigger_version_p4key = trigger_version.split(":")[0].strip() if trigger_version else '0'
        if trigger_version_p4key != p4gf_const.P4GF_TRIGGER_VERSION:
            LOG.error("Incompatible trigger version: {0} should be {1} but got {2}".format(
                p4gf_const.P4GF_P4KEY_TRIGGER_VERSION,
                p4gf_const.P4GF_TRIGGER_VERSION, trigger_version_p4key))
            if trigger_version_p4key == '0':
                raise RuntimeError(_('Git Fusion submit triggers are not installed.'
                                     ' Please contact your admin.'))
            else:
                raise RuntimeError(_('Git Fusion submit triggers need updating.'
                                     ' Please contact your admin.'))

    def _raise_p4gf_perm(self):
        """User-visible permission failure."""
        p4gf_util.raise_gfuser_insufficient_perm(p4port=self.p4.port)

    def _check_lock_perm(self):
        """Permission check: see if git-fusion-user has adequate permissions to use locks."""
        # try deleting a p4 key to see if we can.  If so, we have enough rights
        # to be able to use locks.  P4 checks perms before checking if the key
        # actually exists, so it doesn't need to be a key that actually exists.
        try:
            P4Key.delete(self.p4, p4gf_const.P4GF_P4KEY_PERM_CHECK)
        except P4Exception:
            # expect a protect error if we don't have access to use keys
            if p4gf_p4msg.contains_protect_error(self.p4):
                self._raise_p4gf_perm()
                return
            if p4gf_p4msg.find_msgid(self.p4, p4gf_p4msgid.MsgDm_NoSuchKey):
                return
            raise

    def _check_reviews_users_login(self):
        """Check the reviews users have logins."""
        # For testing purposes skip this test if so configured.
        if p4gf_const.P4GF_TEST_CHECK_REVIEWS_USERS_LOGIN in os.environ:
            return

        for user in [p4gf_util.gf_reviews_user_name(), p4gf_const.P4GF_REVIEWS__ALL_GF]:
            (result, msg) = p4_has_login(user, self.p4.port, self.p4.client)
            LOG.debug("_check_reviews_users_login user={} {} {}".format(user, result, msg))
            if not result:
                msg = _("Perforce: Git Fusion user='{0}' not logged in.\n{1}".
                        format(user, msg))
                raise RuntimeError(msg)

    def _check_protects(self):
        """Check that the protects table does not deny the Git Fusion user.

        It must either be empty or grant the Git Fusion user sufficient privileges.
        Return False if this is not the case.
        """
        if not p4gf_version_3.p4d_supports_protects(self.p4):
            self._raise_p4gf_perm()

    def check_lfs_enabled(self):
        """Validate repo configuration if processing an LFS request.

        If we're processing a Git LFS request, but the current repo is not
        configured to allow Git LFS requests, reject.

        Cannot check until after you load the repo config.
        """
        pass

    def _check_gf_depot_permissions(self):
        """Verify P4GF_USER has admin access to //P4GF_DEPOT/...

        Fetch the admin permisions and raise perm exception
        if any of the test files are not mapped.
        The set of tested files are those required the given repo_name
        """
        # init admin filter
        gf_client_map = Map()
        gf_client_map.insert("//{0}/...".format(p4gf_const.P4GF_DEPOT), "//client/...")
        gf_admin_filter = p4gf_protect.UserToProtect(self.p4).user_to_protect(
            p4gf_const.P4GF_USER).map_for_perm(p4gf_protect.ADMIN)
        gf_admin_filter = Map.join(gf_admin_filter, gf_client_map)

        # Exhaustive list of RW paths required by GF for a given repo
        files_to_test = ["//{0}/branch_info/foo",
                         "//{0}/branches/{1}/foo",
                         "//{0}/objects/repos/{1}/foo",
                         "//{0}/objects/trees/foo",
                         "//{0}/p4gf_config",
                         "//{0}/repos/{1}/p4gf_config",
                         "//{0}/repos/{1}/p4gf_config2",
                         "//{0}/users/p4gf_usermap"
                         ]
        for f in files_to_test:
            f = f.format(p4gf_const.P4GF_DEPOT,  # pylint: disable=too-many-format-args
                         self.repo_name)
            if not gf_admin_filter.includes(f):
                LOG.error("check_gf_depot_permissions FAILED test {0}".format(f))
                self._raise_p4gf_perm()

    def check_special_command(self):
        """If repo is actually a special command, run it and raise SpecialCommandException."""
        if not self.repo_name.startswith('@'):
            return
        SpecialCommandHandler.create(self).run()
        raise SpecialCommandException()

    def _check_valid_command(self):
        """Verify requested command is valid."""
        if self.command not in COMMAND_TO_PERM:
            LOG.debug("command %s not in %s", self.command, COMMAND_TO_PERM)
            raise BadRequestException(_("Unrecognized service\n"))

    def _check_user_exists(self, user):
        """Check that the user actually exists."""
        if not p4gf_p4spec.spec_exists(self.p4, 'user', user):
            form = _('User {user} does not exist in Perforce. Please contact your admin.')
            msg = form.format(user=user)
            self.record_error(msg)
            raise RuntimeError(msg)
        if not p4gf_p4spec.spec_values_match(self.p4, 'user', user, {'Type': 'standard'}):
            form = _("Perforce: User '{user}' has invalid 'Type'."
                     " User Type must be 'standard'. Please contact your admin.")
            msg = form.format(user=user)
            self.record_error(msg)
            raise RuntimeError(msg)

    def check_user_exists(self):
        """Check that the user(s) actually exist."""
        self._check_user_exists(self.user)
        if self.foruser:
            self._check_user_exists(self.foruser)

    def _raise_perms_error(self):
        """Raise an exception for insufficient permissions."""
        msg = _("User '{user}' not authorized for '{command}' on '{repo}'.").format(
            user=self.repo_perm.p4user_name, command=self.command, repo=self.repo_name)
        # if user permissions prevent the pull provide verbose message.
        if self.repo_perm.user_read_permission_checked and self.repo_perm.error_msg:
            msg += self.repo_perm.error_msg
        self.record_error(msg)
        raise CommandError(msg)

    def _check_perms(self):
        """Check that user has permission to run the command.

        If not, raise an exception.
        We use the translated internal repo name here for perm authorization
        """
        if self.skip_perms_check:
            return
        required_perm = COMMAND_TO_PERM[self.command]
        # first, check foruser if set.  This will leave self.repo_perm set
        # to RepoPerm for foruser, so any error can be reported
        if self.foruser:
            LOG.debug("_check_perms for {}".format(self.foruser))
            if not self.check_permissions(required_perm, user=self.foruser):
                self._raise_perms_error()
        # if that worked, check for authenticated user, resetting self.repo_perm
        if not self.check_permissions(required_perm):
            self._raise_perms_error()

    def check_permissions(self, required_perm, repo_name=None, user=None):
        """Check that user has permission to run the command.

        We use the translated internal repo name here for perm authorization.

        :type required_perm: str
        :param required_perm: either p4gf_group.PERM_PULL or p4gf_group.PERM_PUSH

        :type repo_name: str
        :param repo_name: name of repository, or None to use default.

        :rtype: bool
        :return: True if access permitted, False otherwise.

        """
        user = user or self.user
        if repo_name is None:
            repo_name = self.repo_name
        self.repo_perm = p4gf_group.RepoPerm.for_user_and_repo(
            self.p4, user, repo_name, required_perm)
        LOG.debug2('check_permissions() require %s perm %s', required_perm, self.repo_perm)
        if self.repo_perm.can(required_perm):  # check group permissions
            if required_perm == p4gf_group.PERM_PULL:
                # if group grants permissions - then check for user read perms
                if p4gf_read_permission.user_has_read_permissions(
                        self.p4, self.repo_perm, required_perm):
                    LOG.debug2('check_permissions() read accept perm %s', self.repo_perm)
                    return True
            else:  # PERM_PUSH
                LOG.debug2('check_permissions() push accept perm %s', self.repo_perm)
                return True
        if p4gf_read_permission.can_create_depot_repo(self.p4, repo_name):
            LOG.debug2('check_permissions() create accept perm %s', self.repo_perm)
            return True
        LOG.debug2('check_permissions() reject perm %s', self.repo_perm)
        return False

    def _write_motd(self):
        """If there is a .git-fusion/motd.txt file, return it on stderr."""
        motd = self._read_motd()
        if motd:
            sys.stderr.write(motd)

    @staticmethod
    def _read_motd():
        """If there is a message of the day file, return its contents.

        If not, return None.
        """
        p4gf_dir = p4gf_const.P4GF_HOME
        motd_file_path = p4gf_const.P4GF_MOTD_FILE.format(P4GF_DIR=p4gf_dir)
        if not os.path.exists(motd_file_path):
            return None
        with open(motd_file_path, 'r') as f:
            content = f.read()
        return content

    def _init_repo(self, ctx):
        """Create Git Fusion per-repo client view mapping and config.

        :return: True if repo created, False otherwise.

        """
        LOG.debug("ensuring repo {} is initialized".format(self.repo_name))
        repo_initer = p4gf_init_repo.InitRepo(self.p4, None).set_repo_name(self.repo_name)
        if repo_initer.is_init_needed():
            # Set the context first, and then the config, using the init
            # repo setters so that they are coordinated.
            repo_initer.context = ctx
            # Lazily fetch/create the repo configuration.
            repo_initer.set_repo_config(self.repo_config)
            with ExitStack() as stack:
                # Temporarily ignore SIGTERM while we take out several locks.
                stack.enter_context(ignore_sigterm())
                LOG.info("_init_repo() initializing repo {}".format(self.repo_name))
                # Hold all the locks while initializing this repository and
                # its configuration, as well as possibly modifying the
                # group membership.
                stack.enter_context(p4gf_git_repo_lock.write_lock(
                    self.repo_name, username=self.effective_user))
                # If in read-only mode, do not block trying to acquire the
                # p4key lock and exit immediately with an error message.
                blocking = not p4gf_const.READ_ONLY
                repo_lock = p4gf_lock.RepoLock(self.p4, self.repo_name, blocking=blocking,
                                               username=self.effective_user)
                try:
                    stack.enter_context(repo_lock)
                except p4gf_lock.LockBusy:
                    raise ReadOnlyInstanceException(_("Repo currently busy, try again later."))
                stack.enter_context(Timer('with Lock'))
                try:
                    repo_initer.repo_lock = repo_lock
                    # Do not permit initialization from scratch during
                    # pull/push when in read-only instance mode.
                    repo_initer.set_fail_when_read_only(True)
                    repo_created = repo_initer.init_repo()
                except p4gf_init_repo.InitRepoMissingView as e:
                    LOG.debug('missing view for repo, %s', self.repo_name)
                    raise RepoNotFoundException(str(e))
                except p4gf_init_repo.InitRepoMissingConfigFile as e:
                    LOG.debug('missing config file for repo, %s', self.repo_name)
                    raise RepoNotFoundException(str(e))
                except RuntimeError as e:
                    LOG.exception('repo initialization failed')
                    raise RepoInitFailedException(str(e))
                # If authorization came from default, not explicit group
                # membership, copy that authorization to a group now. Could
                # not do this until after p4gf_init_repo() has a chance to
                # create not-yet-existing groups.
                if self.repo_perm:
                    self.repo_perm.write_if(self.p4)
                return repo_created
        if not p4gf_const.READ_ONLY and self.repo_perm and self.repo_perm.needs_write():
            # Not convinced we need the lock for modifying the group,
            # but since this will be rare, it is an acceptable cost.
            with p4gf_lock.RepoLock(self.p4, self.repo_name):
                self.repo_perm.write_if(self.p4)
        return False

    def _init_host(self, ctx):
        """Initialize the Git Fusion host for this repo, if needed."""
        if p4gf_init_host.is_init_needed(ctx.repo_dirs):
            with p4gf_git_repo_lock.write_lock(self.repo_name, username=self.effective_user):
                p4gf_init_host.init_host(ctx.repo_dirs, ctx)

    @with_timer('copy to Git')
    def _copy_p2g(self, ctx):
        """Copy any recent changes from Perforce to Git."""
        try:
            # While we have the write lock, make sure the repo is marked as
            # bare, since that is how it should be by default anyway.
            p4gf_git.set_bare(True)

            # Recover from a failed mirror operation, if needed, before
            # going any further and attempting p4-to-git translation.
            if not p4gf_const.READ_ONLY:
                # Read-only does not take a lock and does not do mirroring,
                # so nothing to be done here.
                p4gf_gitmirror.recover_failed_mirror(ctx)

            # Since we fetch the configuration before we acquire the
            # exclusive lock, there is a chance that another pull/push has
            # modified/created that configuration. Now that we have the
            # exclusive lock, preventing submit to config/config2, refresh
            # if a new file revision for config or config2 has been
            # submitted since we originally loaded it, but before we
            # acquired the lock.
            self.repo_config.refresh_if(self.p4, create_if_missing=True)
            ctx.repo_config = self.repo_config

            stream_imports = p4gf_imports.StreamImports(ctx)
            if stream_imports.missing_submodule_import_url():
                raise MissingSubmoduleImportUrlException()

            p4gf_copy_p2g.copy_p2g_ctx(ctx)

            if not p4gf_const.READ_ONLY:
                stream_imports.process()

            # Now is also an appropriate time to clear out any stale Git
            # Swarm reviews. We're pre-pull, pre-push, time when we've
            # got exclusive write access to the Git repo,
            GSReviewCollection.delete_refs_for_closed_reviews(ctx)

        except:
            # Dump failure to log, BEFORE cleanup, just in case
            # cleanup ALSO fails and throws its own error (which
            # happens if we're out of memory).
            LOG.exception('copy p2g failed')
            p4gf_lock_status.print_lock_status(ctx.p4gf, self.server_id, LOG.error)
            raise

    def _increment_push_counter(self, ctx):
        """Increment the push counter value now that the push is underway."""
        # Increment if we are getting the actual push payload.
        if self.push_received():
            id_key_name = P4Key.calc_repo_push_id_p4key_name(self.repo_name)
            push_id = P4Key.increment(ctx.p4gf, id_key_name)
            msg = _("Push {push_id} started").format(push_id=push_id)
            ctx.record_push_status_p4key(msg)
            return push_id
        return None

    def _lock_transfer_cb(self):
        """Lock ownership was transferred successfully."""
        # Lock was successfully transferred to the background process, do
        # _not_ release the atomic lock in this "foreground" process.
        LOG.debug2('canceling repo lock removal')
        self._should_remove_atomic_lock = False

    def _maybe_remove_atomic_lock(self, ctx):
        """Remove the atomic lock for the repo."""
        LOG.debug2('_maybe_remove_atomic_lock() remove? %s', self._should_remove_atomic_lock)
        if self._should_remove_atomic_lock:
            LOG.debug('removing atomic-lock for %s', self.repo_name)
            # In the event of a preflight error, it is highly likely the
            # connection has been dropped, so temporarily reconnect.
            with ExitStack() as stack:
                if not ctx.p4gf.connected():
                    p4gf_create_p4.p4_connect(ctx.p4gf)
                    stack.callback(p4gf_create_p4.p4_disconnect, ctx.p4gf)
                p4gf_atomic_lock.lock_update_repo_reviews(ctx, action=p4gf_atomic_lock.REMOVE)

    def _call_git(self, ctx):
        """Delegate to the appropriate Git command defined in git_caller.

        Call git (e.g. git-upload-pack, git-receive-pack) while keeping reviews updated.

        Returns the exit code of the Git command.

        """
        # Record the event of transferring information (i.e. audit log).
        self.record_access()

        retval = None

        # Detach git repo's HEAD before calling original git, otherwise we
        # won't be able to push the current branch (if any).
        if not p4gf_git.is_bare_git_repo():
            p4gf_git.checkout_detached_head()

        # Flush stderr before returning control to Git. Otherwise Git's own
        # output might interrupt ours.
        sys.stderr.flush()

        # Keep the idle P4 connection open during call to git since we have
        # to keep the temporary client (and likewise the lock) alive.
        LOG.info('call git: delegating to git for {}'.format(ctx.config.repo_name))

        # Ignore the SIGTERM signal during this particularly delicate stage
        # in the process. Allowing this signal to get through to Git could
        # possibly result in the pre-receive hook running but not the post-
        # receive hook, which may leave the git references ahead of the
        # last translated commit. This is very difficult to detect and
        # prevent after the fact. For instance, HTTP push seemingly always
        # ends up with Git references that are ahead of Perforce changes.
        # The same happens with a push that introduces a change rejected by
        # a trigger.
        with ignore_sigterm():
            retval = self.git_caller(ctx)  # pylint:disable=not-callable

        LOG.info('call git: returning {}'.format(retval))
        return retval


# functions and log messages for tag or branch rollback in _rollback_prt().
RollbackHow = namedtuple("RollbackHow", [
    "log_nop",
    "log_force",
    "func_force",
    "log_delete",
    "func_delete"
])
_ROLLBACK_BRANCH = RollbackHow(
    "git branch no change required",
    "git branch -f {ref:<20} {old_sha1}",
    p4gf_git.force_branch_ref,
    "git branch -D {ref:<20}",
    p4gf_git.delete_branch_ref
)
_ROLLBACK_TAG = RollbackHow(
    "git tag no change required",
    "git tag -f {ref:<20} {old_sha1}",
    p4gf_git.force_tag_ref,
    "git tag -d {ref:<20}",
    p4gf_git.delete_tag_ref
)


def _rollback_prt(prt, how):
    """Roll one tag or branch ref back, with logging."""
    curr_sha1 = p4gf_util.git_rev_list_1(prt.ref)
    if curr_sha1 is None:
        curr_sha1 = p4gf_const.NULL_COMMIT_SHA1
    prefix = "rollback: prt={prt:<60} curr={curr_sha1} "\
             .format(prt=prt, curr_sha1=p4gf_util.abbrev(curr_sha1))
    # Surely we have a "strip /refs/xxx/ prefix" utility function somewhere?
    short_ref = prt.ref
    for p in ["refs/heads/", "refs/tags/"]:
        if short_ref.startswith(p):
            short_ref = short_ref[len(p):]
            break
    if curr_sha1 == prt.old_sha1:
        LOG.warning(prefix + how.log_nop)
    elif prt.old_sha1 == p4gf_const.NULL_COMMIT_SHA1:
        LOG.warning(prefix + how.log_delete.format(ref=short_ref))
        how.func_delete(short_ref)
    else:
        LOG.warning(prefix + how.log_force.format(ref=short_ref, old_sha1=prt.old_sha1))
        how.func_force(short_ref, prt.old_sha1)


def rollback_prl(ctx, blocking):
    """If a prl file exists, then a previous push has failed. Git might have
    refs that point to history not yet copied to Perforce. Roll those
    refs back to where they were before the previous, failed, push.

    NOP if no prl file: post-receive took over and (should have) finished
    the translation to Perforce.

    :param blocking: if True, will block until performing rollback,
                      if false, raise LockBusy.

    Returns with Git repo checked out to master-ish branch (if any),
    since rollback_prl() has to detach head to change any branch refs.
    """
    prlfile = PRLFile(ctx.config.repo_name)
    prl = prlfile.read()
    if not prl:
        return
    with p4gf_git_repo_lock.write_lock(ctx.config.repo_name, blocking=blocking):
        # When we can get the lock, then log as an error, otherwise it
        # means we had lock contention, which is not an error.
        LOG.warning("rollback: checking Git refs after previous push failed: {}"
                    .format(ctx.config.repo_name))
        # Re-fetch prl AFTER we hold the lock. Can't trust pre-lock data
        # that some other process might have changed out from under us.
        prl = prlfile.read()
        if not prl:
            return
        if not p4gf_git.is_bare_git_repo():
            p4gf_git.checkout_detached_head()
        for prt in prl.heads():
            _rollback_prt(prt, _ROLLBACK_BRANCH)
        for prt in prl.tags():
            _rollback_prt(prt, _ROLLBACK_TAG)
        prlfile.delete()
        if not p4gf_git.is_bare_git_repo():
            ctx.checkout_master_ish()


@contextmanager
def run_before_after(server):
    """Wrap something with calls to before() and after()."""
    server.before()
    try:
        yield
    finally:
        server.after()


@contextmanager
def log_start_end(server):
    """Log the start and end of the processing of a request."""
    LOG.info("process-start of {} for '{}' on repo '{}' on server='{}' READ_ONLY={}".format(
        server.command, server.user, server.repo_name_git,
        server.server_id, p4gf_const.READ_ONLY))
    try:
        yield
    finally:
        LOG.info("process-end of {} for '{}' on repo '{}' on server='{}' READ_ONLY={}".format(
            server.command, server.user, server.repo_name_git,
            server.server_id, p4gf_const.READ_ONLY))


@contextmanager
def gc_debug():
    """Yield to caller with garbage collection debugging, if enabled."""
    p4gf_mem_gc.init_gc()
    try:
        yield
    finally:
        p4gf_mem_gc.process_garbage(NTR('at end of server process'))


@contextmanager
def raise_on_sigterm():
    """Raise an exception if the SIGTERM signal is received."""
    def raise_exception(_signum, _frame):
        """Raise an exception."""
        LOG.error("SIGTERM received. Raising exception.")
        # Avoid receiving another SIGTERM again while we are trying to exit
        # cleanly already, so ignore any further SIGTERM signals.
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        raise TerminatingException('SIGTERM received')
    LOG.debug("raise_on_sigterm() will raise exception on SIGTERM")
    term_handler = signal.signal(signal.SIGTERM, raise_exception)
    try:
        yield
    finally:
        if term_handler:
            signal.signal(signal.SIGTERM, term_handler)
        else:
            signal.signal(signal.SIGTERM, signal.SIG_DFL)


@contextmanager
def ignore_sigterm():
    """Temporarily ignore the SIGTERM signal."""
    if 'REMOTE_ADDR' in os.environ:
        LOG.debug("ignore_sigterm() not ignoring SIGTERM because http")
        # For HTTP, secretly do nothing as we really need to exit when
        # Apache sends us a SIGTERM, lest it employs SIGKILL after we fail
        # to politely shut down in a timely manner.
        yield
    else:
        LOG.debug("ignore_sigterm() ignoring SIGTERM signal")
        term_handler = signal.signal(signal.SIGTERM, signal.SIG_IGN)
        try:
            yield
        finally:
            if term_handler:
                signal.signal(signal.SIGTERM, term_handler)
            else:
                signal.signal(signal.SIGTERM, signal.SIG_DFL)


class NoSuchRepoCleaner:
    """Clean up after an attempt to clone a non-existent repo."""

    def __init__(self, repo_name):
        """Initialize an instance of NoSuchRepoCleaner."""
        self.repo_name = repo_name

    def __enter__(self):
        """Enter the context."""
        return None

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Based on the exception, clean up or not."""
        if exc_type and issubclass(exc_type, RepoNotFoundException):
            # Makes a huge assumption that nothing else is trying to lock
            # this same repository. But since we're called only when a pull
            # is made on a non-existent repository, that seems unlikely.
            locks_dir = "{0}/locks/{1}".format(p4gf_const.P4GF_HOME, self.repo_name)
            p4gf_util.remove_tree(locks_dir, contents_only=False)
            # Could remove the containers directory tree here, too, but the
            # init_repo code already handles that.
        return False


class StatusReporter:

    """Reports current repo status occassionally."""

    def __init__(self, p4, repo_name, wait_push_id, active_push_id):
        """Construct a new instance with the given wait time."""
        self.last_time = time.time()
        self.p4 = p4
        self.repo_name = repo_name
        self.wait_push_id = wait_push_id
        self.active_push_id = active_push_id
        self.last_status = None

    def maybe_report(self):
        """If it has been a while, report the current status."""
        new_time = time.time()
        # check for updated status every 5 seconds
        if new_time - self.last_time > 5:
            self.last_time = new_time
            msg = _status_for_push(self.p4, self.repo_name, self.wait_push_id, self.active_push_id)
            if msg != self.last_status:
                self.last_status = msg
                sys.stderr.write("{}\n".format(msg))
                sys.stderr.flush()


class SpecialCommandHandler:

    """Base class for handlers of special commands."""

    def __init__(self, server):
        """Init the handler."""
        self.server = server
        self.repo_name = server.repo_name
        self.p4 = server.p4
        self.user = server.user
        self.foruser = server.foruser

    @staticmethod
    def create(server):
        """Factory method for SpecialCommandHandler.

        :param server: instance of Server.
        :return: instance of SpecialCommandHandler.

        """
        special_names = [
            (p4gf_const.P4GF_UNREPO_HELP,     SpecialCommandHelp),
            (p4gf_const.P4GF_UNREPO_INFO,     SpecialCommandInfo),
            (p4gf_const.P4GF_UNREPO_LIST,     SpecialCommandList),
            (p4gf_const.P4GF_UNREPO_FEATURES, SpecialCommandFeatures),
            (p4gf_const.P4GF_UNREPO_PS,       SpecialCommandProcesses),
            (NTR("@mirror_wait"),             SpecialCommandMirrorWait)  # deprecated special
        ]
        command = server.repo_name
        for (s, f) in special_names:
            if s == command:
                return f(server)

        special_patterns = [
            (p4gf_const.P4GF_UNREPO_WAIT,     SpecialCommandWaitRepo),
            (p4gf_const.P4GF_UNREPO_PROGRESS, SpecialCommandProgressRepo),
            (p4gf_const.P4GF_UNREPO_FEATURES, SpecialCommandFeatureRepo),
            (p4gf_const.P4GF_UNREPO_STATUS,   SpecialCommandStatusRepo),
            (p4gf_const.P4GF_UNREPO_CONFIG,   SpecialCommandConfigRepo),
            (p4gf_const.P4GF_UNREPO_CL,       SpecialCommandChangelist)
        ]
        for (s, f) in special_patterns:
            if re.compile("^" + s + "@").match(command):
                repo_name = command[len(s)+1:]
                try:
                    repo_name, wait_id, active_id, timeout = \
                        _parse_push_identifiers(server.p4, repo_name)
                except RuntimeError as e:
                    # use the normal process to report the error
                    return SpecialCommandError(server, s, str(e))
                return f(server, repo_name, wait_id, active_id, timeout)

        # Did not match any known command, fall back to listing supported commands.
        return SpecialCommandHandler(server)

    def run(self):
        """Print the list of special commands."""
        # pylint:disable=no-self-use
        special_cmds = " ".join(p4gf_const.P4GF_UNREPO)
        sys.stderr.write(
            _("Git Fusion: unrecognized special command.\n"
              "Valid commands are: {commands}\n")
            .format(commands=special_cmds))


class SpecialCommandError(SpecialCommandHandler):

    """Reports an error in parsing the special command."""

    def __init__(self, server, command, error_msg):
        """Init the handler."""
        SpecialCommandHandler.__init__(self, server)
        self.command = command
        self.error_msg = error_msg

    def run(self):
        """Print the error message and write an error to the log."""
        LOG.warning("{} unable to proceed: {}".format(self.command, self.error_msg))
        sys.stderr.write(self.error_msg + "\n")


class SpecialCommandHelp(SpecialCommandHandler):

    """Processes the '@help' special command."""

    def run(self):
        """Dump the contents of the help.txt file, if it exists."""
        # pylint:disable=no-self-use
        help_text = p4gf_util.read_bin_file('help.txt')
        if help_text is False:
            sys.stderr.write(_("file 'help.txt' not found\n"))
        else:
            sys.stderr.write(help_text)


class SpecialCommandInfo(SpecialCommandHandler):

    """Processes the '@info' special command."""

    def run(self):
        """Print version information to stderr."""
        sys.stderr.write(p4gf_version_3.as_string_extended(p4=self.p4, include_checksum=True))


class SpecialCommandList(SpecialCommandHandler):

    """Processes the '@list' special command."""

    def run(self):
        """Print a list of known repositories to stderr."""
        def _merge_perm(p1, p2):
            """Return the lower permission of p1 and p2."""
            if p1 == NTR('pull') or p2 == NTR('pull'):
                return NTR('pull')
            return NTR('push')

        def _merge_lists(rl1, rl2):
            """merge two repo lists.

            Result contains only repos that are in both lists.
            perm for each repo in result is the lower of that from the two lists.
            """
            result = []
            for r1 in rl1:
                for r2 in rl2:
                    if r1[0] != r2[0]:
                        continue
                    result.append([
                        r1[0],
                        _merge_perm(r1[1], r2[1]),
                        r1[2],
                        r1[3]
                    ])
            result.sort(key=lambda tup: tup[0])
            return result

        def _format_repo(r):
            """format info for a single repo."""
            # pylint can't see the nested {width}
            # pylint:disable=unused-format-string-argument
            return ("{name:<{width}} {perm} {charset:<10} {desc}"
                    .format(width=width,
                            name=p4gf_translate.TranslateReponame.repo_to_git(r[0]),
                            perm=r[1],
                            charset=r[2],
                            desc=r[3]))

        try:
            self.server.check_user_exists()

            repos = RepoList.list_for_user(self.p4, self.user).repos
            if self.foruser:
                repos2 = RepoList.list_for_user(self.p4, self.foruser).repos
                repos = _merge_lists(repos, repos2)
            if len(repos):
                width = max(len(r[0]) for r in repos)
                sys.stderr.write("\n".join([_format_repo(r) for r in repos]) + "\n")
            else:
                sys.stderr.write(_('no repositories found\n'))
        except RuntimeError:
            LOG.exception('repo list retrieval failed')
            sys.stderr.write(_('no repositories found\n'))


class SpecialCommandProcesses(SpecialCommandHandler):

    """Processes the '@ps' special command."""

    def run(self):
        """Print a list of active pull/push operations."""
        self.server.check_user_exists()
        # collect list of active pull/push operations
        requests = query_active_requests(self.p4)
        # filter that list by user permissions
        repos = []
        for repo_name in requests.keys():
            user = self.foruser if self.foruser else self.user
            if self.server.check_permissions(p4gf_group.PERM_PULL, repo_name, user):
                repos.append(repo_name)
        # print the resulting details in sorted order
        for name in sorted(repos):
            repo_requests = requests[name]
            for request in repo_requests:
                sys.stderr.write('{0} {1} {2}\n'.format(name, request.action, request.process_id))


class SpecialCommandFeatures(SpecialCommandHandler):

    """Processes the '@features' special command."""

    def run(self):
        """Print a list of all available features."""
        # pylint:disable=no-self-use
        sys.stderr.write(_('Available features:\n'))
        for k in p4gf_config.configurable_features():
            sys.stderr.write("{} : {}\n".format(k, p4gf_config.FEATURE_KEYS[k]))


class SpecialCommandHandlerRepo(SpecialCommandHandler):

    """Base class for special commands targeting a single repo.

    Some (but not all) of these commands also take wait_id and active_id.
    """
    #  pylint: disable=too-many-arguments
    def __init__(self, server, repo_name, wait_id, active_id, timeout=None):
        """Init the handler."""
        SpecialCommandHandler.__init__(self, server)
        self.repo_name = repo_name
        self.wait_push_id = wait_id
        self.active_push_id = active_id
        self.timeout = timeout


class SpecialCommandChangelist(SpecialCommandHandlerRepo):

    """Processes the '@cl@value' special command."""

    def run(self):
        """Map a changelist to a SHA1 and vice versa."""
        # self.repo_name is either a SHA1 or a changelist number
        not_a_change = False
        try:
            cl = int(self.repo_name)
            # it is likely a change number
            try:
                # extract the SHA1 from the change description
                change = self.p4.fetch_change(cl)
                di = DescInfo.from_text(change['Description'])
                if di:
                    sys.stderr.write('change {} => commit {}\n'.format(cl, di.sha1))
                else:
                    sys.stderr.write('change {} has no commit information\n'.format(cl))
            except P4Exception:
                if p4gf_p4msg.find_msgid(self.p4, p4gf_p4msgid.MsgDm_NoSuchChange):
                    not_a_change = True
                else:
                    LOG.exception('unexpected p4 error')
        except ValueError:
            # it must be a SHA1 value
            not_a_change = True
        if not_a_change:
            # search the key/values for our change index to find the SHA1
            found_commit = False

            def key_callback(key):
                """Invoked for each key dict from p4 keys."""
                if key['value'] == self.repo_name:
                    (_pre, cl, _suf) = key['key'].split(',')
                    sys.stderr.write('commit {} => change {}\n'.format(self.repo_name, cl))
                    nonlocal found_commit
                    found_commit = True
                    return True

            P4Key.query_keys(self.p4, 'git-fusion-index-branch-*', key_callback)
            if not found_commit:
                sys.stderr.write('commit {} not found\n'.format(self.repo_name))


class SpecialCommandProgressRepo(SpecialCommandHandlerRepo):

    """Processes the '@progress@repo' special command."""

    def run(self):
        """Wait for a repo lock and report changes to the repo status."""
        if self.server.check_permissions(p4gf_group.PERM_PULL, self.repo_name):
            _wait_for_push(self.p4, self.repo_name, self.wait_push_id, self.active_push_id,
                           progress=True)
        else:
            LOG.warning("@progress unable to report status for {} to {}".format(
                self.repo_name, self.user))
            sys.stderr.write(_('Status not available due to permissions\n'))


class SpecialCommandWaitRepo(SpecialCommandHandlerRepo):

    """Processes the '@wait@repo' special command."""

    def run(self):
        """Wait for repo lock to be released."""
        if self.server.check_permissions(p4gf_group.PERM_PULL, self.repo_name):
            _wait_for_push(self.p4, self.repo_name, self.wait_push_id,
                           self.active_push_id, timeout=self.timeout)
        else:
            LOG.warning("@wait unable to report status for {} to {}".format(
                self.repo_name, self.user))
            sys.stderr.write(_('Status not available due to permissions\n'))


class SpecialCommandFeatureRepo(SpecialCommandHandlerRepo):

    """Processes the '@features@repo' special command."""

    def run(self):
        """Report which features are enabled for a repo."""
        config = p4gf_config.RepoConfig.from_depot_file(self.repo_name, self.p4)
        sys.stderr.write(_("Enabled features for repo '{repo_name}':\n")
                         .format(repo_name=self.repo_name))
        for k in p4gf_config.configurable_features():
            sys.stderr.write("{} : {}\n".format(
                k, config.is_feature_enabled(k)))


class SpecialCommandStatusRepo(SpecialCommandHandlerRepo):

    """Processes the '@status@repo' special command."""

    def run(self):
        """Show the status of a push operation."""
        if self.server.check_permissions(p4gf_group.PERM_PULL, self.repo_name):
            msg = _status_for_push(self.p4, self.repo_name, self.wait_push_id, self.active_push_id)
            sys.stderr.write("{}\n".format(msg))
        else:
            LOG.warning("@status denied for {} on {}".format(self.user, self.repo_name))
            sys.stderr.write(_('Permission denied for @status\n'))


class SpecialCommandConfigRepo(SpecialCommandHandlerRepo):

    """Dump the configuration of the named repository."""

    def run(self):
        """Dump the repository configuration."""
        if self.server.check_permissions(p4gf_group.PERM_PULL, self.repo_name):
            try:
                config = p4gf_config.RepoConfig.from_depot_file(self.repo_name, self.p4)
                config_content = p4gf_config.to_text('', config.repo_config)
                sys.stderr.write(config_content)
            except p4gf_config.ConfigLoadError as err:
                sys.stderr.write(_('Unable to load configuration: {}\n').format(err))
        else:
            LOG.warning("@config denied for {} on {}".format(self.user, self.repo_name))
            sys.stderr.write(_('Permission denied for @config\n'))


class SpecialCommandMirrorWait(SpecialCommandHandler):

    """Processes the '@mirror_wait' special command."""

    def run(self):
        """Warn user that @mirror_wait is no longer implemented."""
        # pylint:disable=no-self-use
        sys.stderr.write(_('@mirror_wait is no longer implemented, try @wait instead\n'))


def _status_for_push(p4, repo_name, wait_push_id, active_push_id):
    """Return the status for the requested push, or something reasonable.

    :param p4: instance of P4API.
    :param repo_name: name fo the repository for which to retrieve status.
    :param wait_push_id: push identifier of interest.
    :param active_push_id: identifier of the active push.

    """
    # check for the push-specific status key, if any
    n_key_name = P4Key.calc_repo_status_p4key_name(repo_name, wait_push_id)
    result = P4Key.get(p4, n_key_name)
    if result and result != '0':
        return result
    # otherwise, retrieve the current status, if appropriate
    if wait_push_id == active_push_id:
        key_name = P4Key.calc_repo_status_p4key_name(repo_name)
        result = P4Key.get(p4, key_name)
        if result and result != '0':
            return result
    # and if that didn't work, just return some generic message
    if wait_push_id != '0':
        # try to keep this text the same as in record_push_success_p4key()
        return _('Push {push_id} completed successfully').format(
            push_id=wait_push_id)
    return _('No status available')


def _parse_push_identifiers(p4, repo_name):
    """Return a tuple of adjusted repository name, requested push id, and active push id.

    :param p4: instance of P4API.
    :param repo_name: name fo the repository for which to retrieve status; may contain
                      the requested push identifier (e.g. repo_name@push_id).

    Returns a 3-tuple of repository name, requested push identifer, and active push id.
    If no push identifier was requested, the active push identifier is substituted.
    The returned repository name will also have been translated.

    """
    # pylint: disable=too-many-branches
    if '@' in repo_name:
        # (repo_name, wait_push_id) = repo_name.rsplit('@', 1)
        parms = repo_name.rsplit('@', 2)
        repo_name = parms[0]
        wait_push_id = None
        timeout = None
        for p in parms[1:]:
            if p.startswith('timeout='):
                timeout = p[8:]
                try:
                    timeout = int(timeout)
                    if timeout < 0:
                        raise ValueError()
                except ValueError:
                    raise RuntimeError(_("Timeout period must be a non-negative integer."))
            elif '=' in p:
                raise RuntimeError(_("Improperly formatted @wait command: @wait{}".format(
                    repo_name)))
            else:
                wait_push_id = p
                try:
                    if int(wait_push_id) < 1:
                        raise ValueError()
                except ValueError:
                    raise RuntimeError(_("Push identifier must be a positive integer."))
        # we have both push_id and timeout
        if len(parms) > 2 and not wait_push_id and not timeout:
            raise RuntimeError(_("Improperly formatted @wait command: @wait{}".format(repo_name)))
    else:
        wait_push_id = None
        timeout = None

    repo_name = p4gf_translate.TranslateReponame.url_to_repo(repo_name, p4)
    key_name = P4Key.calc_repo_push_id_p4key_name(repo_name)
    active_push_id = P4Key.get(p4, key_name)
    if wait_push_id is None:
        wait_push_id = active_push_id
    elif int(wait_push_id) > int(active_push_id):
        raise RuntimeError(_("Push {push_id} unrecognized\n")
                           .format(push_id=wait_push_id))
    return repo_name, wait_push_id, active_push_id, timeout


def _wait_for_push(p4, repo_name, wait_push_id, active_push_id, progress=False, timeout=None):
    """Wait for the completion of a push on a particular repository.

    :param p4: P4API instance.
    :param str repo_name: name of repository.
    :param int wait_push_id: push identifier on which to wait.
    :param int active_push_id: identifier of active push.
    :param bool progress: True to display progress of push periodically.

    """
    # pylint: disable=too-many-arguments
    if wait_push_id == active_push_id:
        LOG.debug("checking for active lock on {}".format(repo_name))
        # quick check to see if the lock is held or not
        lock_key_name = p4gf_const.P4GF_P4KEY_LOCK_VIEW_OWNERS.format(repo_name=repo_name)
        lock_value = P4Key.get(p4, lock_key_name)
        if lock_value and lock_value != '0':
            LOG.info("waiting for lock on {}".format(repo_name))
            # wait for either the lock status or the active push id to change
            sys.stderr.write(_("Waiting for push {push_id}...\n".format(push_id=wait_push_id)))
            sys.stderr.flush()
            key_name = P4Key.calc_repo_push_id_p4key_name(repo_name)
            label = "@wait on {}".format(repo_name)
            wait_reporter = p4gf_log.LongWaitReporter(label, LOG)
            status_reporter = StatusReporter(p4, repo_name, wait_push_id, active_push_id)
            start_time = time.time()
            while lock_value and lock_value != '0' and wait_push_id == active_push_id:
                if timeout is not None:
                    lapsed_time = int(time.time() - start_time)
                    if lapsed_time > timeout:
                        LOG.debug("@wait returning after timeout={} (actual={})"
                                  " and before push changed or lock released for {}".
                                  format(timeout, lapsed_time, repo_name))
                        break
                time.sleep(1)
                lock_value = P4Key.get(p4, lock_key_name)
                active_push_id = P4Key.get(p4, key_name)
                wait_reporter.been_waiting()
                if progress:
                    status_reporter.maybe_report()
            LOG.info("active push changed or lock released for {}".format(
                repo_name))
    # retrieve the status while we're here
    msg = _status_for_push(p4, repo_name, wait_push_id, active_push_id)
    sys.stderr.write("{}\n".format(msg))


def _set_pre_receive_flag(ctx):
    """Write a file that pre-receive hook should delete when it runs."""
    fname = os.path.join(ctx.repo_dirs.repo_container, p4gf_const.P4GF_PRE_RECEIVE_FLAG)
    with open(fname, 'w') as fobj:
        fobj.write(str(os.getpid()))


def _detect_pre_receive_flag(ctx):
    """Check for the file that indicates whether pre-receive ran or not.

    :return: True if file still exists, indicating pre-receive did _not_ run.

    """
    fname = os.path.join(ctx.repo_dirs.repo_container, p4gf_const.P4GF_PRE_RECEIVE_FLAG)
    return os.path.exists(fname)


def p4_has_login(user, port, client):
    """Return login state for user."""
    login = None
    msg = ''
    login_p4 = p4gf_create_p4.create_p4(port=port, user=user, client=client)
    if not login_p4:
        raise RuntimeError(_("Failed to connect to P4."))
    try:
        login = login_p4.run('login', '-s')[0]
        if 'TicketExpiration' in login:
            login = True
            msg = 'TicketExpiration'
        else:
            login = True
            msg = 'No password'
    except P4Exception as e:
        lines = str(e).splitlines()
        msg = ""
        login = None
        for l in lines:
            if 'Perforce password (P4PASSWD) invalid or unset.' in l or \
               'Your session was logged out, please login again.' in l or \
               'Your session has expired, please login again.' in l:
                msg = l
                break
        if msg:
            login = False
        else:
            emsg = _("Failed to check login status for user={0}\n{1}").format(user, str(e))
            raise RuntimeError(emsg)

    p4gf_create_p4.p4_disconnect(login_p4)
    return (login, msg)


# ActiveRequest represents an active request in Git Fusion.
ActiveRequest = namedtuple('ActiveRequest', ['action', 'process_id'])


def query_active_requests(p4):
    """Query the active requests and return a dict of repos and details.

    :param p4: P4API object with which to query p4key values.
    :return: map of repo names to sets of ActiveRequest.
    :rtype: dict

    """
    # Use a defaultdict of sets to ensure we only include a request once
    # for each process and behavior. Since a process may have both a p4key
    # lock and a file lock, we want to count that as one 'writer' entry and
    # not as two separate requests.
    results = defaultdict(set)

    # All p4key based locks are considered writers since the owner can make
    # changes to anything related to the repository.
    pattern = p4gf_const.P4GF_P4KEY_LOCK_VIEW_OWNERS.format(repo_name='*')
    repo_re = re.compile(pattern.replace('*', '(?P<repo>.+)'))
    repo_locks = P4Key.get_all(p4, pattern)
    for name, raw_value in repo_locks.items():
        repo_name = repo_re.match(name).group('repo')
        content = json.loads(raw_value)
        for owner in content['owners']:
            results[repo_name].add(ActiveRequest('writer', owner['process_id']))

    # File-based locks are either writers or readers, and we can know that
    # by the name of the file. The name of the file or its contents gives
    # us the process identifier.
    lock_dir_name = '{0}/locks'.format(p4gf_const.P4GF_HOME)
    if os.path.isdir(lock_dir_name):
        for repo_name in os.listdir(lock_dir_name):
            repo_path = os.path.join(lock_dir_name, repo_name)
            if os.path.isdir(repo_path):
                for lock_name in os.listdir(repo_path):
                    if lock_name == 'write':
                        with open(os.path.join(repo_path, lock_name)) as fobj:
                            pids = fobj.read().splitlines()
                            for writer_pid in pids:
                                results[repo_name].add(ActiveRequest('writer', writer_pid))
                    elif lock_name == 'lock':
                        continue
                    else:
                        results[repo_name].add(ActiveRequest('reader', lock_name))

    return results
