#! /usr/bin/env python3.3
"""Functions to support storing and reconstituting Git tags."""

import logging
import os
import re
import sys
import zlib

import P4
import pygit2

import p4gf_const
import p4gf_git
import p4gf_p4key as P4Key
from p4gf_l10n import _, NTR
from p4gf_profiler import with_timer
import p4gf_util

LOG = logging.getLogger(__name__)
_BITE_SIZE = 1000  # How many files to pass in a single 'p4 xxx' operation.


def _client_path(ctx, sha1):
    """Construct the client path for the given tag object.

    For example, 5716ca5987cbf97d6bb54920bea6adde242d87e6 might return as
    objects/repos/foobar/tags/57/16/ca5987cbf97d6bb54920bea6adde242d87e6
    """
    return os.path.join("objects", "repos", ctx.config.repo_name, "tags",
                        sha1[:2], sha1[2:4], sha1[4:])


def _add_tag(ctx, name, sha1, edit_list, add_list):
    """Add a tag to the object cache.

    If adding another lightweight tag that refers to the same object,
    edit the file rather than add.
    """
    LOG.debug("adding tag %s", name)
    git_obj = ctx.repo.get(sha1)
    is_annotated = git_obj.type == pygit2.GIT_OBJ_TAG
    tag_of_a_tag = is_annotated and git_obj.name != name
    fpath = os.path.join(ctx.gitlocalroot, _client_path(ctx, sha1))
    if tag_of_a_tag:
        fpath = fpath + '.txt'
    if os.path.exists(fpath):
        if is_annotated and not tag_of_a_tag:
            LOG.warning("cannot overwrite annotated tag with itself %s", name)
            return
        with open(fpath, 'ab') as f:
            f.write(b'\n')
            f.write(name.encode('UTF-8'))
        edit_list.append(fpath)
    else:
        if is_annotated and not tag_of_a_tag:
            LOG.debug("annotated tag %s added via %s", name, fpath)
            p4gf_git.write_git_object_from_sha1(ctx.repo, sha1, fpath)
        else:
            # Lightweight tags can be anything: commit, tree, blob, annotated tag
            LOG.debug("lightweight tag %s added via %s", name, fpath)
            p4gf_util.ensure_parent_dir(fpath)
            with open(fpath, 'wb') as f:
                f.write(name.encode('UTF-8'))
        add_list.append(fpath)


def _remove_tag(ctx, name, sha1, edit_list, delete_list):
    """Remove the tag from the object cache.

    If removing one of several lightweight tags which reference the same object,
    the corresponding file will be edited rather than deleted.
    """
    LOG.debug("removing tag %s", name)
    fpath = os.path.join(ctx.gitlocalroot, _client_path(ctx, sha1))
    fpath_txt = fpath + '.txt'

    def _remove_lw_tag_entry(contents):
        """Remove the tag name from the file contents."""
        tag_names = contents.decode('UTF-8').splitlines()
        if name in tag_names:
            tag_names.remove(name)
            if tag_names:
                contents = '\n'.join(tag_names).encode('UTF-8')
                os.chmod(fpath, 0o644)
                with open(fpath, 'wb') as f:
                    f.write(contents)
                edit_list.append(fpath)
                LOG.debug("removed tag %s from %s", name, fpath)
            else:
                delete_list.append(fpath)
                LOG.debug("removed tag %s, deleting %s", name, fpath)
            return True
        return False

    # New behavior of looking for a .txt file that contains the names of
    # lightweight tags of annotated tags. If such a file exists, and the
    # name is inside, remove it and edit/delete the .txt file.
    if os.path.exists(fpath_txt):
        with open(fpath, 'rb') as f:
            contents = f.read()
        if _remove_lw_tag_entry(contents):
            return

    # Otherwise we fall back to the original behavior, in which we could
    # never have a lightweight tag of an annotated tag, and handling
    # tag removal is very simple.
    if not os.path.exists(fpath):
        # Already gone (or never stored), nothing else to do.
        LOG.debug("tag %s does not exist, nothing to do", name)
        return
    with open(fpath, 'rb') as f:
        contents = f.read()
    try:
        zlib.decompress(contents)
        # Must be an annotated tag...
        delete_list.append(fpath)
        LOG.debug("removed tag %s, deleting %s", name, fpath)
    except zlib.error:
        _remove_lw_tag_entry(contents)


@with_timer('preflight tags')
def preflight_tags(tags):
    """Validate the incoming tags.

    :param tags: list of PreReceiveTuple objects for tags

    Warnings are printed to stderr so the user knows about them.

    Returns None if successful and an error string otherwise.

    """
    if not tags:
        LOG.debug("preflight_tags() no incoming tags to process")
        return None
    regex = re.compile(r'[*@#,]|\.\.\.|%%')
    for prt in tags:
        tag = prt.ref[10:]
        # Screen the tags to ensure their names won't cause problems
        # sometime in the future (i.e. when we create Perforce labels).
        # Several of these characters are not allowed in Git tag names
        # anyway, but better to check in case that changes in the future.
        # In particular git disallows a leading '-'.
        if regex.search(tag) or tag.startswith('-'):
            return _("illegal characters (@#*,...%%) in tag name: '{tag}'").format(tag=tag)
        if prt.old_sha1 == p4gf_const.NULL_COMMIT_SHA1:
            if prt.new_sha1 == p4gf_const.NULL_COMMIT_SHA1:
                # No idea how this happens, but it did, so guard against it.
                sys.stderr.write(_('Ignoring double-zero pre-receive-tuple line'))
                continue
            # We used to prohibit various types of tags, now we let them
            # through as-is and restore on clone.
        elif prt.new_sha1 != p4gf_const.NULL_COMMIT_SHA1:
            # Older versions of Git allowed moving a tag reference, while
            # newer ones seemingly do not. We will take the new behavior as
            # the correct one and reject such changes.
            return _('Updates were rejected because the tag already exists in the remote.')


@with_timer('process tags')
def process_tags(ctx, tags):
    """Add or remove tags objects from the Git Fusion mirror.

    :param ctx: P4GF context with initialized pygit2 Repository.
    :param tags: list of PreReceiveTuple objects for tags

    """
    # pylint:disable=too-many-branches
    if not tags:
        LOG.debug("process_tags() no incoming tags to process")
        return

    # Re-sync the tags since preflight_tags() synced with a different temp client.
    tags_path = "objects/repos/{repo}/tags".format(repo=ctx.config.repo_name)
    with ctx.p4gf.at_exception_level(P4.P4.RAISE_NONE):
        # Raises an exception when there are no files to sync?
        ctx.p4gfrun('sync', '-q', "//{}/{}/...".format(ctx.p4gf.client, tags_path))

    # Decide what to do with the tag references.
    tags_to_delete = []
    tags_to_add = []
    tags_to_edit = []
    for prt in tags:
        tag = prt.ref[10:]
        if prt.old_sha1 == p4gf_const.NULL_COMMIT_SHA1:
            if prt.new_sha1 == p4gf_const.NULL_COMMIT_SHA1:
                # No idea how this happens, but it did, so guard against it.
                continue
            _add_tag(ctx, tag, prt.new_sha1, tags_to_edit, tags_to_add)
        elif prt.new_sha1 == p4gf_const.NULL_COMMIT_SHA1:
            # Removing an existing tag
            _remove_tag(ctx, tag, prt.old_sha1, tags_to_edit, tags_to_delete)

    # Seemingly nothing to do.
    if not tags_to_add and not tags_to_edit and not tags_to_delete:
        LOG.debug("process_tags() came up empty, nothing to do")
        return

    # Add and remove tags as appropriate, doing so in batches.
    LOG.info("adding %s tags, removing %s tags, editing %s tags from Git mirror",
             len(tags_to_add), len(tags_to_delete), len(tags_to_edit))
    desc = _("Git Fusion '{repo}' tag changes").format(repo=ctx.config.repo_name)
    with p4gf_util.NumberedChangelist(gfctx=ctx, description=desc) as nc:
        while len(tags_to_add):
            bite = tags_to_add[:_BITE_SIZE]
            tags_to_add = tags_to_add[_BITE_SIZE:]
            ctx.p4gfrun('add', '-t', 'binary+F', bite)
        while len(tags_to_edit):
            bite = tags_to_edit[:_BITE_SIZE]
            tags_to_edit = tags_to_edit[_BITE_SIZE:]
            ctx.p4gfrun('edit', '-k', bite)
        while len(tags_to_delete):
            bite = tags_to_delete[:_BITE_SIZE]
            tags_to_delete = tags_to_delete[_BITE_SIZE:]
            ctx.p4gfrun('delete', bite)
        nc.submit()
        if nc.submitted:
            _write_last_copied_tag(ctx, nc.change_num)
    LOG.debug("process_tags() complete")


def _calc_last_copied_tag_p4key_name(repo_name, server_id):
    """Return a name of a p4key that holds the changelist number of the most recently
    updated tag on the given Git Fusion server.
    """
    return p4gf_const.P4GF_P4KEY_LAST_COPIED_TAG.format(repo_name=repo_name, server_id=server_id)


def _last_copied_tag_p4key_name(ctx):
    """Return the name of a p4key that holds the latest tag changelist
    number for this Git Fusion server.
    """
    return _calc_last_copied_tag_p4key_name(ctx.config.repo_name, p4gf_util.get_server_id())


def _read_last_copied_tag(ctx):
    """Return the changelist number for the most recent tags change for this
    Git Fusion server.
    """
    return P4Key.get(ctx.p4gf, _last_copied_tag_p4key_name(ctx))


def _write_last_copied_tag(ctx, change_num):
    """Update the changelist number for the most recent tags change for this
    Git Fusion server.
    """
    return P4Key.set(ctx.p4gf, _last_copied_tag_p4key_name(ctx), change_num)


def any_tags_since_last_copy(ctx):
    """Return the first tags change not yet copied between Git and Perforce.

    If a tags change exists and is > last copied tags change, return it.
    Else return None.
    """
    last_copied_change = _read_last_copied_tag(ctx)
    if not last_copied_change:
        return None
    tags_path = '{root}/repos/{repo}/tags/...'.format(
        root=p4gf_const.objects_root(),
        repo=ctx.config.repo_name)
    head_change = p4gf_util.head_change_as_string(ctx, submitted=True, view_path=tags_path)

    if head_change and int(head_change) > int(last_copied_change):
        pass  # return this head_change
    else:
        head_change = None
    LOG.debug('any_tags_since_last_copy() found new tags: %s', head_change)
    return head_change


def _create_tag_ref(name, sha1):
    """Create a single tag reference in the repository."""
    if not name or not sha1:
        LOG.warning("_create_tag_ref() invalid params: (%s, %s)", name, sha1)
        return
    # Do _not_ check for the presence of the object referred to by `sha1`
    # as sometimes we process the lightweight tag of an object before the
    # object itself is received (e.g. when we get a lightweight tag of an
    # annotated tag).
    tag_file = os.path.join('.git', 'refs', 'tags', name)
    p4gf_util.ensure_parent_dir(tag_file)
    with open(tag_file, 'w') as f:
        f.write(sha1)


def _remove_tag_ref(name, sha1):
    """Remove a single tag reference from the repository."""
    if not name or not sha1:
        LOG.warning("_remove_tag_ref() invalid params: (%s, %s)", name, sha1)
        return
    tag_file = os.path.join('.git', 'refs', 'tags', name)
    if os.path.exists(tag_file):
        os.unlink(tag_file)


def _install_tag(repo, fname):
    """Given the path of a tag copied from Perforce object cache, copy
    the tag to the repository, with the appropriate name and SHA1.
    There may be multiple lightweight tags associated with the same
    SHA1, in which case multiple tags will be created.

    Arguments:
        repo -- pygit2 repository
        fname -- clientFile attr for sync'd tag
    """
    if fname.endswith('.txt'):
        sha1 = fname[-46:-4].replace('/', '')
    else:
        sha1 = fname[-42:].replace('/', '')
    LOG.debug("_install_tag() examining %s...", sha1)
    with open(fname, 'rb') as f:
        contents = f.read()
    try:
        zlib.decompress(contents)
        # Must be an annotated tag...
        blob_path = os.path.join('.git', 'objects', sha1[:2], sha1[2:])
        p4gf_util.ensure_parent_dir(blob_path)
        os.link(fname, blob_path)
        tag_obj = repo.get(sha1)
        tag_name = tag_obj.name
        LOG.debug("_install_tag() annotated tag %s", tag_name)
        _create_tag_ref(tag_name, sha1)
    except zlib.error:
        # Lightweight tags are stored simply as the tag name, but
        # there may be more than one name for a single SHA1.
        tag_names = contents.decode('UTF-8')
        for name in tag_names.splitlines():
            LOG.debug("_install_tag() lightweight tag %s", name)
            _create_tag_ref(name, sha1)


def _uninstall_tag(repo, fname):
    """Given the path of a tag copied from Perforce object cache, remove
    the tag from the repository.

    Arguments:
        repo -- pygit2 repository
        fname -- clientFile attr for sync'd tag
    """
    if fname.endswith('.txt'):
        sha1 = fname[-46:-4].replace('/', '')
    else:
        sha1 = fname[-42:].replace('/', '')
    LOG.debug("_uninstall_tag() examining %s...", sha1)
    with open(fname, 'rb') as f:
        contents = f.read()
    try:
        zlib.decompress(contents)
        # Must be an annotated tag...
        tag_obj = repo.get(sha1)
        tag_name = tag_obj.name
        LOG.debug("_uninstall_tag() annotated tag %s", tag_name)
    except zlib.error:
        # Lightweight tags are stored simply as the tag name
        tag_name = contents.decode('UTF-8')
        LOG.debug("_uninstall_tag() lightweight tag %s", tag_name)
    # Remove the tag reference
    tag_refs = os.path.join('.git', 'refs', 'tags')
    if not os.path.exists(tag_refs):
        return
    tag_file = os.path.join(tag_refs, tag_name)
    if os.path.exists(tag_file):
        os.unlink(tag_file)


def _read_tags(ctx, depot_path, rev=None):
    """Return the set of (lightweight) tag names read from the given file.

    :type ctx: :class:`p4gf_context.Context`
    :param ctx: Git Fusion context

    :type depot_path: str
    :param depot_path: file path within depot

    :type rev: int
    :param rev: if not None, specifies file revision to retrieve

    """
    if rev:
        depot_tag_path =  "{}#{}".format(depot_path, rev)
    else:
        depot_tag_path = depot_path
    local_tag_path = p4gf_util.print_depot_objects_to_object_root(ctx, depot_tag_path)
    with open(local_tag_path, 'rb') as f:
        contents = f.read()
    tag_names = contents.decode('UTF-8').splitlines()
    return set(tag_names)


@with_timer('update tags')
def update_tags(ctx):
    """Based on the recent changes to the tags, update our repository
    (remove deleted tags, add new pushed tags).
    """
    last_copied_change = _read_last_copied_tag(ctx)
    tags_prefix = '{root}/repos/{repo}/tags/'.format(
        root=p4gf_const.objects_root(), repo=ctx.config.repo_name)
    tags_path = '{prefix}...'.format(prefix=tags_prefix)
    num = 1 + int(last_copied_change)
    r = ctx.p4gfrun('changes', '-s', 'submitted', '-e', num, tags_path)
    changes = sorted(r, key=lambda k: int(k['change']))
    for change in changes:
        LOG.debug2('update_tags() processing %s', change['change'])
        d = ctx.p4gfrun('describe', change['change'])
        d = p4gf_util.first_dict(d)
        for d_file, rev, action in zip(d['depotFile'], d['rev'], d['action']):
            LOG.debug2('update_tags() processing %s - %s - %s', d_file, action, rev)
            if not d_file.startswith(tags_prefix):
                LOG.info('non-tag file %s in tag-related Git Fusion change %s',
                         d_file, change['change'])
                continue
            if action == 'add':
                depot_tag_path =  "{}#{}".format(d_file, rev)
                local_tag_path = p4gf_util.print_depot_objects_to_object_root(ctx, depot_tag_path)
                _install_tag(ctx.repo, local_tag_path)
            elif action == 'delete':
                depot_tag_path =  "{}#{}".format(d_file, int(rev) - 1)
                local_tag_path = p4gf_util.print_depot_objects_to_object_root(ctx, depot_tag_path)
                _uninstall_tag(ctx.repo, local_tag_path)
            elif action == 'edit':
                # get the tags named in the file prior to this change
                tags_before = _read_tags(ctx, d_file, int(rev) - 1)
                # get the tags named in the file after this change
                tags_after = _read_tags(ctx, d_file)
                # remove old (lightweight) tags and add new ones
                sha1 = d_file[-42:].replace('/', '')
                for old_tag in tags_before - tags_after:
                    _remove_tag_ref(old_tag, sha1)
                for new_tag in tags_after - tags_before:
                    _create_tag_ref(new_tag, sha1)
            else:
                LOG.error("update_tags() received an unexpected change action: " +
                          "@%s, '%s' on %s", change['change'], action, d_file)
    _write_last_copied_tag(ctx, changes[-1]['change'])


@with_timer('generate tags')
def generate_tags(ctx):
    """Regenerate the original tags into the (rebuilt) Git repository.

    This should only be called when the repository was just rebuilt
    from Perforce, otherwise it will do a bunch of work for nothing.
    """
    # Fetch everything under //.git-fusion/objects/repos/<repo>/tags/...
    tags_path = NTR('objects/repos/{repo}/tags').format(repo=ctx.config.repo_name)
    depot_tags_path = '//{}/{}/...'.format(p4gf_const.P4GF_DEPOT, tags_path)
    p4gf_util.print_depot_objects_to_object_root(ctx, depot_tags_path)

    # Walk the tree looking for tags, reconstituting those we encounter.
    tags_root = os.path.join(ctx.gitlocalroot, tags_path)
    for walk_root, _dirs, files in os.walk(tags_root):
        for name in files:
            fname = os.path.join(walk_root, name)
            _install_tag(ctx.repo, fname)

    # Update the tag change p4key to avoid repeating our efforts.
    head_change = any_tags_since_last_copy(ctx)
    if head_change:
        LOG.debug("generate_tags: head_change %s", head_change)
        _write_last_copied_tag(ctx, head_change)
