#!/usr/bin/env bash

set -e
#check to make sure this is a 64-bit machine
getconf LONG_BIT | grep 64 > /dev/null  || ( echo "Git Fusion requires a 64-bit machine. Exiting." && exit 1)

echo ''
echo 'This script will upgrade pygit2 and libgit2 to version 0.22.0 for Git Fusion'
echo 'It will also install cffi and setuptool modules (build prerequisites) for the existing Python3 installation'
echo 'It is not intended for use with package installs'
echo ''
echo "Do you wish to continue?"
select yn in "Yes" "No"; do
    case $yn in
        Yes ) break;;
        No ) exit;;
    esac
done

# Enforce /usr/local/bin for RHEL/Centos because the install script does. 
# All else can use what is found in the environment
PYTHON33="python3.3"
if [ -e "/etc/redhat-release" ]; then
    PYTHON33="/usr/local/bin/python3.3"
    # This is a CentOS or RedHat
    CENTOS=1
fi

if [ -e /opt/perforce/git-fusion/bin/python3.3 ]; then
    echo 'Found: /opt/perforce/git-fusion/bin/python3.3'
    echo 'Package install of Git Fusion detected.'
    echo 'Use the package installer to upgrade perforce-3rdparty-pygit2 & perforce-3rdparty-libgit2 packages'
    exit 1
fi

PYTHON_VER=`$PYTHON33 -c 'import sys; print(sys.version_info[:])' | 
    awk -F'(' '{print $2}' | awk -F', ' '{print "Python "$1"."$2"."$3}'`
if [ "$PYTHON_VER" != "Python 3.3.2" ]; then
        echo 'Expected Python 3.3.2'
        echo 'Found: '$PYTHON_VER
    exit 1
fi

if [ ! -e build_pygit2 ]; then
    mkdir build_pygit2
else
    sudo rm -rf build_pygit2
    mkdir build_pygit2
fi
cd build_pygit2

if [ ! -e build_prereq ]; then
    mkdir build_prereq
else
    sudo rm -rf build_prereq
    mkdir build_prereq
fi
cd build_prereq

echo 'Checking prerequisites...'
if [ -n "$CENTOS" ]
then
    sudo yum -y install libffi-devel cmake
else
    sudo apt-get -y install libffi-dev
    CMAKE=`which cmake`
    if [ -z "$CMAKE" ]
    then
        wget http://www.cmake.org/files/v2.8/cmake-2.8.3.tar.gz
        tar xzf cmake-2.8.3.tar.gz
        cd cmake-2.8.3
        ./configure
        make
        sudo make install
        cd ..
    fi
fi

echo 'Grabbing the required versions of cffi and setuptools...'

wget --no-check-certificate https://pypi.python.org/packages/source/s/setuptools/setuptools-14.3.tar.gz
wget --no-check-certificate https://pypi.python.org/packages/source/c/cffi/cffi-0.9.2.tar.gz

echo 'Building setuptools and cffi...'
tar xzvf setuptools-14.3.tar.gz
tar xzvf cffi-0.9.2.tar.gz
cd setuptools-14.3
sudo $PYTHON33 setup.py install
cd ../cffi-0.9.2
sudo $PYTHON33 setup.py install
cd ..

echo 'Verifying install of setuptools and cffi...'
set +e
$PYTHON33 -c 'import setuptools'
if [ $? -ne 0 ]
then
    echo 'Error with setuptools install'
    exit 1
fi
$PYTHON33 -c 'import cffi'
if [ $? -ne 0 ]
then
    echo 'Error with cffi install'
    exit 1
fi

set -e
echo 'Grabbing the required versions of libgit2 and pygit2...'
git clone https://github.com/libgit2/libgit2.git
mkdir -p libgit2/build
cd libgit2
git checkout v0.22.0
cd build
cmake ..
cmake --build .
sudo cmake --build . --target install
cd ../..

git clone https://github.com/libgit2/pygit2.git
cd pygit2
git checkout v0.22.0
export LIBGIT2="/usr/local"
export LDFLAGS="-Wl,-rpath='$LIBGIT2/lib',--enable-new-dtags $LDFLAGS"
$PYTHON33 setup.py build
sudo $PYTHON33 setup.py install
$PYTHON33 setup.py test
cd ..

echo 'Testing the installation...'
set +e
PYGIT2_VERSION="$($PYTHON33 -c 'import pygit2; print(pygit2.__version__)' | grep 0.22.0)"
if [ -n "$PYGIT2_VERSION" ]
then
    echo ''
    echo 'You have successfully upgraded pygit2 and libgit2 to version 0.22.0!'
    echo ''
else
    echo 'pygit2 version is not 0.22.0'
    exit 1
fi
